--
-- OclType
--

lin OclType = mkClassNoHum ["OCL type"] "OclType";

    class2oclType c = constant (class2id c);

    otName t = np2inst (appFunSg (funOfReg "name") t);
    otAttributes t = np2inst (appFunSg (funOfReg ["set of the attributes"]) t);
    otOperations t = np2inst (appFunSg (funOfReg ["set of the operations"]) t);
    otSupertypes t = np2inst (appFunSg (funOfReg ["set of the direct supertypes"]) t);
    otAllSupertypes t = np2inst (appFunSg (funOfReg ["set of all supertypes"]) t);
    otAllInstancesStrict c t = np2inst (appFunSg (funOfReg ["set of all instances"]) t);

--
-- OclAny
--
lin OclAny = mkClassNoHum "object" "OclAny";

    conforms2OclAny c = ss (class2id c ++ ["is a subtype of OclAny"]);

    anyEq a b = s2sent ((inst2np a).s!NomP ++ ["is the same object as"]  ++ (inst2np b).s!NomP);
    anyNeq a b = s2sent ((inst2np a).s!NomP ++ ["is not the same object as"]  ++ (inst2np b).s!NomP);
    anyOclIsKindOf a t = s2sent ( ["one of the (super-)types of"] ++ (inst2np a).s!AccP ++ 
         ["is equal to"] ++ (inst2np t).s!NomP);
    anyOclIsTypeOf a t = s2sent ( ["one of the types of"] ++ (inst2np a).s!AccP ++ 
         ["is equal to"] ++ (inst2np t).s!NomP);
    anyOclAsTypeStrict a c t = postfixIndFun (["considered as an instance of"] ++ (inst2np t).s!AccP) a;
    anyOclAsTypeClass a c = postfixIndFun (["considered as"] ++ artIndef ++ c.s!Sg!Nom) a;
--    anyOclAsTypeExpr a c _ = postfixIndFun (["considered as"] ++ artIndef ++ c.s!Sg!Nom) a;
    anyOclInState a s = s2sent ((inst2np a).s!NomP ++ ["is in state"] ++ (inst2np s).s!NomP);
    anyOclIsNew a = s2sent ((inst2np a).s!NomP ++ ["is created during the performing the operation"]);

--
-- OclState
--
lin OclState = mkClassNoHum "state" "OclState";


--
-- OclExpression
--

lin OclExpression = mkClassNoHum "expression" "OclExpression";

    inst2OclExpr _ i = i;  -- circular, of course

    oeEvaluationType i = np2inst (appFunSg (funOfReg ["type of"]) i);



--
-- Integer
--
lin Integer = mkClassNoHum  "integer" "Integer";
	int i = np2inst (numeral i.s);
	intAdd = instInfix "plus"; 
	intSub = instInfix "minus";
	intUnaryMinus = instFunOf ["negative value"];
	intTimes = instInfix "times"; 
	intDiv2Real = instInfix ["divided by"]; 
	intDiv2Int = instInfix ["integer divided by"];
	intMod = instInfix "modulo";

{- 
	-- These could be added as variants:
	
	intAdd a b = np2inst (appFunSg (funOfReg "sum") (conjoinNP "and" a b));  -- the sum of a and b
	intSub a b = np2inst (appFunSg (funXReg "difference" "between") (conjoinNP "and" a b)); 
	intTimes a b = np2inst (appFunSg (funOfReg "product") (conjoinNP "and" a b));
-}

	intAbs = instFunOf ["absolute value"];
	intMax a b = instFunOf "maximum" (conjoinInst "and" a b);
	intMin a b = instFunOf "minimum" (conjoinInst "and" a b);
	
	intEq = equalTo;
	intLT = lessThan;
	intGT = greaterThan;
	-- the following is not pretty, a better way?
	intLTE a b = let {ltb =complAdj {s = ["less than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	intGTE a b = let {ltb =complAdj {s = ["greater than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	
	Zero = numeral "zero";
	One = numeral "one";
	Two = numeral "two";
	
--
-- Real
--

lin
	Real = mkClassNoHum "real" "Real";
	realEq = equalTo;
	realNeq = notEqualTo;
	realAdd = instInfix "add"; 
	realSub = instInfix "sub"; 
	realTimes = instInfix "times"; 
	realDiv = instInfix ["divided by"];
	realUnaryMinus = instFunOf ["negative value"];
	realAbs = instFunOf ["absolute value"];
	realFloor a = np2inst (appFunSg (funXReg ["largest integer less than or equal"] "to") a);
	realRound a = np2inst (appFunSg (funXReg ["integer closest"] "to") a);
	realMax a b = instFunOf "maximum" (conjoinInst "and" a b);
	realMin a b = instFunOf "minimum" (conjoinInst "and" a b);
	realLT = lessThan; 
	realGT = greaterThan; 
	-- the following is not pretty, a better way?
	-- we need "or" on adjectives in resource grammar
	realLTE a b = let {ltb =complAdj {s = ["less than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	realGTE a b = let {ltb =complAdj {s = ["greater than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	 

--
-- String
--
lin StringC = mkClassNoHum "string" "String";
	stringEq a b = equalTo a b;
	stringLiteral lit = constant lit.s;
	stringSize  = instFunOf "size";
	stringConcat a b = instFunOf "concatenation" (conjoinInst "and" a b);
	stringToUpper = postfixIndFun ["converted to uppercase"];
	stringToLower = postfixIndFun ["converted to lowercase"];
	stringSubstring s a b = np2inst ( -- here we have the problems of indices...
			postfixIndFun (["ranging from index"] ++ a.s!NomP ++ ["to index"] ++ b.s!NomP) 
			(appFunSg (funOfReg "substring") s)
	);
			
	

--
-- Bool
--

lin BoolC = mkClassNoHum "bool" "Boolean";
	true = (constant "true");
	false = (constant "false");
	boolEq = equalTo;
	orBool = instInfix "or"; 
	xorBool = instInfix "xor";
	andBool = instInfix "and";
	notBool = instPrefix "not"; --np2inst (constant ("not" ++ a.s!NomP));
	impliesBool = instInfix "implies";
			
--
-- Sent 
--
lin sentEq a b = s2sent (a.s ++ ["if and only if"] ++ b.s);
-- use conjunctions
	orS a b = s2sent (a.s ++ "or" ++ b.s);
	xorS a b = s2sent ("either" ++ a.s ++ "or" ++ b.s);
	andS a b = s2sent (a.s ++ "and" ++ b.s);
	notS a = s2sent (["it is not the case that"] ++ a.s);
	impliesS a b = s2sent (a.s ++ ["implies that"] ++ b.s);
	-- ifThenElse is a bit tricky: do we really know that t 
	-- and e have the same number and person?
	-- ideally we should have a resource grammar which supports conditional expressions
 	ifThenElse _ cond t e = {
 		s = \\c => t.s!c ++ [", if"] ++ cond.s ++ "," ++ e.s!c ++ "otherwise";
 		n = t.n; p = t.p};
	ifThenElseS cond t e = s2sent ("if" ++ cond.s ++ "then" ++ t.s ++ "else" ++ e.s);



--
-- Iterators
--

-- a InstIter is an instance and a list of bound variables
-- this is not abstract, we would like to simply extend Inst with 
-- a string field for the variable list. Note that Inst = NounPhrase = Pronoun
lintype InstIterL = {s : CaseP => Str ; n : Num ; p : Person; s1 : Str};
lincat InstIter = InstIterL;

lin instIterSingle varC _ body = {
		s1 = body.v;
		s0 = body.s0; n = body.n; p = body.p};
	instIterMany varC _ body = {
		s1 = body.s1 ++ "," ++ body.v;
		s0 = body.s0; n = body.n; p = body.p};

-- a SentIter is a sentence + a list of bound variables
lintype SentIterL = {s0, s1 : Str; moreThanOne : Bool};
lincat SentIter = SentIterL;
lin	sentIterSingle varC  body = {
		s1 = body.v;
		s0 = body.s0;
		moreThanOne = False};
	sentIterMany varC body = {
		s1 = body.s1 ++ "," ++ body.v;
		s0 = body.s0;
		moreThanOne = True};
		
--
-- Collections
--

-- general operations

-- are all forms here really correct?
oper mkCollClass : Str -> Str -> ClassL-> ClassL = \collection, Collection, Integer -> 
	let {integers = Integer.s!Pl!Nom} in 
		(mkNoun (collection ++ "of" ++ integers) (collection+"s" ++ "of" ++ integers)
		(collection ++ "of" ++ integers+"'") (collection+"s" ++ "of" ++ integers+"'")
		) ** {g = NoHum; s42 = Collection ++ class2id Integer};

    collIncluding : Str -> Inst -> Inst -> Inst = \collType, coll, elem -> 
        instFunOf (collType ++ ["containing all elements"]) (instInfix "plus" coll elem);
    collExcluding : Str -> Inst -> Inst -> Inst = \collType, coll, elem -> 
        instFunOf (collType ++ ["containing all elements"]) (instInfix "without" coll elem);
    collSelect : Str -> Inst -> SentIterL -> Inst = \collType, coll, iter -> 
	-- let {satisfy = if_then_else Str	iter.moreThanOne "satisfies" "satisfy"} in 
	postfixIndFun (["which satisfy"] ++ iter.s0) 
         (instFunOf ("the" ++ collType ++ ["all elements"] ++ iter.s1) coll);
    collReject : Str -> Inst -> SentIterL -> Inst = \collType, coll, iter -> 
	postfixIndFun (["which do not satisfy"] ++ iter.s0) 
         (instFunOf ("the" ++ collType ++ ["all elements"] ++ iter.s1) coll);
    collCollect : Str -> Inst -> InstIterL -> Inst = \collType, coll, iter -> 
        -- the collType resulting from evaluating iter.s0 for all iter.s1 in coll
	-- the result of evaluating "iter.s0" for all iter.s1 in coll
	-- NomP correct for coll.s?
         postfixIndFun (["for all"] ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP) 
	 (instFunOf "result" (constant (["evaluating \""] ++ (inst2np iter).s!NomP ++ "\"")));
    collCount : Inst -> Inst -> Inst = \coll, elem ->
        np2inst 
	  (appFunSg (funXReg (["number of times"] ++ (inst2np elem).s!NomP ++ "occurs") "in") 
	  coll);
		
lin Collection c = mkCollClass "collection" "Collection" c; --mkClassNoHum (["collection of"] ++ c.s!Pl!Nom) ("Collection" ++ class2id c);

lin size _ coll = np2inst (appFunSg (funOfReg "size") coll);
	includes _ _ coll elem = predVerbPhrase (inst2np coll) 
		(complTransVerb True (mkTransVerbDir (regVerbP3 "contain")) (inst2np elem));
	excludes _ _ coll elem = predVerbPhrase (inst2np coll) 
		(complTransVerb False (mkTransVerbDir (regVerbP3 "contain")) (inst2np elem)); -- correct?
	count _ = collCount; 
	includesAll _ coll1 coll2 = predVerbPhrase (inst2np coll1) 
		(complTransVerb True (mkTransVerbDir (mkVerbP3 ["contains all elements of"] 
		["contain all elements of"])) (inst2np coll2));
	excludesAll _ coll1 coll2 = predVerbPhrase (inst2np coll1) 
		(complTransVerb True (mkTransVerbDir (mkVerbP3 ["contains none of the elements of"]
		["contain none of the elements of"])) 
		(inst2np coll2));
	isEmpty _ coll = (predVerbPhrase coll (predAdjective True (ss "empty")));
	notEmpty _ coll = (predVerbPhrase coll (predAdjective True (ss ["not empty"])));
	sum _ coll = np2inst (appFunSg (funOfReg ["the sum of all elements"]) coll);
	forAll c coll iter = s2sent (["for all"] ++ (class2cn c).s!Pl!Nom ++ iter.s1 ++
		"in" ++ (inst2np coll).s!NomP ++ "," ++ iter.s0);
	exists c coll iter = s2sent ((case iter.moreThanOne of {
		True => ["there exist"] ++ (class2cn c).s!Pl!Nom ;   
		False => ["there exists"] ++ artIndef ++ (class2cn c).s!Sg!Nom
	}) ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP ++ ["such that"] ++ iter.s0);
	isUnique c d coll iter = s2sent (iter.s!NomP ++ ["is unique for every"] ++ (class2cn c).s!Pl!Nom ++
		iter.s1 ++ "in" ++ (inst2np coll).s!NomP); -- need quotes here?
	sortedBy c d coll iter = constant (["the sequence of all"] ++ (class2cn c).s!Pl!Nom ++ iter.s1 ++
		"in" ++ (inst2np coll).s!NomP ++ ["sorted by"] ++ iter.s0!NomP);
	iterate c a coll init body = constant (["the result of iterating over every"]  ++ (class2cn c).s!Sg!Nom ++ 
		body.v0 ++ "in" ++ (inst2np coll).s!NomP ++ ["with accumulator"] ++ body.v1 ++ ["initialized to"] ++
		(inst2np init).s!NomP ++ [", in each step evaluating"] ++ body.s!NomP ); -- (inst2np body).s!NomP does not work???
	any c coll iter = constant ("any" ++ (class2cn c).s!Sg!Nom ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP ++
		["such that"] ++ iter.s0);
	one c coll iter = s2sent ((case iter.moreThanOne of {
		True => ["there exist exactly one"] ++ (class2cn c).s!Pl!Nom ;   -- should be variants{} ?
		False => ["there exists exactly one"] ++ artIndef ++ (class2cn c).s!Sg!Nom
	}) ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP ++ ["such that"] ++ iter.s0);

--
-- sets 
--

lin Set c = mkCollClass "set" "Set" c;
    unionSS _ a b = instFunOf "union" (conjoinInst "and" a b);
    unionSB _ a b = instFunOf "union" (conjoinInst "and" a b);
    setEq c = equalTo;
    intersectionSS c a b = instFunOf "intersection" (conjoinInst "and" a b);
    intersectionSB c a b = instFunOf "intersection" (conjoinInst "and" a b);
    subS c a b = instFunOf ["set difference"] (conjoinInst "minus" a b); 
    includingS c = collIncluding "set"; 
    excludingS c = collExcluding "set";
    symmetricDiff c a b = instFunOf ["symmetric difference"] (conjoinInst "and" a b);
    selectSet c = collSelect "set";   
    rejectSet c = collReject "set";
    collectSet c d = collCollect "bag";
    countSet _ = collCount;
    setAsSequence _ = postfixIndFun ["as a sequence"]; -- alternative is a circular rule "= c"
    setAsBag _ = postfixIndFun ["as a bag"];

--
-- bags
--
lin Bag c = mkCollClass "bag" "Bag" c;
    bagEq _ = equalTo;
    unionBB _ a b = instFunOf "union" (conjoinInst "and" a b);
    unionBS _ a b = instFunOf "union" (conjoinInst "and" a b);
    intersectionBB _ a b = instFunOf "intersection" (conjoinInst "and" a b);
    intersectionBS _ a b = instFunOf "intersection" (conjoinInst "and" a b);
    includingB c = collIncluding "bag";
    excludingB c = collExcluding "bag";
    selectBag c = collSelect "bag";
    rejectBag c = collReject "bag";
    collectBag _ _ = collCollect "bag";
    countBag _ = collCount;
    bagAsSequence _ = postfixIndFun ["as a sequence"];
    bagAsSet _ = postfixIndFun ["as a set"];

--
-- sequences
--
lin Sequence c = mkCollClass "sequence" "Sequence" c;
    sequenceEq _ = equalTo;
    unionSeq _ a b = instFunOf "union" (conjoinInst "and" a b);
    append _ a b = instInfix ["appended to the end of"] b a;
    prepend _ a b = instInfix ["appended to the start of"] b a; 
    subSequence _ a start end = np2inst ( -- here we have the problems of indices...
			postfixIndFun (["ranging from index"] ++ start.s!NomP ++ ["to index"] ++ end.s!NomP) 
			(appFunSg (funOfReg "subsequence") a)
	); 
    at _ a index = instFunOf (["element at index"] ++ (inst2np index).s!NomP) a; -- NomP?
    first _  = instFunOf ["first element"] ;
    last _ = instFunOf ["last element"];
    includingSeq _ = collIncluding "sequence";
    excludingSeq _ = collExcluding "sequence";
    selectSeq _ = collReject "sequence";
    rejectSeq _ = collReject "sequence";
    collectSeq _ _ = collCollect "sequence";
    seqAsBag _ = postfixIndFun ["as a bag"];
    seqAsSet _ = postfixIndFun ["as a set"];
