// $ANTLR 2.7.2: "decls.g" -> "DeclParser.java"$


package de.uka.ilkd.key.parser;

import java.io.*;
import antlr.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.visitor.*;
import java.util.List;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.HashSet;




import antlr.TokenBuffer;
import antlr.TokenStreamException;
import antlr.TokenStreamIOException;
import antlr.ANTLRException;
import antlr.LLkParser;
import antlr.Token;
import antlr.TokenStream;
import antlr.RecognitionException;
import antlr.NoViableAltException;
import antlr.MismatchedTokenException;
import antlr.SemanticException;
import antlr.ParserSharedInputState;
import antlr.collections.impl.BitSet;

/** 
 * A parser for signature declarations,
 * i.e. sorts, schema variabls, predicates, functions and heuristics.
 */
public class DeclParser extends antlr.LLkParser       implements DeclParserTokenTypes
 {

    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    
    static HashMap name2sort = new HashMap(50);
    
    static {
          name2sort.put("statement",ProgramSVSort.STATEMENT);
	  name2sort.put("expression",ProgramSVSort.EXPRESSION);
	  name2sort.put("javaboolean expression",ProgramSVSort.JAVABOOLEANEXPRESSION);
	  name2sort.put("executioncontext",ProgramSVSort.EXECUTIONCONTEXT);
	  name2sort.put("multiplevardecl",ProgramSVSort.MULTIPLEVARDECL);
	  name2sort.put("simple expression",ProgramSVSort.SIMPLEEXPRESSION);
	  name2sort.put("simple javabyte expression",ProgramSVSort.SIMPLEJAVABYTEEXPRESSION);
	  name2sort.put("simple javachar expression",ProgramSVSort.SIMPLEJAVACHAREXPRESSION);
	  name2sort.put("simple javashort expression",ProgramSVSort.SIMPLEJAVASHORTEXPRESSION);
	  name2sort.put("simple javaint expression",ProgramSVSort.SIMPLEJAVAINTEXPRESSION);
	  name2sort.put("simple javalong expression",ProgramSVSort.SIMPLEJAVALONGEXPRESSION);
	  name2sort.put("simple javabyteshortint expression",ProgramSVSort.SIMPLEJAVABYTESHORTINTEXPRESSION);
	  name2sort.put("simple anyjavatype expression",ProgramSVSort.SIMPLEANYJAVATYPEEXPRESSION);
	  name2sort.put("nonsimple expression",ProgramSVSort.NONSIMPLEEXPRESSION);
	  name2sort.put("nonsimple methodreference",ProgramSVSort.NONSIMPLEMETHODREFERENCE);
	  name2sort.put("loopinit",ProgramSVSort.LOOPINIT);
	  name2sort.put("guard",ProgramSVSort.GUARD);
	  name2sort.put("forupdates",ProgramSVSort.FORUPDATES);
	  name2sort.put("label",ProgramSVSort.LABEL);
	  name2sort.put("type",ProgramSVSort.TYPE);
	  name2sort.put("notprimitive type",ProgramSVSort.TYPENOTPRIMITIVE);
	  name2sort.put("explicit variable",ProgramSVSort.EXPLICITVARIABLE);
          name2sort.put("static variable",ProgramSVSort.STATICVARIABLE);
	  name2sort.put("variable",ProgramSVSort.VARIABLE);
	  name2sort.put("constant variable",ProgramSVSort.CONSTANTVARIABLE);
	  name2sort.put("variableinitializer",ProgramSVSort.VARIABLEINIT);
	  name2sort.put("implicit variable",ProgramSVSort.IMPLICITVARIABLE);
	  name2sort.put("implicit reference field",ProgramSVSort.IMPLICITREFERENCE);
	  name2sort.put("implicit next", ProgramSVSort.IMPLICITNEXT);
          name2sort.put("implicit nextToCreate", ProgramSVSort.IMPLICITNEXTTOCREATE);
          name2sort.put("implicit classInitialized", ProgramSVSort.IMPLICITCLINIT);
          name2sort.put("implicit classInitializationInProgress",ProgramSVSort.IMPLICITINITINPROGRESS);
          name2sort.put("implicit classErroneous", ProgramSVSort.IMPLICITERRONEOUS);
          name2sort.put("implicit classPrepared", ProgramSVSort.IMPLICITPREPARED);
	  name2sort.put("begin transaction",ProgramSVSort.JCBEGINTRANSACTION);
	  name2sort.put("commit transaction",ProgramSVSort.JCCOMMITTRANSACTION);
	  name2sort.put("abort transaction",ProgramSVSort.JCABORTTRANSACTION);
	  name2sort.put("transient booleanarray",ProgramSVSort.JCMAKETRANSIENTBOOLEANARRAY);
	  name2sort.put("transient bytearray",ProgramSVSort.JCMAKETRANSIENTBYTEARRAY);
	  name2sort.put("transient shortarray",ProgramSVSort.JCMAKETRANSIENTSHORTARRAY);
	  name2sort.put("transient objectarray",ProgramSVSort.JCMAKETRANSIENTOBJECTARRAY);
	  name2sort.put("transient istransient",ProgramSVSort.JCISTRANSIENT);
	  name2sort.put("array initializer",ProgramSVSort.ARRAYINITIALIZER);
	  name2sort.put("array length",ProgramSVSort.ARRAYLENGTH);
	  name2sort.put("array creation expression",ProgramSVSort.NEWARRAY);
	  name2sort.put("array post declaration",ProgramSVSort.ARRAYPOSTDECL);
	  name2sort.put("instance creation expression",ProgramSVSort.NEW);
	  name2sort.put("switch",ProgramSVSort.SWITCH);
	  name2sort.put("catch",ProgramSVSort.CATCH);
	  name2sort.put("methodname",ProgramSVSort.METHODNAME);
	  name2sort.put("special constructorreference",ProgramSVSort.SPECIALCONSTRUCTORREFERENCE);
	  name2sort.put("methodbody",ProgramSVSort.METHODBODY);
	  name2sort.put("lefthandside",ProgramSVSort.LEFTHANDSIDE);
    }
    
    private Prioritization.StaticPrioritization heur_prioritization
        = new Prioritization.StaticPrioritization ();
    
    private NamespaceSet nss;
    private Choice defaultChoice = null;
    protected HashMap category2Default = new HashMap();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();

    private Services services;

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_schemavariables;
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_heuristics;

    private boolean schemaMode = false;
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();

    public DeclParser(TokenStream lexer,
                      String filename,
                      Services services,
                      NamespaceSet nss,
                      Choice defaultChoice) {
        this(lexer);
        setFilename(filename);
        this.nss = nss;
        this.services = services;
        setSkips(false, false, false, false, false);
        switchToNormalMode();
        this.defaultChoice = defaultChoice;
    }

    public DeclParser(TokenStream lexer,
                      String filename,
                      Services services,
                      NamespaceSet nss) {
        this(lexer, filename, services, nss, null);
    }
    
    public JavaInfo getJavaInfo() {
        return getServices().getJavaInfo();
    }

    public Services getServices() {
        return services;
    }

    public Namespace sorts() {
        return nss.sorts();
    }

    public Namespace functions() {
        return nss.functions();
    }

    public Namespace heuristics() {
        return nss.heuristics();
    }

    public Namespace variables() {
        return nss.variables();
    }

    public Namespace programVariables() {
        return nss.programVariables();
    }

    public Namespace choices(){
        return nss.choices();
    }

    public NamespaceSet namespaces() {
        return nss;
    }
    
    public HashMap getCategory2Default(){
        return category2Default;
    }

    private void switchToSchemaMode() {
        schemaMode = true;      
    }

    private boolean inSchemaMode() {
        return schemaMode;
    }

    private void switchToNormalMode() {
        schemaMode = false;  
    }

    public Prioritization getHeuristicPrioritization () {
        return heur_prioritization;
    }

    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    protected int getColumn() {
        int col = -1;
        try {
            col = LT(0).getColumn();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }   

    protected void setSkips(boolean svars, boolean funcs, 
                            boolean preds, boolean sorts,
                            boolean heurs) {
        skip_schemavariables = svars;
        skip_functions       = funcs;
        skip_predicates      = preds;
        skip_sorts       = sorts;
        skip_heuristics      = heurs;
    }

    public Named lookup(Name n) {
	final Namespace[] lookups = {
		functions(), variables(), 
		programVariables()
	};
	for (int i = 0; i<lookups.length; i++) {
	    if (lookups[i].lookup(n) != null) {
		return lookups[i].lookup(n);
            }
	}
	return null;	
    }

    public void parseVariables()    
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(false, true, true, true, true); 
            top();
            setSkips(false, false, false, false, false);                
        }

    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if (relativePath) {
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
        } else {
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if (ldt) {
            includes.putLDT(filename, source);
        } else {
            includes.put(filename, source);
        }
    }  

    public void parseFunctions()
        throws antlr.RecognitionException, antlr.TokenStreamException

        {
            setSkips(true, false, true, true, true); 
            top();
            setSkips(false, false, false, false, false);
        }

    public void parsePredicates()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, true, false, true, true); 
            top();
            setSkips(false, false, false, false, false);        
        }

    public void parseSorts() 
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, true, true, false, true); 
            top();
            setSkips(false, false, false, false, false);
        }

    public void parseHeuristics()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, true, true, true, false); 
            top();
            setSkips(false, false, false, false, false);        
        }

    public void parseFuncAndPred()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, false, false, true, true); 
            top();
            setSkips(false, false, false, false, false);
        }
    

    public void schema_var_decl(String name,Sort s,
                                    boolean makeVariableSV,
                                    boolean makeDependingSV,
                                    boolean rigid,
                                    boolean listModifier)
	{
              if (!skip_schemavariables) {
                  SchemaVariable v;
                  if ( s == Sort.FORMULA ) {
                      v = SchemaVariableFactory.createFormulaSV
                          (new Name(name), listModifier, rigid);
                      } else if ( s instanceof ProgramSVSort ) {
                          v = SchemaVariableFactory.createProgramSV
                         (new ProgramElementName(name),(ProgramSVSort) s, listModifier);
                      } else {
                        if ( makeVariableSV ) {
                             v=SchemaVariableFactory.createVariableSV
                             (new Name(name), s, listModifier);
                        } else if ( makeDependingSV ) {
                               v = SchemaVariableFactory.createDependingSV
                             (new Name(name), s, listModifier);
                        } else { v = SchemaVariableFactory.createTermSV
                              (new Name(name), s, listModifier, rigid);
                        }
                     }
                     if (inSchemaMode()) {
                         variables().add(v);
                     }
               }

         }
     

protected DeclParser(TokenBuffer tokenBuf, int k) {
  super(tokenBuf,k);
  tokenNames = _tokenNames;
}

public DeclParser(TokenBuffer tokenBuf) {
  this(tokenBuf,1);
}

protected DeclParser(TokenStream lexer, int k) {
  super(lexer,k);
  tokenNames = _tokenNames;
}

public DeclParser(TokenStream lexer) {
  this(lexer,1);
}

public DeclParser(ParserSharedInputState state) {
  super(state,1);
  tokenNames = _tokenNames;
}

	public final void top() throws RecognitionException, TokenStreamException {
		
		
		decls();
	}
	
	public final void decls() throws RecognitionException, TokenStreamException {
		
		
		{
		_loop7:
		do {
			if ((LA(1)==INCLUDE)) {
				match(INCLUDE);
				{
				switch ( LA(1)) {
				case LDTS:
				{
					match(LDTS);
					match(COLON);
					match(LBRACE);
					include(true);
					{
					_loop6:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							include(true);
						}
						else {
							break _loop6;
						}
						
					} while (true);
					}
					match(RBRACE);
					break;
				}
				case STRING_LITERAL:
				case IDENT:
				{
					include(false);
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				match(SEMI);
			}
			else {
				break _loop7;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			
			if(parse_includes) return;
			activatedChoices = SetAsListOfChoice.EMPTY_SET;  
			
		}
		{
		switch ( LA(1)) {
		case WITH:
		{
			match(WITH);
			match(LPAREN);
			activated_choice();
			{
			_loop10:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					activated_choice();
				}
				else {
					break _loop10;
				}
				
			} while (true);
			}
			match(RPAREN);
			break;
		}
		case EOF:
		case PROGRAM:
		case HEURISTICS:
		case SORTS:
		case OPTIONS:
		case SCHEMA:
		case PREDICATES:
		case FUNCTIONS:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			if(onlyWith) return;
		}
		{
		_loop12:
		do {
			switch ( LA(1)) {
			case OPTIONS:
			{
				option_decls();
				break;
			}
			case SORTS:
			{
				sort_decls();
				break;
			}
			case PROGRAM:
			{
				prog_var_decls();
				break;
			}
			case SCHEMA:
			{
				schema_var_decls();
				break;
			}
			case PREDICATES:
			{
				pred_decls();
				break;
			}
			case FUNCTIONS:
			{
				func_decls();
				break;
			}
			case HEURISTICS:
			{
				heur_decls();
				break;
			}
			default:
			{
				break _loop12;
			}
			}
		} while (true);
		}
	}
	
	public final void include(
		boolean ldt
	) throws RecognitionException, TokenStreamException {
		
		Token  absfile = null;
		Token  relfile = null;
		
		{
		switch ( LA(1)) {
		case IDENT:
		{
			absfile = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				
				if(parse_includes){
				addInclude(absfile.getText(),false,ldt);
				}
				
			}
			break;
		}
		case STRING_LITERAL:
		{
			relfile = LT(1);
			match(STRING_LITERAL);
			if ( inputState.guessing==0 ) {
				
				if(parse_includes){
				addInclude(relfile.getText(),true,ldt);
				}
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
	}
	
	public final void activated_choice() throws RecognitionException, TokenStreamException {
		
		Token  cat = null;
		Token  choice = null;
		
		String name;
		Choice c;
		
		
		cat = LT(1);
		match(IDENT);
		match(COLON);
		choice = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			if(usedChoiceCategories.contains(cat.getText())){
			throw new IllegalArgumentException("You have already chosen a different option for "+cat.getText());
			}
			usedChoiceCategories.add(cat.getText());
			name = cat.getText()+":"+choice.getText();
			c = (Choice) choices().lookup(new Name(name));
			if(c==null){
			throw new NotDeclException("Option", choice,
			getFilename());
			}else{
			activatedChoices=activatedChoices.add(c);
			namespaces().setFunctions(functions().extended(c.funcNS().allElements()));
			}
			
		}
	}
	
	public final void option_decls() throws RecognitionException, TokenStreamException {
		
		
		
		
		match(OPTIONS);
		match(LBRACE);
		choice();
		{
		_loop18:
		do {
			if ((LA(1)==SEMI)) {
				match(SEMI);
				choice();
			}
			else {
				break _loop18;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void sort_decls() throws RecognitionException, TokenStreamException {
		
		
		boolean isObjectSort=false;
		boolean isGenericSort=false;
		
		//For OCL simplification
		boolean isOclSort = false;
		
		Sort[] sortExt=new Sort [0];
		Sort[] sortOneOf=new Sort [0];
		ListOfName sortIds;
		
		
		match(SORTS);
		match(LBRACE);
		{
		_loop29:
		do {
			if ((_tokenSet_0.member(LA(1)))) {
				if ( inputState.guessing==0 ) {
					
					isObjectSort=false; 
					isGenericSort=false; 
					isOclSort = false;
					sortExt = new Sort[0];
					sortOneOf = new Sort[0];
					
				}
				{
				switch ( LA(1)) {
				case OBJECT:
				{
					match(OBJECT);
					if ( inputState.guessing==0 ) {
						isObjectSort=true;
					}
					sortIds=sortIdentifiers();
					break;
				}
				case GENERIC:
				{
					match(GENERIC);
					if ( inputState.guessing==0 ) {
						isGenericSort=true;
					}
					sortIds=sortIdentifiers();
					{
					switch ( LA(1)) {
					case ONEOF:
					{
						match(ONEOF);
						sortOneOf=oneof_sorts();
						break;
					}
					case EXTENDS:
					case SEMI:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					{
					switch ( LA(1)) {
					case EXTENDS:
					{
						match(EXTENDS);
						sortExt=extends_sorts();
						break;
					}
					case SEMI:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					break;
				}
				case OCL:
				{
					match(OCL);
					if ( inputState.guessing==0 ) {
						isOclSort=true;
					}
					sortIds=sortIdentifiers();
					break;
				}
				case IDENT:
				{
					sortIds=sortIdentifiers();
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				match(SEMI);
				if ( inputState.guessing==0 ) {
					
					if (!skip_sorts) {
					Name sort_name;
					IteratorOfName it = sortIds.iterator ();
					while ( it.hasNext () ) {
					sort_name = it.next ();
					
					if (sorts().lookup(sort_name) == null) {
					Sort s;
					if (isObjectSort) {
					if ("Object".equals(sort_name)) {
					if (sorts().lookup(new Name("Object")) == null) {
					s = new ClassInstanceSortImpl(sort_name);
					}
					s=(Sort)sorts().lookup(new Name("Object"));
					} else {
					s = new ClassInstanceSortImpl(sort_name, 
					(Sort)sorts().lookup
					(new Name("Object")));
					}	
					}                            
					//Just constant for now
					else if (isOclSort) {
					s = Sort.OCL_ANY;
					}
					//
					else if (isGenericSort) {
					int        i;
					SetOfSort  ext   = SetAsListOfSort.EMPTY_SET;
					SetOfSort  oneOf = SetAsListOfSort.EMPTY_SET;
					
					for ( i = 0; i != sortExt.length; ++i )
					ext = ext.add ( sortExt[i] );
					
					for ( i = 0; i != sortOneOf.length; ++i )
					oneOf = oneOf.add ( sortOneOf[i] );
					
					try {
					s = new GenericSort ( sort_name,
					ext,
					oneOf );
					} catch ( GenericSupersortException e ) {
					throw new GenericSortException ( "sort",
					"Illegal sort given", e.getIllegalSort (),
					getFilename (), getLine (), getColumn () );
					}
					} else {
					s = new PrimitiveSort(sort_name);
					}
					sorts().add(s);
					if (s instanceof NonCollectionSort) {
					NonCollectionSort ns
					= (NonCollectionSort)s;
					sorts().add(ns.getSetSort());
					sorts().add(ns.getSequenceSort());
					sorts().add(ns.getBagSort());
					}
					if ( s instanceof SortDefiningSymbols )
					((SortDefiningSymbols)s).createSymbols ( defaultChoice.funcNS(),
					sorts() );
					}
					
					}
					}
					
				}
			}
			else {
				break _loop29;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void prog_var_decls() throws RecognitionException, TokenStreamException {
		
		Token  v0 = null;
		Token  v1 = null;
		Token  pid = null;
		
		String var_name;
		String type = null;
		
		
		if ( inputState.guessing==0 ) {
			switchToNormalMode();
		}
		match(PROGRAM);
		match(VARIABLES);
		match(LBRACE);
		{
		_loop47:
		do {
			if ((LA(1)==PIDENT||LA(1)==IDENT)) {
				{
				switch ( LA(1)) {
				case IDENT:
				{
					{
					v0 = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						type= "" + v0.getText();
					}
					{
					_loop44:
					do {
						if ((LA(1)==DOT)) {
							match(DOT);
							v1 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								type= type + "." + v1.getText();
							}
						}
						else {
							break _loop44;
						}
						
					} while (true);
					}
					}
					break;
				}
				case PIDENT:
				{
					{
					pid = LT(1);
					match(PIDENT);
					if ( inputState.guessing==0 ) {
						type = pid.getText();
					}
					}
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				{
				switch ( LA(1)) {
				case LBRACKET:
				{
					match(LBRACKET);
					match(RBRACKET);
					if ( inputState.guessing==0 ) {
						
						type = type + "[]";
						
					}
					break;
				}
				case PIDENT:
				case IDENT:
				{
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				var_name=id();
				if ( inputState.guessing==0 ) {
					
					KeYJavaType kjt = getJavaInfo().getKeYJavaType(type);
							ProgramElementName pvName = new ProgramElementName(var_name);
					if (kjt == null) {
					throw new antlr.SemanticException
					("Program Variable " + var_name + 
					" declared with unknown type " + 
					type);
					
					} else if (lookup(pvName)!=null) {
							  // commented out as pv do not have unique name (at the moment)
							    //throw new AmbigiousDeclException
							//	(var_name, getFilename(), getLine(), getColumn()); 
							}
					namespaces().programVariables().add(new ProgramVariable
					(new ProgramElementName(var_name), kjt));
					
					
				}
				match(SEMI);
			}
			else {
				break _loop47;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void schema_var_decls() throws RecognitionException, TokenStreamException {
		
		Token  id1 = null;
		Token  id2 = null;
		Token  id3 = null;
		Token  id5 = null;
		Token  id6 = null;
		Token  id7 = null;
		Token  id8 = null;
		
		Sort s = null;
		boolean makeVariableSV  = false;
		boolean makeDependingSV = false; // skolem functions
		boolean rigid           = false;
		boolean listModifier    = false;
		String type = "";    
		
		
		match(SCHEMA);
		match(VARIABLES);
		match(LBRACE);
		if ( inputState.guessing==0 ) {
			switchToSchemaMode();
		}
		{
		_loop72:
		do {
			switch ( LA(1)) {
			case MODAL:
			{
				{
				match(MODAL);
				match(OPERATOR);
				one_schema_modal_decl();
				{
				_loop52:
				do {
					if ((LA(1)==COMMA)) {
						match(COMMA);
						one_schema_modal_decl();
					}
					else {
						break _loop52;
					}
					
				} while (true);
				}
				match(SEMI);
				}
				break;
			}
			case PROGRAM:
			case FORMULA:
			case PIDENT:
			case IDENT:
			{
				if ( inputState.guessing==0 ) {
					makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;
				}
				{
				switch ( LA(1)) {
				case PROGRAM:
				{
					{
					match(PROGRAM);
					{
					switch ( LA(1)) {
					case SVLIST:
					{
						match(SVLIST);
						if ( inputState.guessing==0 ) {
							listModifier=true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						type = "";
					}
					{
					int _cnt59=0;
					_loop59:
					do {
						boolean synPredMatched58 = false;
						if (((LA(1)==IDENT))) {
							int _m58 = mark();
							synPredMatched58 = true;
							inputState.guessing++;
							try {
								{
								match(IDENT);
								match(IDENT);
								}
							}
							catch (RecognitionException pe) {
								synPredMatched58 = false;
							}
							rewind(_m58);
							inputState.guessing--;
						}
						if ( synPredMatched58 ) {
							id1 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								type += " "+id1.getText();
							}
						}
						else if ((LA(1)==IDENT)) {
							id2 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								type = type.substring(1,type.length());
								s = (Sort) DeclParser.name2sort.get(type);
								if (s == null) {
								throw new antlr.SemanticException
								("SchemaVariable of sort "+type+ " not found.");
								}
								schema_var_decl(id2.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
						}
						else {
							if ( _cnt59>=1 ) { break _loop59; } else {throw new NoViableAltException(LT(1), getFilename());}
						}
						
						_cnt59++;
					} while (true);
					}
					{
					_loop61:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							id3 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								schema_var_decl(id3.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
						}
						else {
							break _loop61;
						}
						
					} while (true);
					}
					match(SEMI);
					}
					break;
				}
				case FORMULA:
				{
					match(FORMULA);
					{
					switch ( LA(1)) {
					case RIGID:
					{
						match(RIGID);
						if ( inputState.guessing==0 ) {
							rigid = true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						s = Sort.FORMULA;type = "";
					}
					{
					id5 = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						schema_var_decl(id5.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
					}
					}
					{
					_loop66:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							{
							id6 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								schema_var_decl(id6.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
							}
						}
						else {
							break _loop66;
						}
						
					} while (true);
					}
					match(SEMI);
					break;
				}
				case PIDENT:
				case IDENT:
				{
					s=any_sortId_check(true);
					{
					switch ( LA(1)) {
					case VARIABLES:
					{
						match(VARIABLES);
						if ( inputState.guessing==0 ) {
							makeVariableSV = true;
						}
						break;
					}
					case DEPENDING:
					{
						match(DEPENDING);
						if ( inputState.guessing==0 ) {
							makeDependingSV = true;
						}
						break;
					}
					case RIGID:
					{
						match(RIGID);
						if ( inputState.guessing==0 ) {
							rigid = true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						type = "";
					}
					{
					id7 = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						schema_var_decl(id7.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
					}
					}
					{
					_loop71:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							{
							id8 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								schema_var_decl(id8.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
							}
						}
						else {
							break _loop71;
						}
						
					} while (true);
					}
					match(SEMI);
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				break;
			}
			default:
			{
				break _loop72;
			}
			}
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			switchToNormalMode();
		}
		match(RBRACE);
	}
	
	public final void pred_decls() throws RecognitionException, TokenStreamException {
		
		Token  cat = null;
		Token  choice = null;
		
		Sort[] argSorts;
		String pred_name;
		Namespace oldFNS=null;
		String ctext = null;
		Choice c=null;
		
		
		match(PREDICATES);
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			cat = LT(1);
			match(IDENT);
			match(COLON);
			choice = LT(1);
			match(IDENT);
			match(RPAREN);
			break;
		}
		case LBRACE:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			if(cat!=null){
			ctext = cat.getText()+":"+choice.getText();
			c = (Choice) choices().lookup(new Name(ctext));
			if(c==null){
			throw new NotDeclException("Option", choice,
			getFilename());
			}else{
			namespaces().setFunctions(c.funcNS());
			}
			}else{
			namespaces().setFunctions(defaultChoice.funcNS());
			}
			cat=null;
			
		}
		match(LBRACE);
		{
		_loop81:
		do {
			if ((LA(1)==PIDENT||LA(1)==IDENT)) {
				pred_name=id();
				argSorts=arg_sorts(!skip_predicates);
				if ( inputState.guessing==0 ) {
					
					if (!skip_predicates) {
					Name predicate = new Name(pred_name);		    		    
					if (lookup(predicate) != null) {
					throw new AmbigiousDeclException
					(pred_name, getFilename(), getLine(), getColumn());
					} else {
					Function p = new Function(predicate,
					Sort.FORMULA, 
					argSorts);
					functions().add(p);         
					}
					} 
					
				}
				match(SEMI);
			}
			else {
				break _loop81;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void func_decls() throws RecognitionException, TokenStreamException {
		
		Token  cat = null;
		Token  choice = null;
		
		Sort[] argSorts;
		Sort retSort;
		String func_name;
		Namespace oldFNS=null;
		String ctext = null;
		Choice c=null;
		
		
		match(FUNCTIONS);
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			cat = LT(1);
			match(IDENT);
			match(COLON);
			choice = LT(1);
			match(IDENT);
			match(RPAREN);
			break;
		}
		case LBRACE:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			if(cat!=null){
			ctext = cat.getText()+":"+choice.getText();
			c = (Choice) choices().lookup(new Name(ctext));
			if(c==null){
			throw new NotDeclException("Option", choice,
			getFilename());
			}
			}else{
			c=defaultChoice;
			}
			namespaces().setFunctions(c.funcNS());
			cat=null;
			
		}
		match(LBRACE);
		{
		_loop85:
		do {
			if ((LA(1)==PIDENT||LA(1)==IDENT)) {
				retSort=sortId_check(!skip_functions);
				func_name=id();
				argSorts=arg_sorts(!skip_functions);
				if ( inputState.guessing==0 ) {
					
					if (!skip_functions) {
					Name fct_name = new Name(func_name);
					if (lookup(fct_name) != null) {
					throw new AmbigiousDeclException
					(func_name, getFilename(), getLine(), getColumn());
					} else {
					functions().add(new Function(fct_name,
						                                             retSort, 
						                                     argSorts));
					}
					} 
					
				}
				match(SEMI);
			}
			else {
				break _loop85;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void heur_decls() throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		int     t_priority = 0;
		Integer priority   = null;
		
		
		match(HEURISTICS);
		match(LBRACE);
		{
		_loop94:
		do {
			if ((LA(1)==IDENT)) {
				if ( inputState.guessing==0 ) {
					priority = null;
				}
				id = LT(1);
				match(IDENT);
				{
				switch ( LA(1)) {
				case LPAREN:
				{
					match(LPAREN);
					t_priority=integer_num();
					match(RPAREN);
					if ( inputState.guessing==0 ) {
						priority = new Integer ( t_priority );
					}
					break;
				}
				case SEMI:
				{
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				match(SEMI);
				if ( inputState.guessing==0 ) {
					
					if (!skip_heuristics) {
					Heuristic h = new Heuristic(new Name(id.getText()));
					if(heuristics().lookup(new Name(id.getText()))==null){
					heuristics().add(h);
					}
					if ( priority != null )
					heur_prioritization.setPriority ( h, priority.intValue () );
					}
					
				}
			}
			else {
				break _loop94;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void choice() throws RecognitionException, TokenStreamException {
		
		Token  category = null;
		
		String cat=null;
		
		
		category = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			cat=category.getText();
		}
		{
		switch ( LA(1)) {
		case COLON:
		{
			match(COLON);
			match(LBRACE);
			choice_option(cat);
			{
			_loop22:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					choice_option(cat);
				}
				else {
					break _loop22;
				}
				
			} while (true);
			}
			match(RBRACE);
			break;
		}
		case SEMI:
		case RBRACE:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			if(!category2Default.containsKey(cat)){
			choices().add(new Choice("On",cat));
			choices().add(new Choice("Off",cat)); 
			category2Default.put(cat, cat+":On");               
			}
			
		}
	}
	
	public final void choice_option(
		String cat
	) throws RecognitionException, TokenStreamException {
		
		Token  choice = null;
		
		String name;
		
		
		choice = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			name=cat+":"+choice.getText();
			Choice c = (Choice) choices().lookup(new Name(name));
			if(c==null){
			c = new Choice(choice.getText(),cat);
			choices().add(c);
			}
			if(!category2Default.containsKey(cat)){
			category2Default.put(cat, name);
			}
			
		}
	}
	
	public final ListOfName  sortIdentifiers() throws RecognitionException, TokenStreamException {
		ListOfName ids = SLListOfName.EMPTY_LIST;
		
		Token  id = null;
		Token  id2 = null;
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			ids = ids.append ( new Name(id.getText()) );
		}
		{
		_loop32:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				id2 = LT(1);
				match(IDENT);
				if ( inputState.guessing==0 ) {
					ids = ids.append ( new Name(id2.getText()) );
				}
			}
			else {
				break _loop32;
			}
			
		} while (true);
		}
		return ids;
	}
	
	public final Sort[]  oneof_sorts() throws RecognitionException, TokenStreamException {
		Sort[] oneOfSorts = null;
		
		
		List args = new LinkedList();
		Sort s;
		
		
		match(LBRACE);
		s=sortId_check(true);
		if ( inputState.guessing==0 ) {
			args.add(s);
		}
		{
		_loop38:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s=sortId_check(true);
				if ( inputState.guessing==0 ) {
					args.add(s);
				}
			}
			else {
				break _loop38;
			}
			
		} while (true);
		}
		match(RBRACE);
		if ( inputState.guessing==0 ) {
			
			oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
			
		}
		return oneOfSorts;
	}
	
	public final Sort[]  extends_sorts() throws RecognitionException, TokenStreamException {
		Sort[] extendsSorts = null;
		
		
		List args = new LinkedList();
		Sort s;
		
		
		s=any_sortId_check(true);
		if ( inputState.guessing==0 ) {
			args.add(s);
		}
		{
		_loop35:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s=any_sortId_check(true);
				if ( inputState.guessing==0 ) {
					args.add(s);
				}
			}
			else {
				break _loop35;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			
			extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
			
		}
		return extendsSorts;
	}
	
	public final Sort  any_sortId_check(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		Sort t;
		
		
		t=any_sortId_check_help(checkSort);
		s=array_decls(t);
		return s;
	}
	
	public final Sort  sortId_check(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		Sort t;
		
		
		t=sortId_check_help(checkSort);
		s=array_decls(t);
		return s;
	}
	
	public final String  id() throws RecognitionException, TokenStreamException {
		String name = null;
		
		Token  prefix = null;
		Token  pprefix = null;
		Token  nameID = null;
		
		{
		switch ( LA(1)) {
		case IDENT:
		{
			prefix = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				name = prefix.getText();
			}
			break;
		}
		case PIDENT:
		{
			pprefix = LT(1);
			match(PIDENT);
			if ( inputState.guessing==0 ) {
				name = pprefix.getText();
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		{
		switch ( LA(1)) {
		case COLON:
		{
			match(COLON);
			match(COLON);
			nameID = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				
				name += "::" + nameID.getText();
				
			}
			break;
		}
		case EOF:
		case DEPENDING:
		case VARIABLES:
		case RIGID:
		case SEMI:
		case COMMA:
		case LPAREN:
		case RPAREN:
		case RBRACE:
		case LBRACKET:
		case PIDENT:
		case IDENT:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return name;
	}
	
	public final void one_schema_modal_decl() throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		Modality[] mods;
		
		
		id = LT(1);
		match(IDENT);
		match(EQUALS);
		mods=modalities_decl();
		if ( inputState.guessing==0 ) {
			
			SchemaVariable msv = (SchemaVariable)variables().lookup(new Name(id.getText()));
			if(msv != null)
			throw new antlr.SemanticException("Schema variable "+id.getText()+ " already defined");
			int arity = mods[0].arity();
			for(int i=1;i<mods.length;i++)
			if(arity != mods[i].arity())
			throw new antlr.SemanticException("Arity mismatch for schema variable "+id.getText());
			msv = SchemaVariableFactory.createModalOperatorSV(new Name(id.getText()), arity, mods);
			if (inSchemaMode()) {
			variables().add(msv);
			}
			
		}
	}
	
	public final Modality[]  modalities_decl() throws RecognitionException, TokenStreamException {
		Modality[] modalities = null;
		
		Token  s1 = null;
		Token  s2 = null;
		
		List args = new LinkedList();
		
		
		{
		match(LBRACE);
		s1 = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			
			if(s1.getText().charAt(0) == '#') {
			ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s1.getText()));
			if(msv == null)
			throw new antlr.SemanticException("Schema variable "+s1.getText()+ " not defined");
			for(int i=0; i<msv.modalities().length; i++)
			args.add(msv.modalities()[i]);
			}else{
			Modality op = Op.getModality(s1.getText());
			if(op == null)
			throw new antlr.SemanticException("Unrecognised modal operator "+s1.getText());
			args.add(op);
			}
			
		}
		{
		_loop77:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s2 = LT(1);
				match(IDENT);
				if ( inputState.guessing==0 ) {
					
					if(s2.getText().charAt(0) == '#') {
					ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s2.getText()));
					if(msv == null)
					throw new antlr.SemanticException("Schema variable "+s2.getText()+ " not defined");
					for(int i=0; i<msv.modalities().length; i++)
					args.add(msv.modalities()[i]);
					} else {
					Modality op = Op.getModality(s2.getText());
					if(op == null)
					throw new antlr.SemanticException("Unrecognised modal operator "+s2.getText());
					args.add(op);
					}
					
				}
			}
			else {
				break _loop77;
			}
			
		} while (true);
		}
		match(RBRACE);
		}
		if ( inputState.guessing==0 ) {
			
			modalities = (Modality[])args.toArray(new Modality[args.size()]);
			
		}
		return modalities;
	}
	
	public final Sort[]  arg_sorts(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort[] argSorts = null;
		
		
		List args = new LinkedList();
		Sort s;
		
		
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			s=sortId_check(checkSort);
			if ( inputState.guessing==0 ) {
				args.add(s);
			}
			{
			_loop89:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					s=sortId_check(checkSort);
					if ( inputState.guessing==0 ) {
						args.add(s);
					}
				}
				else {
					break _loop89;
				}
				
			} while (true);
			}
			match(RPAREN);
			break;
		}
		case SEMI:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
			
		}
		return argSorts;
	}
	
	public final int  integer_num() throws RecognitionException, TokenStreamException {
		 int num = 0 ;
		
		
		if ( inputState.guessing==0 ) {
			
			try {
			Token t = LT(1);
			num = Integer.decode ( t.getText () ).intValue ();
			consume ();
			} catch ( TokenStreamException e ) {
			throw new antlr.SemanticException ( "Expected a number",
			getFilename (), getLine (), getColumn () );
			} catch ( NumberFormatException e ) {
			throw new antlr.SemanticException ( "Expected a number",
			getFilename (), getLine (), getColumn () );
			} catch ( NullPointerException e ) {
			throw new antlr.SemanticException ( "Expected a number",
			getFilename (), getLine (), getColumn () );
			}
			
		}
		return num;
	}
	
	public final Sort  sortId() throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		s=sortId_check(true);
		return s;
	}
	
	public final Sort  sortId_check_help(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		s=any_sortId_check_help(checkSort);
		if ( inputState.guessing==0 ) {
			
			// don't allow generic sorts or collection sorts of
			// generic sorts at this point
			Sort t = s;
			while ( t instanceof CollectionSort )
			t = ((CollectionSort)t).elementSort ();
			if ( t instanceof GenericSort )
			throw new GenericSortException ( "sort",
			"Non-generic sort expected", t,
			getFilename (), getLine (), getColumn () );
			
		}
		return s;
	}
	
	public final Sort  array_decls(
		Sort p
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		int  n=0;
		
		
		{
		_loop102:
		do {
			if ((LA(1)==LBRACKET)) {
				match(LBRACKET);
				match(RBRACKET);
				if ( inputState.guessing==0 ) {
					n++;
				}
			}
			else {
				break _loop102;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			
			if(n != 0){
			s = ArraySortImpl.getArraySortForDim(
			p, n, getJavaInfo().getObjectSort(),
			getJavaInfo().getCloneableSort());
			}else{
			s = p;
			}
			
		}
		return s;
	}
	
	public final Sort  any_sortId_check_help(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		String name;
		
		
		name=id();
		if ( inputState.guessing==0 ) {
			s= (Sort) sorts().lookup(new Name(name));
			if (checkSort && s == null) {
			throw new NotDeclException("sort", name, 
			getFilename(), 
			getLine(), getColumn());
			}
			
		}
		return s;
	}
	
	public final IdDeclaration  id_declaration() throws RecognitionException, TokenStreamException {
		 IdDeclaration idd = null ;
		
		Token  id = null;
		
		Sort s = null;
		
		
		id = LT(1);
		match(IDENT);
		{
		switch ( LA(1)) {
		case COLON:
		{
			match(COLON);
			s=sortId_check(true);
			break;
		}
		case EOF:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			idd = new IdDeclaration ( id.getText (), s );
			
		}
		return idd;
	}
	
	
	public static final String[] _tokenNames = {
		"<0>",
		"EOF",
		"<2>",
		"NULL_TREE_LOOKAHEAD",
		"\"new\"",
		"\"branch\"",
		"\"notsimple\"",
		"\"include\"",
		"\"LDTs\"",
		"\"nostandardrules\"",
		"\"typeof\"",
		"\"object\"",
		"\"number\"",
		"\"intliteral\"",
		"\"longliteral\"",
		"\"generic\"",
		"\"extends\"",
		"\"oneof\"",
		"\"program\"",
		"\"svlist\"",
		"\"all\"",
		"\"ex\"",
		"\"true\"",
		"\"false\"",
		"\"find\"",
		"\"if\"",
		"\"varcond\"",
		"\"containsquery\"",
		"\"noninteractive\"",
		"\"recursive\"",
		"\"displayname\"",
		"\"helptext\"",
		"\"not\"",
		"\"free\"",
		"\"in\"",
		"\"depending\"",
		"\"on\"",
		"\"close\"",
		"\"goal\"",
		"\"replacewith\"",
		"\"add\"",
		"\"addrules\"",
		"\"addprogvars\"",
		"\"heuristics\"",
		"\"hasSort\"",
		"\"sorts\"",
		"\"options\"",
		"\"with\"",
		"\"schema\"",
		"\"variables\"",
		"\"local\"",
		"\"rigid\"",
		"\"predicates\"",
		"\"functions\"",
		"\"formula\"",
		"\"modal\"",
		"\"operator\"",
		"\"rules\"",
		"\"proof\"",
		"\"rule\"",
		"\"term\"",
		"\"inst\"",
		"\"ifseqformula\"",
		"\"interactive\"",
		"\"heur\"",
		"\"modifies\"",
		"\"contracts\"",
		"\"problem\"",
		"\"#inType\"",
		"\"#isObject\"",
		"\"#inLong\"",
		"\"#inInt\"",
		"\"#inChar\"",
		"\"#inShort\"",
		"\"#inByte\"",
		"\"#add\"",
		"\"#sub\"",
		"\"#mul\"",
		"\"#div\"",
		"\"#jdiv\"",
		"\"#mod\"",
		"\"#jmod\"",
		"\"#less\"",
		"\"#greater\"",
		"\"#leq\"",
		"\"#geq\"",
		"\"#eq\"",
		"\"#JavaIntUnaryMinus\"",
		"\"#JavaIntAdd\"",
		"\"#JavaIntSub\"",
		"\"#JavaIntMul\"",
		"\"#JavaIntDiv\"",
		"\"#JavaIntMod\"",
		"\"#JavaIntAnd\"",
		"\"#JavaIntOr\"",
		"\"#JavaIntXor\"",
		"\"#JavaIntComplement\"",
		"\"#JavaIntShiftRight\"",
		"\"#JavaIntShiftLeft\"",
		"\"#JavaIntUnsignedShiftRight\"",
		"\"#JavaLongUnaryMinus\"",
		"\"#JavaLongAdd\"",
		"\"#JavaLongSub\"",
		"\"#JavaLongMul\"",
		"\"#JavaLongDiv\"",
		"\"#JavaLongMod\"",
		"\"#JavaLongAnd\"",
		"\"#JavaLongOr\"",
		"\"#JavaLongXor\"",
		"\"#JavaLongComplement\"",
		"\"#JavaLongShiftRight\"",
		"\"#JavaLongShiftLeft\"",
		"\"#JavaLongUnsignedShiftRight\"",
		"\"#moduloByte\"",
		"\"#moduloShort\"",
		"\"#moduloInteger\"",
		"\"#moduloLong\"",
		"\"#staticanalysis\"",
		"\"#ResolveQuery\"",
		"\"#constantvalue\"",
		"\"#lengthReference\"",
		"\"#transient\"",
		"\"#shadowed\"",
		"\"#concat\"",
		"\"#introduce\"",
		"\"waryAll\"",
		"\"waryEx\"",
		"\"same\"",
		"\"compatible\"",
		"\"sameUpdateLevel\"",
		"\"smaller\"",
		"\"than\"",
		"\"quotes\"",
		"\"ThisReference\"",
		"\"keyLog\"",
		"\"keyUser\"",
		"\"keyVersion\"",
		"\"keySettings\"",
		"\"ocl\"",
		"\"#allSubtypes\"",
		"VOCAB",
		"`;'",
		"`/'",
		"`\\'",
		"`:'",
		"`?='",
		"`/->'",
		"`:='",
		"`.'",
		"`,'",
		"`('",
		"`)'",
		"`{'",
		"`}'",
		"`['",
		"']'",
		"`at'",
		"`|'",
		"`&'",
		"`->'",
		"`<->'",
		"`='",
		"] followed by a sequence of primes (at least one)",
		"`==>'",
		"'^'",
		"java operator",
		"white space",
		"a string in double quotes",
		"PRIMES_OR_CHARLITERAL",
		"Sequence of primes (at least one)",
		"a char in single quotes",
		"ESC",
		"DIAMOND",
		"BOX",
		"THROUGHOUT",
		"MODALITYARG",
		"BLOCK",
		"a string with double quotes",
		"a comment",
		"a comment",
		"a digit",
		"a hexadeciamal number",
		"a letter",
		"an admissible character for identifiers",
		"an admissible character for identifiers protected with ``",
		"PIDENT",
		"an identifer"
	};
	
	private static final long[] mk_tokenSet_0() {
		long[] data = { 34816L, 0L, 288230376151712768L, 0L, 0L, 0L};
		return data;
	}
	public static final BitSet _tokenSet_0 = new BitSet(mk_tokenSet_0());
	
	}
