// $ANTLR 2.7.2: "expandedtaclets.g" -> "TacletParser.java"$


package de.uka.ilkd.key.parser;

import java.io.*;
import antlr.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;    
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.metaconstruct.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.init.Includes;

import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.SchemaJavaReader;
import de.uka.ilkd.key.java.SchemaRecoder2KeY;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.ListOfKeYJavaType;
import de.uka.ilkd.key.java.abstraction.SLListOfKeYJavaType;
import de.uka.ilkd.key.java.visitor.DeclarationProgramVariableCollector;
import java.util.List;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import de.uka.ilkd.key.gui.AbbrevMap;

import antlr.TokenBuffer;
import antlr.TokenStreamException;
import antlr.TokenStreamIOException;
import antlr.ANTLRException;
import antlr.LLkParser;
import antlr.Token;
import antlr.TokenStream;
import antlr.RecognitionException;
import antlr.NoViableAltException;
import antlr.MismatchedTokenException;
import antlr.SemanticException;
import antlr.ParserSharedInputState;
import antlr.collections.impl.BitSet;

public class TacletParser extends antlr.LLkParser       implements TacletParserTokenTypes
 {

            
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected HashMap category2Default = new HashMap();
    protected HashMap taclet2Builder;
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();

    private Prioritization.StaticPrioritization heur_prioritization
        = new Prioritization.StaticPrioritization ();

    private NamespaceSet nss;
    private AbbrevMap scm = new AbbrevMap();


    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    private boolean skip_functions;
    private boolean skip_predicates;
    private boolean skip_sorts;
    private boolean skip_schemavariables;
    private boolean skip_heuristics;
    private Choice defaultChoice;

    private TermFactory tf;
    private SchemaJavaReader javaReader;
    private Services services;

    private boolean errors = false;

    public TacletParser(TokenStream lexer,
                        String filename,
                        TermFactory tf,
                        SchemaJavaReader jr,
                        Services services,  
                        NamespaceSet nss,
                        Choice defaultChoice,
                        HashMap taclet2Builder) {
        this(lexer,2);
        setFilename(filename);
        this.tf = tf;
        this.javaReader = jr;
        this.services = services;
        this.nss  = nss;    
        switchToSchemaMode();
        this.defaultChoice = defaultChoice;
        this.taclet2Builder = taclet2Builder;
    }

    public TacletParser(TokenStream lexer,
                        String filename,
                        TermFactory tf,
                        Services services,
                        NamespaceSet nss,
                        Choice defaultChoice) {
        this(lexer, filename, tf,
             new SchemaRecoder2KeY(services, nss), services, nss, defaultChoice, new HashMap());
    }

    public TacletParser(TokenStream lexer,
                        String filename,
                        TermFactory tf,
                        Services services,
                        NamespaceSet nss) {
        this(lexer, filename, tf,
             new SchemaRecoder2KeY(services, nss), services, nss, 
             new Choice(new Name("Default")), new HashMap());
    }

    public Namespace sorts() {
        return nss.sorts();
    }

    public Namespace choices() {
        return nss.choices();
    }

    public Namespace functions() {
        return nss.functions();
    }

    public Namespace heuristics() {
        return nss.heuristics();
    }

    public Namespace variables() {
        return nss.variables();
    }

    private Namespace programVariables() {
        return nss.programVariables();
    }

    public NamespaceSet namespaces() {
        return nss;
    }

    public Named lookup(Name n) {
	final Namespace[] lookups = {
		functions(), variables(), 
		programVariables()
	};
	for (int i = 0; i<lookups.length; i++) {
	    if (lookups[i].lookup(n) != null) {
		return lookups[i].lookup(n);
            }
	}
	return null;	
    }

    protected boolean isPackage(String name){
        try {
            return getJavaInfo().isPackage(name);
        } catch(RuntimeException e){	    
	    // may be thrown in case of invalid an java identifier
            return false;
        } 
    }

    public Term createAttributeTerm(Term prefix, 
				    TermSymbol attribute,
				    int shadowPrimes) {
	Term result = prefix;
	if (!inSchemaMode()) {
	    if (((ProgramVariable)attribute).isStatic()){
		result = tf.createVariableTerm((ProgramVariable)attribute);
	    } else {
		if (shadowPrimes > 0) {
		    Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
		    result = tf.createShadowAttributeTerm((ProgramVariable)attribute, 
                                                              result, shadownum);
		} else {
		    result = tf.createAttributeTerm((ProgramVariable)attribute, result);
		}
	    }
	} else {
	    if(shadowPrimes>0){
		Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
                    result = tf.createShadowAttributeTerm((SchemaVariable)attribute, result, shadownum);         
	    } else
                result = tf.createAttributeTerm((SchemaVariable)attribute, result);         
	}

	return result;
    }

    public TermSymbol getAttribute(Sort prefixSort, 
				   String attributeName) 
	throws antlr.SemanticException 	{
	    final JavaInfo javaInfo = getJavaInfo();
	    TermSymbol v = null;
            if (!inSchemaMode()) {
		if (attributeName.indexOf(':') != -1) {
		   v = javaInfo.getAttribute(attributeName, 
			javaInfo.getKeYJavaType(attributeName.substring(0, 
						attributeName.indexOf(':'))));
	        } else {
		    v = javaInfo.getAttribute(attributeName, (ObjectSort) prefixSort);
                }
            } else {
                v = (SortedSchemaVariable)variables().lookup(new Name(attributeName));
            }
            if ( v == null && !"length".equals(attributeName) ) {
                throw new NotDeclException
                    ("Attribute ", attributeName, getFilename(), 
                     getLine(), getColumn());

            }

	return v;
    }


    // taclets are always parsed in in schema mode
    public void switchToSchemaMode() {
    }

    private boolean inSchemaMode() {
        return true;
    }

    public void switchToNormalMode() {
    }

    private TacletBuilder createTacletBuilderFor
        (Object find, boolean sameUpdatePrefix) 
        throws InvalidFindException {
        if ( sameUpdatePrefix && !( find instanceof Term ) )
            throw new InvalidFindException
                ("\"sameUpdateLevel\" may only be used for rewrite taclets:"+find,
                 getFilename(), getLine(), getColumn());
        if ( find == null ) {
            return new NoFindTacletBuilder();
        } else if ( find instanceof Term ) {
            return new RewriteTacletBuilder().setFind((Term)find)
                .setSameUpdatePrefix(sameUpdatePrefix);
        } else if ( find instanceof Sequent ) {
            Sequent findSeq = (Sequent) find;
            if ( findSeq == Sequent.EMPTY_SEQUENT ) {
                return new NoFindTacletBuilder();
            } else if (   findSeq.antecedent().size() == 1
                          && findSeq.succedent().size() == 0 ) {
                Term findFma = findSeq.antecedent().get(0).formula();
                return new AntecTacletBuilder().setFind(findFma);
            } else if (   findSeq.antecedent().size() == 0
                          && findSeq.succedent().size() == 1 ) {
                Term findFma = findSeq.succedent().get(0).formula();
                return new SuccTacletBuilder().setFind(findFma);
            } else {
                throw new InvalidFindException
                    ("Unknown find-sequent (perhaps null?):"+findSeq,
                     getFilename(), getLine(), getColumn());
            }
        } else {
            throw new RuntimeException("createTacletBuilderFor called with"
                                       + find.getClass().getName() + "object");
        }
    }       

    protected static HashSet progVars(JavaBlock jb) {
        DeclarationProgramVariableCollector pvc
            = new DeclarationProgramVariableCollector(jb.program());
        pvc.start();
        return pvc.result();
    }

    
    protected void bindProgVars(HashSet progVars) {
        namespaces().setVariables(new Namespace(variables()));
        Iterator it=progVars.iterator();
        while (it.hasNext()) {
            variables().add((IProgramVariable)it.next());
        }
    }

    protected void unbindProgVars() {
        namespaces().setVariables(variables().parent());
    }

    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }


    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }

    
    private static ListOfRulevariable NO_RULE_VARIABLES =
        SLListOfRulevariable.EMPTY_LIST;

    private void addGoalTemplate(TacletBuilder b,
                                 String id,
                                 Object rwObj,
                                 Sequent addSeq,
                                 ListOfTaclet addRList,
                                 SetOfSchemaVariable pvs,
                                 SetOfChoice soc) 
        throws SemanticException
        {
            TacletGoalTemplate gt = null;
            if ( rwObj == null ) {
                // there is no replacewith, so we take
                gt = new TacletGoalTemplate(addSeq,
                                            addRList,
                                            NO_RULE_VARIABLES,
                                            pvs);
            } else {
                if ( b instanceof NoFindTacletBuilder ) {
                    // there is a replacewith without a find.
                    throw 
                        new UnfittingReplacewithException
                        ("Replacewith without find", getFilename(),
                         getLine(), getColumn());
                } else if ( b instanceof SuccTacletBuilder
                            || b instanceof AntecTacletBuilder ) {
                    if ( rwObj instanceof Sequent ) {
                        gt = new AntecSuccTacletGoalTemplate(addSeq,
                                                             addRList,
                                                             NO_RULE_VARIABLES,
                                                             (Sequent)rwObj,
                                                             pvs);  
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a Antec-or SuccTaclet has "+
                             "to contain a sequent (not a term)", 
                             getFilename(), getLine(), getColumn());
                    }
                } else if ( b instanceof RewriteTacletBuilder ) {
                    if ( rwObj instanceof Term ) {
                        gt = new RewriteTacletGoalTemplate(addSeq,
                                                           addRList,
                                                           NO_RULE_VARIABLES,
                                                           (Term)rwObj,
                                                           pvs);  
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a RewriteTaclet has "+
                             "to contain a term (not a sequent)", 
                             getFilename(), getLine(), getColumn());
                    }
                }
            }
            gt.setName(id); 
            b.addTacletGoalTemplate(gt);
            if(soc != null) b.addGoal2ChoicesMapping(gt,soc);
        }
    


    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    protected int getColumn() {
        int col = -1;
        try {
            col = LT(0).getColumn();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }   

    protected Term termForParsedVariable(ParsableVariable v) 
        throws antlr.SemanticException 
        {
            if ( v instanceof LogicVariable ) {
                return tf.createVariableTerm((LogicVariable)v);
            } else if ( v instanceof SchemaVariable ) {
                return tf.createVariableTerm((SchemaVariable)v);
            } else if ( v instanceof ProgramVariable ) {
                return tf.createVariableTerm((ProgramVariable)v);
                //      } else if ( v instanceof ParsableVariable ) {
                //          return tf.createVariableTerm((ParsableVariable)v);

            } else {
                throw new antlr.SemanticException
                    (""+ v +" is not a logic or schema or program variable");
            }
        }

    protected TermSymbol progVar(Token id,Sort s) throws NotDeclException{
        TermSymbol v = (TermSymbol)
            variables().lookup(new Name(id.getText()));

        if (v == null) {
            throw new NotDeclException("program variable", id,
                                       getFilename());
        }
        return v;
    }

    protected LogicVariable bindVar(Token id,Sort s) {
        LogicVariable v=new LogicVariable(new Name(id.getText()), s);
        namespaces().setVariables(variables().extended(v));
        return v;
    }
    
    protected void unbindVars() {
        namespaces().setVariables(variables().parent());
    }

    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
        String s=t.getText();
        if (!inSchemaMode()) {
            try {
                return javaReader.readBlockWithProgramVariables(variables(), s);
            } catch (Exception e) {
                if (s.length()>10) {
                    s=s.substring(0,10)+"...";
                }
                throw new antlr.SemanticException("Reading Java block "
                                                  + s + " failed. "
                                                  +e.getMessage()); 
            }
        }
        javaReader.setSVNamespace(variables());
        return javaReader.readBlockWithEmptyContext(s);
    }

    /** returns the java information object */
    public JavaInfo getJavaInfo() {
        return getServices().getJavaInfo();
    }

    public Services getServices() {
        return services;
    }




       
    public void schema_var_decl(String name,Sort s,
                                    boolean makeVariableSV,
                                    boolean makeDependingSV,
                                    boolean rigid,
                                    boolean listModifier)
	{
              if (!skip_schemavariables) {
                  SchemaVariable v;
                  if ( s == Sort.FORMULA ) {
                      v = SchemaVariableFactory.createFormulaSV
                          (new Name(name), listModifier, rigid);
                      } else if ( s instanceof ProgramSVSort ) {
                          v = SchemaVariableFactory.createProgramSV
                         (new ProgramElementName(name),(ProgramSVSort) s, listModifier);
                      } else {
                        if ( makeVariableSV ) {
                             v=SchemaVariableFactory.createVariableSV
                             (new Name(name), s, listModifier);
                        } else if ( makeDependingSV ) {
                               v = SchemaVariableFactory.createDependingSV
                             (new Name(name), s, listModifier);
                        } else { v = SchemaVariableFactory.createTermSV
                              (new Name(name), s, listModifier, rigid);
                        }
                     }
                     if (inSchemaMode()) {
                         variables().add(v);
                     }
               }

         }
     


    public void testLiteral(String l1, String l2)
    throws KeYSemanticException
    {
     if (!l1.equals(l2)){
        throw new KeYSemanticException
        ("Expecting '"+l1+"', found '"+l2+"'.",getLine(),getColumn(),getFilename());
	};
    }
    /** looks up a function, (program) variable or static query of the 
     * given name varfunc_id and the argument terms args in the namespaces 
     * and java info. 
     * @param args is null iff no argument list is given, for instance `f', 
     * and is an array of size zero, if an empty argument list was given,
     * for instance `f()'.
     */
    protected Operator lookupVarfuncId(String varfunc_name, Term[] args) 
                                                    throws NotDeclException{
        // case 1: variable
        Operator v = (TermSymbol) variables().lookup(new Name(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        // case 2: function
        v = (Function) functions().lookup(new Name(varfunc_name));
        if (v != null) { // we allow both args==null (e.g. `c')
                         // and args.length=0 (e.g. 'c())' here 
            return v;
        }
        // case 3: static query
        if (varfunc_name.lastIndexOf("::")>0 && args!=null) {
            String classname = varfunc_name.substring
            (0, varfunc_name.lastIndexOf("::"));
            varfunc_name = varfunc_name.substring
            (varfunc_name.lastIndexOf("::")+2);
            v = getServices().getJavaInfo().getProgramMethodTerm
                (null, varfunc_name, args, classname).op();
            if (v != null) {
                return v;
            }
        }
        // case 4: program variable
        v = (ProgramVariable) programVariables().lookup
            (new ProgramElementName(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        if (args==null) {
            throw new NotDeclException
            ("(program) variable or constant", varfunc_name,
                getFilename(), getLine(), getColumn());
        } else {
            throw new NotDeclException
            ("function or static query", varfunc_name,
                getFilename(), getLine(), getColumn());
        }
    }



protected TacletParser(TokenBuffer tokenBuf, int k) {
  super(tokenBuf,k);
  tokenNames = _tokenNames;
}

public TacletParser(TokenBuffer tokenBuf) {
  this(tokenBuf,1);
}

protected TacletParser(TokenStream lexer, int k) {
  super(lexer,k);
  tokenNames = _tokenNames;
}

public TacletParser(TokenStream lexer) {
  this(lexer,1);
}

public TacletParser(ParserSharedInputState state) {
  super(state,1);
  tokenNames = _tokenNames;
}

	public final void top() throws RecognitionException, TokenStreamException {
		
		Taclet r;
		
		r=taclet(SetAsListOfChoice.EMPTY_SET);
	}
	
	public final Taclet  taclet(
		SetOfChoice choices
	) throws RecognitionException, TokenStreamException {
		Taclet r;
		
		Token  name = null;
		
		Sequent ifSeq = Sequent.EMPTY_SEQUENT;
		Object  find = null;
		r = null;
		TacletBuilder b = null;
		boolean sameUpdatePrefix = false;
		
		
		name = LT(1);
		match(IDENT);
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			choices=option(choices);
			{
			_loop5:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					choices=option(choices);
				}
				else {
					break _loop5;
				}
				
			} while (true);
			}
			match(RPAREN);
			break;
		}
		case LBRACE:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		match(LBRACE);
		{
		switch ( LA(1)) {
		case IF:
		{
			match(IF);
			match(LPAREN);
			ifSeq=seq();
			match(RPAREN);
			break;
		}
		case FIND:
		case VARCOND:
		case CLOSE:
		case REPLACEWITH:
		case ADD:
		case ADDRULES:
		case LPAREN:
		case STRING_LITERAL:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		{
		switch ( LA(1)) {
		case FIND:
		{
			match(FIND);
			match(LPAREN);
			find=termorseq();
			match(RPAREN);
			{
			switch ( LA(1)) {
			case SAME_UPDATE_LEVEL:
			{
				match(SAME_UPDATE_LEVEL);
				if ( inputState.guessing==0 ) {
					sameUpdatePrefix = true;
				}
				break;
			}
			case VARCOND:
			case CLOSE:
			case REPLACEWITH:
			case ADD:
			case ADDRULES:
			case LPAREN:
			case STRING_LITERAL:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			break;
		}
		case VARCOND:
		case CLOSE:
		case REPLACEWITH:
		case ADD:
		case ADDRULES:
		case LPAREN:
		case STRING_LITERAL:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			b = createTacletBuilderFor(find, sameUpdatePrefix);
			b.setName(new Name(name.getText()));
			b.setIfSequent(ifSeq);
			
		}
		{
		switch ( LA(1)) {
		case VARCOND:
		{
			match(VARCOND);
			match(LPAREN);
			varexplist(b);
			match(RPAREN);
			break;
		}
		case CLOSE:
		case REPLACEWITH:
		case ADD:
		case ADDRULES:
		case LPAREN:
		case STRING_LITERAL:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		goalspecs(b);
		modifiers(b);
		match(RBRACE);
		if ( inputState.guessing==0 ) {
			
			b.setChoices(choices);
			r = b.getTaclet(); 
			taclet2Builder.put(r,b);
			
		}
		return r;
	}
	
	public final SetOfChoice  option(
		SetOfChoice soc
	) throws RecognitionException, TokenStreamException {
		SetOfChoice result=null;
		
		Token  cat = null;
		Token  choice = null;
		
		Choice c = null;
		
		
		cat = LT(1);
		match(IDENT);
		match(COLON);
		choice = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			
			c = (Choice) choices().lookup(new Name(cat.getText()+":"+choice.getText()));
			if(c==null){
			throw new NotDeclException("Option", choice,
			getFilename());
			}else{
			result = soc.add(c);
			}
			
		}
		return result;
	}
	
	public final Sequent  seq() throws RecognitionException, TokenStreamException {
		Sequent s;
		
		Semisequent ant,suc; s = null;
		
		ant=semisequent();
		match(SEQARROW);
		suc=semisequent();
		if ( inputState.guessing==0 ) {
			s = Sequent.createSequent(ant, suc);
		}
		return s;
	}
	
	public final Object  termorseq() throws RecognitionException, TokenStreamException {
		Object o;
		
		
		Term head = null;
		Sequent s = null;
		Semisequent ss = null;
		o = null; 
		
		
		switch ( LA(1)) {
		case FORALL:
		case EXISTS:
		case TRUE:
		case FALSE:
		case NOT:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case LPAREN:
		case LBRACE:
		case AT:
		case CTRL:
		case ARITH_OP:
		case CHAR_LITERAL:
		case DIAMOND:
		case BOX:
		case THROUGHOUT:
		case PIDENT:
		case IDENT:
		{
			head=term();
			{
			switch ( LA(1)) {
			case COMMA:
			{
				match(COMMA);
				s=seq();
				break;
			}
			case SEQARROW:
			{
				match(SEQARROW);
				ss=semisequent();
				break;
			}
			case RPAREN:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			if ( inputState.guessing==0 ) {
				
				if ( s == null ) {
				if ( ss == null ) {
				// Just a term
				o = head;
				} else {
				// A sequent with only head in the antecedent.
				Semisequent ant = Semisequent.EMPTY_SEMISEQUENT;
				ant = ant.insertFirst(
				new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
				o = Sequent.createSequent(ant,ss);
				}
				} else {
				// A sequent.  Prepend head to the antecedent.
				Semisequent newAnt = s.antecedent();
				newAnt = newAnt .insertFirst(
				new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
				o = Sequent.createSequent(newAnt,s.succedent());
				}
				
			}
			break;
		}
		case SEQARROW:
		{
			match(SEQARROW);
			ss=semisequent();
			if ( inputState.guessing==0 ) {
				
				o = Sequent.createSequent(Semisequent.EMPTY_SEMISEQUENT,ss);
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return o;
	}
	
	public final void varexplist(
		TacletBuilder b
	) throws RecognitionException, TokenStreamException {
		
		
		varexp(b);
		{
		_loop20:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				varexp(b);
			}
			else {
				break _loop20;
			}
			
		} while (true);
		}
	}
	
	public final void goalspecs(
		TacletBuilder b
	) throws RecognitionException, TokenStreamException {
		
		
		switch ( LA(1)) {
		case CLOSE:
		{
			match(CLOSE);
			match(GOAL);
			break;
		}
		case REPLACEWITH:
		case ADD:
		case ADDRULES:
		case LPAREN:
		case STRING_LITERAL:
		{
			goalspecwithoption(b);
			{
			_loop36:
			do {
				if ((LA(1)==SEMI)) {
					match(SEMI);
					goalspecwithoption(b);
				}
				else {
					break _loop36;
				}
				
			} while (true);
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
	}
	
	public final void modifiers(
		TacletBuilder b
	) throws RecognitionException, TokenStreamException {
		
		Token  dname = null;
		Token  htext = null;
		
		{
		_loop12:
		do {
			switch ( LA(1)) {
			case HEURISTICS:
			{
				heuristics(b);
				break;
			}
			case NONINTERACTIVE:
			{
				match(NONINTERACTIVE);
				if ( inputState.guessing==0 ) {
					
					//      b.setNoninteractive(true);  
					// "noninteractive" (as it is now) is confusing
					// dropped this completely until a better solution (->Uwe)
					
				}
				break;
			}
			case RECURSIVE:
			{
				match(RECURSIVE);
				if ( inputState.guessing==0 ) {
					b.setRecursive(true);
				}
				break;
			}
			case DISPLAYNAME:
			{
				match(DISPLAYNAME);
				dname = LT(1);
				match(STRING_LITERAL);
				if ( inputState.guessing==0 ) {
					b.setDisplayName(dname.getText());
				}
				break;
			}
			case HELPTEXT:
			{
				match(HELPTEXT);
				htext = LT(1);
				match(STRING_LITERAL);
				if ( inputState.guessing==0 ) {
					b.setHelpText(htext.getText());
				}
				break;
			}
			default:
			{
				break _loop12;
			}
			}
		} while (true);
		}
	}
	
	public final void heuristics(
		TacletBuilder b
	) throws RecognitionException, TokenStreamException {
		
		
		match(HEURISTICS);
		match(LPAREN);
		heuristic(b);
		{
		_loop63:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				heuristic(b);
			}
			else {
				break _loop63;
			}
			
		} while (true);
		}
		match(RPAREN);
	}
	
	public final Semisequent  semisequent() throws RecognitionException, TokenStreamException {
		Semisequent ss;
		
		
		Term head = null;
		ss = Semisequent.EMPTY_SEMISEQUENT; 
		
		
		switch ( LA(1)) {
		case RPAREN:
		case SEQARROW:
		{
			break;
		}
		case FORALL:
		case EXISTS:
		case TRUE:
		case FALSE:
		case NOT:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case LPAREN:
		case LBRACE:
		case AT:
		case CTRL:
		case ARITH_OP:
		case CHAR_LITERAL:
		case DIAMOND:
		case BOX:
		case THROUGHOUT:
		case PIDENT:
		case IDENT:
		{
			head=term();
			{
			switch ( LA(1)) {
			case COMMA:
			{
				match(COMMA);
				ss=semisequent();
				break;
			}
			case RPAREN:
			case SEQARROW:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			if ( inputState.guessing==0 ) {
				ss = ss.insertFirst(new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return ss;
	}
	
	public final Term  term() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		a=term20();
		return a;
	}
	
	public final void varexp(
		TacletBuilder b
	) throws RecognitionException, TokenStreamException {
		
		Token  mId = null;
		Token  sId1 = null;
		Token  sId2 = null;
		Token  sId4 = null;
		Token  sId3 = null;
		
		ParsableVariable x;
		ParsableVariable y; 
		ParsableVariable z;
		Sort s = null;
		boolean negation = false;
		
		
		{
		switch ( LA(1)) {
		case TYPEOF:
		{
			match(TYPEOF);
			match(LPAREN);
			y=varId();
			match(RPAREN);
			{
			switch ( LA(1)) {
			case IDENT:
			{
				x=varId();
				match(NEW);
				if ( inputState.guessing==0 ) {
					b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, false);
				}
				break;
			}
			case NOT:
			{
				match(NOT);
				{
				switch ( LA(1)) {
				case SAME:
				{
					match(SAME);
					match(TYPEOF);
					match(LPAREN);
					x=varId();
					match(RPAREN);
					if ( inputState.guessing==0 ) {
						b.addVariableCondition
									   (new TypeComparisionCondition((SchemaVariable) y, (SchemaVariable) x, 
										TypeComparisionCondition.NOT_SAME));
					}
					break;
				}
				case COMPATIBLE:
				{
					match(COMPATIBLE);
					match(TYPEOF);
					match(LPAREN);
					x=varId();
					match(RPAREN);
					if ( inputState.guessing==0 ) {
						b.addVariableCondition
									   (new TypeComparisionCondition((SchemaVariable) y, (SchemaVariable) x, 
										TypeComparisionCondition.NOT_COMPATIBLE));
					}
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			break;
		}
		case LITERAL_elemtypeof:
		{
			match(LITERAL_elemtypeof);
			match(LPAREN);
			y=varId();
			match(RPAREN);
			x=varId();
			match(NEW);
			if ( inputState.guessing==0 ) {
				b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, true);
			}
			break;
		}
		default:
			boolean synPredMatched26 = false;
			if (((LA(1)==PIDENT||LA(1)==IDENT))) {
				int _m26 = mark();
				synPredMatched26 = true;
				inputState.guessing++;
				try {
					{
					sortId_check(true);
					varId();
					match(NEW);
					}
				}
				catch (RecognitionException pe) {
					synPredMatched26 = false;
				}
				rewind(_m26);
				inputState.guessing--;
			}
			if ( synPredMatched26 ) {
				s=sortId_check(true);
				x=varId();
				match(NEW);
				if ( inputState.guessing==0 ) {
					b.addVarsNew((SchemaVariable) x, s);
				}
			}
			else {
				boolean synPredMatched28 = false;
				if (((LA(1)==IDENT))) {
					int _m28 = mark();
					synPredMatched28 = true;
					inputState.guessing++;
					try {
						{
						match(IDENT);
						varId();
						varId();
						varId();
						}
					}
					catch (RecognitionException pe) {
						synPredMatched28 = false;
					}
					rewind(_m28);
					inputState.guessing--;
				}
				if ( synPredMatched28 ) {
					{
					mId = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						testLiteral("methodreference",mId.getText());
					}
					x=varId();
					y=varId();
					z=varId();
					{
					switch ( LA(1)) {
					case NOT:
					{
						match(NOT);
						if ( inputState.guessing==0 ) {
							negation = true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					sId1 = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						
						testLiteral("static",sId1.getText());
								    b.addVariableCondition(new StaticMethodCondition
						(negation, (SchemaVariable)x, (SchemaVariable)y,
						(SchemaVariable)z));
						negation = false; 
						
					}
					}
				}
				else if ((LA(1)==IDENT)) {
					x=varId();
					{
					switch ( LA(1)) {
					case NOT:
					{
						match(NOT);
						{
						switch ( LA(1)) {
						case FREE:
						{
							match(FREE);
							match(IN);
							y=varId();
							if ( inputState.guessing==0 ) {
								b.addVarsNotFreeIn((SchemaVariable)x,(SchemaVariable)y);
							}
							break;
						}
						case IDENT:
						{
							{
							if (((LA(1)==IDENT))&&("static".equals(LT(1).getText()))) {
								sId2 = LT(1);
								match(IDENT);
								if ( inputState.guessing==0 ) {
									testLiteral("static",sId2.getText());
											      b.addVariableCondition(new StaticReferenceCondition
									((SchemaVariable) x, true));
								}
							}
							else if ((LA(1)==IDENT)) {
								sId4 = LT(1);
								match(IDENT);
								y=varId();
								if ( inputState.guessing==0 ) {
									testLiteral("sameLiteral",sId4.getText());
											      b.addVariableCondition(new TestLiteral
									((SchemaVariable) x, (SchemaVariable) y));
											
								}
							}
							else {
								throw new NoViableAltException(LT(1), getFilename());
							}
							
							}
							break;
						}
						default:
						{
							throw new NoViableAltException(LT(1), getFilename());
						}
						}
						}
						break;
					}
					case IDENT:
					{
						sId3 = LT(1);
						match(IDENT);
						if ( inputState.guessing==0 ) {
							testLiteral("static",sId3.getText());
									   b.addVariableCondition
							(new StaticReferenceCondition((SchemaVariable) x,
							false));
						}
						break;
					}
					case NEW:
					{
						match(NEW);
						match(DEPENDING);
						match(ON);
						y=varId();
						if ( inputState.guessing==0 ) {
							b.addVarsNewDependingOn((SchemaVariable)x,(SchemaVariable)y);
						}
						break;
					}
					case SMALLER:
					{
						match(SMALLER);
						match(THAN);
						y=varId();
						if ( inputState.guessing==0 ) {
							b.addVariableCondition(
							new OrderingCondition ((SchemaVariable)x,(SchemaVariable)y));
						}
						break;
					}
					case CONTAINSQUERY:
					{
						match(CONTAINSQUERY);
						if ( inputState.guessing==0 ) {
							b.addVariableCondition(new ContainsQueryCondition
							((SchemaVariable)x));
						}
						break;
					}
					case HASSORT:
					{
						match(HASSORT);
						s=any_sortId_check(true);
						if ( inputState.guessing==0 ) {
							if ( !( s instanceof GenericSort ) )
							throw new GenericSortException ( "sort",
							"Generic sort expected", s,
							getFilename (), getLine (), getColumn () );
							if ( !( x instanceof SortedSchemaVariable
							&& ( (SortedSchemaVariable)x ).sort () ==
							ProgramSVSort.EXPRESSION ) )
							throw new antlr.SemanticException
							("Expected schema variable of kind EXPRESSION, got " + x);
							b.addVariableCondition(new JavaTypeToSortCondition ((SchemaVariable)x, (GenericSort)s));
						}
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
				}
			else {
				throw new NoViableAltException(LT(1), getFilename());
			}
			}}
			}
		}
		
	public final ParsableVariable  varId() throws RecognitionException, TokenStreamException {
		ParsableVariable v = null;
		
		Token  id = null;
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			
			v = (ParsableVariable) variables().lookup(new Name(id.getText()));
			if (v == null) {
			throw new NotDeclException("variable", id, 
			getFilename());
			}
			
		}
		return v;
	}
	
	public final Sort  sortId_check(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		Sort t;
		
		
		t=sortId_check_help(checkSort);
		s=array_decls(t);
		return s;
	}
	
	public final Sort  any_sortId_check(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		Sort t;
		
		
		t=any_sortId_check_help(checkSort);
		s=array_decls(t);
		return s;
	}
	
	public final void goalspecwithoption(
		TacletBuilder b
	) throws RecognitionException, TokenStreamException {
		
		
		SetOfChoice soc = SetAsListOfChoice.EMPTY_SET;
		
		
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			{
			match(LPAREN);
			soc=option(soc);
			{
			_loop41:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					soc=option(soc);
				}
				else {
					break _loop41;
				}
				
			} while (true);
			}
			match(RPAREN);
			match(LBRACE);
			goalspec(b,soc);
			match(RBRACE);
			}
			break;
		}
		case REPLACEWITH:
		case ADD:
		case ADDRULES:
		case STRING_LITERAL:
		{
			goalspec(b,null);
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
	}
	
	public final void goalspec(
		TacletBuilder b, SetOfChoice soc
	) throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		Object rwObj = null;
		Sequent addSeq = Sequent.EMPTY_SEQUENT;
		ListOfTaclet addRList = SLListOfTaclet.EMPTY_LIST;
		SetOfSchemaVariable addpv = SetAsListOfSchemaVariable.EMPTY_SET;
		String name = null;
		
		
		{
		switch ( LA(1)) {
		case STRING_LITERAL:
		{
			id = LT(1);
			match(STRING_LITERAL);
			match(COLON);
			if ( inputState.guessing==0 ) {
				name = id.getText();
			}
			break;
		}
		case REPLACEWITH:
		case ADD:
		case ADDRULES:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		{
		switch ( LA(1)) {
		case REPLACEWITH:
		{
			{
			rwObj=replacewith();
			{
			switch ( LA(1)) {
			case ADD:
			{
				addSeq=add();
				break;
			}
			case NONINTERACTIVE:
			case RECURSIVE:
			case DISPLAYNAME:
			case HELPTEXT:
			case ADDRULES:
			case ADDPROGVARS:
			case HEURISTICS:
			case SEMI:
			case RBRACE:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			{
			switch ( LA(1)) {
			case ADDRULES:
			{
				addRList=addrules();
				break;
			}
			case NONINTERACTIVE:
			case RECURSIVE:
			case DISPLAYNAME:
			case HELPTEXT:
			case ADDPROGVARS:
			case HEURISTICS:
			case SEMI:
			case RBRACE:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			{
			switch ( LA(1)) {
			case ADDPROGVARS:
			{
				addpv=addprogvar();
				break;
			}
			case NONINTERACTIVE:
			case RECURSIVE:
			case DISPLAYNAME:
			case HELPTEXT:
			case HEURISTICS:
			case SEMI:
			case RBRACE:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			}
			break;
		}
		case ADD:
		{
			{
			addSeq=add();
			{
			switch ( LA(1)) {
			case ADDRULES:
			{
				addRList=addrules();
				break;
			}
			case NONINTERACTIVE:
			case RECURSIVE:
			case DISPLAYNAME:
			case HELPTEXT:
			case HEURISTICS:
			case SEMI:
			case RBRACE:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			}
			break;
		}
		case ADDRULES:
		{
			{
			addRList=addrules();
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			addGoalTemplate(b,name,rwObj,addSeq,addRList,addpv,soc);
			
		}
	}
	
	public final Object  replacewith() throws RecognitionException, TokenStreamException {
		Object o;
		
		o = null;
		
		match(REPLACEWITH);
		match(LPAREN);
		o=termorseq();
		match(RPAREN);
		return o;
	}
	
	public final Sequent  add() throws RecognitionException, TokenStreamException {
		Sequent s;
		
		s = null;
		
		match(ADD);
		match(LPAREN);
		s=seq();
		match(RPAREN);
		return s;
	}
	
	public final ListOfTaclet  addrules() throws RecognitionException, TokenStreamException {
		ListOfTaclet lor;
		
		lor = null;
		
		match(ADDRULES);
		match(LPAREN);
		lor=tacletlist();
		match(RPAREN);
		return lor;
	}
	
	public final SetOfSchemaVariable  addprogvar() throws RecognitionException, TokenStreamException {
		SetOfSchemaVariable pvs;
		
		pvs = null;
		
		match(ADDPROGVARS);
		match(LPAREN);
		pvs=pvset();
		match(RPAREN);
		return pvs;
	}
	
	public final ListOfTaclet  tacletlist() throws RecognitionException, TokenStreamException {
		ListOfTaclet lor;
		
		
		Taclet head = null;
		lor = SLListOfTaclet.EMPTY_LIST; 
		
		
		head=taclet(SetAsListOfChoice.EMPTY_SET);
		{
		switch ( LA(1)) {
		case RPAREN:
		{
			break;
		}
		case COMMA:
		{
			match(COMMA);
			lor=tacletlist();
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			lor = lor.prepend(head);
		}
		return lor;
	}
	
	public final SetOfSchemaVariable  pvset() throws RecognitionException, TokenStreamException {
		SetOfSchemaVariable pvs;
		
		
		ParsableVariable pv = null;
		pvs = SetAsListOfSchemaVariable.EMPTY_SET;
		
		
		pv=varId();
		{
		switch ( LA(1)) {
		case RPAREN:
		{
			break;
		}
		case COMMA:
		{
			match(COMMA);
			pvs=pvset();
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			pvs = pvs.add
			((SchemaVariable)pv);
		}
		return pvs;
	}
	
	public final void heuristic(
		TacletBuilder b
	) throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			
			Heuristic h = (Heuristic) heuristics().lookup(new Name(id.getText()));
			if (h == null) {
			throw new NotDeclException("heuristic", id,
			getFilename());
			}
			b.addHeuristic(h);
			
		}
	}
	
	public final Term  bindingterm() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		a=sv_or_local_bindingterm();
		return a;
	}
	
	public final Term  sv_or_local_bindingterm() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		Token  v0 = null;
		
		Operator op = null;
		QuantifiableVariable v = null;
		TermSymbol ts = null;
		Term a1;
		Sort s = null;
		boolean bound = false;
		
		
		switch ( LA(1)) {
		case FORALL:
		case EXISTS:
		{
			{
			{
			switch ( LA(1)) {
			case FORALL:
			{
				match(FORALL);
				if ( inputState.guessing==0 ) {
					op = Op.ALL;
				}
				break;
			}
			case EXISTS:
			{
				match(EXISTS);
				if ( inputState.guessing==0 ) {
					op = Op.EX;
				}
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			v0 = LT(1);
			match(IDENT);
			{
			switch ( LA(1)) {
			case COLON:
			{
				match(COLON);
				s=sortId();
				if ( inputState.guessing==0 ) {
					
					v = bindVar(v0,s);
					bound = true;
					
				}
				break;
			}
			case DOT:
			{
				if ( inputState.guessing==0 ) {
					
					ts = (TermSymbol) variables().lookup(new Name(v0.getText()));   
					if (   ts == null 
					|| ts instanceof LogicVariable ) {
					throw new SemanticException(
					"Quantified variables need a sort.");
					}
					if ( ! (ts instanceof SchemaVariable && 
					((SchemaVariable)ts).isVariableSV())) {
					throw new SemanticException(
					""+ts+" is not allowed in a quantifier.");
					}
					v = (SortedSchemaVariable) ts;
					bound = false;
					
				}
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			match(DOT);
			a1=term60();
			if ( inputState.guessing==0 ) {
				
				a = tf.createQuantifierTerm((Quantifier)op, v, a1);
				if (bound) {unbindVars();}
				
			}
			}
			break;
		}
		case LBRACE:
		{
			a=update_or_substitution();
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return a;
	}
	
	public final Sort  sortId() throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		s=sortId_check(true);
		return s;
	}
	
	public final Term  term60() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		Term a1,a2;
		Sort s;
		
		
		switch ( LA(1)) {
		case NOT:
		{
			match(NOT);
			a1=term60();
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.NOT,new Term[]{a1});
			}
			break;
		}
		case CTRL:
		{
			match(CTRL);
			a1=term60();
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.CTRL,new Term[]{a1});
			}
			break;
		}
		case FORALL:
		case EXISTS:
		case LBRACE:
		{
			a=bindingterm();
			break;
		}
		case TRUE:
		case FALSE:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case LPAREN:
		case AT:
		case ARITH_OP:
		case CHAR_LITERAL:
		case DIAMOND:
		case BOX:
		case THROUGHOUT:
		case PIDENT:
		case IDENT:
		{
			a=dlTermOrTerm70();
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return a;
	}
	
	public final Term  update_or_substitution() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		boolean synPredMatched85 = false;
		if (((LA(1)==LBRACE))) {
			int _m85 = mark();
			synPredMatched85 = true;
			inputState.guessing++;
			try {
				{
				match(LBRACE);
				match(IDENT);
				{
				switch ( LA(1)) {
				case COLON:
				{
					match(COLON);
					sortId();
					break;
				}
				case SLASH:
				{
					match(SLASH);
					term60();
					break;
				}
				case TRUE:
				case FALSE:
				case IN_TYPE:
				case IS_OBJECT:
				case META_IN_LONG:
				case META_IN_INT:
				case META_IN_CHAR:
				case META_IN_SHORT:
				case META_IN_BYTE:
				case META_ADD:
				case META_SUB:
				case META_MUL:
				case META_DIV:
				case META_JDIV:
				case META_MOD:
				case META_JMOD:
				case META_LESS:
				case META_GREATER:
				case META_LEQ:
				case META_GEQ:
				case META_EQ:
				case META_INT_UNARY_MINUS:
				case META_INT_ADD:
				case META_INT_SUB:
				case META_INT_MUL:
				case META_INT_DIV:
				case META_INT_MOD:
				case META_INT_AND:
				case META_INT_OR:
				case META_INT_XOR:
				case META_INT_COMPLEMENT:
				case META_INT_SHIFTRIGHT:
				case META_INT_SHIFTLEFT:
				case META_INT_UNSIGNEDSHIFTRIGHT:
				case META_LONG_UNARY_MINUS:
				case META_LONG_ADD:
				case META_LONG_SUB:
				case META_LONG_MUL:
				case META_LONG_DIV:
				case META_LONG_MOD:
				case META_LONG_AND:
				case META_LONG_OR:
				case META_LONG_XOR:
				case META_LONG_COMPLEMENT:
				case META_LONG_SHIFTRIGHT:
				case META_LONG_SHIFTLEFT:
				case META_LONG_UNSIGNEDSHIFTRIGHT:
				case META_MODULO_BYTE:
				case META_MODULO_SHORT:
				case META_MODULO_INTEGER:
				case META_MODULO_LONG:
				case STATIC_ANALYSIS:
				case RESOLVE_QUERY:
				case CONSTANT_VALUE:
				case META_LENGTH:
				case META_TRANSIENT:
				case META_SHADOW:
				case META_ALLSUBTYPES:
				case LPAREN:
				case LBRACE:
				case AT:
				case ARITH_OP:
				case CHAR_LITERAL:
				case PIDENT:
				case IDENT:
				{
					logicTermReEntry();
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				}
			}
			catch (RecognitionException pe) {
				synPredMatched85 = false;
			}
			rewind(_m85);
			inputState.guessing--;
		}
		if ( synPredMatched85 ) {
			result=substitutionterm();
		}
		else if ((LA(1)==LBRACE)) {
			result=updateterm();
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		return result;
	}
	
	public final TermWithBoundVars  argument() throws RecognitionException, TokenStreamException {
		TermWithBoundVars p = null;
		
		Token  v0 = null;
		Token  v1 = null;
		Token  v2 = null;
		
		List listVars = new LinkedList();
		SortedSchemaVariable var = null;
		ArrayOfQuantifiableVariable vars = null;
		SortedSchemaVariable[] arrayVars = null;
		Term a1;
		
		LogicVariable v = null;
		LogicVariable[] logArrayVars = null;
		Sort s = null;
		
		
		{
		switch ( LA(1)) {
		case BACKSLASH:
		{
			match(BACKSLASH);
			v0 = LT(1);
			match(IDENT);
			{
			switch ( LA(1)) {
			case COLON:
			{
				match(COLON);
				s=sortId();
				if ( inputState.guessing==0 ) {
					
					v = bindVar(v0, s);
					listVars.add(v);
					
				}
				{
				_loop74:
				do {
					if ((LA(1)==BACKSLASH)) {
						match(BACKSLASH);
						v1 = LT(1);
						match(IDENT);
						match(COLON);
						s=sortId();
						if ( inputState.guessing==0 ) {
							
							v = bindVar(v1, s);
							listVars.add(v);
							
						}
					}
					else {
						break _loop74;
					}
					
				} while (true);
				}
				match(SLASH);
				a1=term60();
				if ( inputState.guessing==0 ) {
					
					logArrayVars = (LogicVariable[])listVars.toArray(new LogicVariable[listVars.size()]);
					vars = new ArrayOfQuantifiableVariable(logArrayVars);
					p = new TermWithBoundVars(vars, a1);
					unbindVars();
					
				}
				break;
			}
			case SLASH:
			case BACKSLASH:
			{
				if ( inputState.guessing==0 ) {
					
					var = (SortedSchemaVariable)variables().lookup(new Name(v0.getText())); 
					listVars.add(var);
					
				}
				{
				_loop76:
				do {
					if ((LA(1)==BACKSLASH)) {
						match(BACKSLASH);
						v2 = LT(1);
						match(IDENT);
						if ( inputState.guessing==0 ) {
							
							var = (SortedSchemaVariable)variables().lookup(new Name(v2.getText())); 
							listVars.add(var);
							
						}
					}
					else {
						break _loop76;
					}
					
				} while (true);
				}
				match(SLASH);
				a1=term60();
				if ( inputState.guessing==0 ) {
					
					arrayVars = (SortedSchemaVariable[])listVars.toArray(new SortedSchemaVariable[listVars.size()]);
					vars = new ArrayOfQuantifiableVariable(arrayVars);
					p = new TermWithBoundVars(vars, a1);
					
				}
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			break;
		}
		case FORALL:
		case EXISTS:
		case TRUE:
		case FALSE:
		case NOT:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case LPAREN:
		case LBRACE:
		case AT:
		case CTRL:
		case ARITH_OP:
		case CHAR_LITERAL:
		case DIAMOND:
		case BOX:
		case THROUGHOUT:
		case PIDENT:
		case IDENT:
		{
			a1=term60();
			if ( inputState.guessing==0 ) {
				
				p = new TermWithBoundVars(vars, a1);
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return p;
	}
	
	public final Operator  metaId() throws RecognitionException, TokenStreamException {
		Operator v = null;
		
		
		switch ( LA(1)) {
		case META_SHADOW:
		{
			match(META_SHADOW);
			if ( inputState.guessing==0 ) {
				v = Op.META_SHADOW ;
			}
			break;
		}
		case CONSTANT_VALUE:
		{
			match(CONSTANT_VALUE);
			if ( inputState.guessing==0 ) {
				v = Op.CONSTANT_VALUE;
			}
			break;
		}
		case META_INT_UNARY_MINUS:
		{
			match(META_INT_UNARY_MINUS);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_UNARY_MINUS ;
			}
			break;
		}
		case META_LONG_UNARY_MINUS:
		{
			match(META_LONG_UNARY_MINUS);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_UNARY_MINUS ;
			}
			break;
		}
		case META_ADD:
		{
			match(META_ADD);
			if ( inputState.guessing==0 ) {
				v = Op.META_ADD;
			}
			break;
		}
		case META_SUB:
		{
			match(META_SUB);
			if ( inputState.guessing==0 ) {
				v = Op.META_SUB;
			}
			break;
		}
		case META_MUL:
		{
			match(META_MUL);
			if ( inputState.guessing==0 ) {
				v = Op.META_MUL;
			}
			break;
		}
		case META_DIV:
		{
			match(META_DIV);
			if ( inputState.guessing==0 ) {
				v = Op.META_DIV;
			}
			break;
		}
		case META_JDIV:
		{
			match(META_JDIV);
			if ( inputState.guessing==0 ) {
				v = Op.META_JDIV;
			}
			break;
		}
		case META_MOD:
		{
			match(META_MOD);
			if ( inputState.guessing==0 ) {
				v = Op.META_MOD;
			}
			break;
		}
		case META_JMOD:
		{
			match(META_JMOD);
			if ( inputState.guessing==0 ) {
				v = Op.META_JMOD;
			}
			break;
		}
		case META_LESS:
		{
			match(META_LESS);
			if ( inputState.guessing==0 ) {
				v = Op.META_LESS;
			}
			break;
		}
		case META_GREATER:
		{
			match(META_GREATER);
			if ( inputState.guessing==0 ) {
				v = Op.META_GREATER;
			}
			break;
		}
		case META_LEQ:
		{
			match(META_LEQ);
			if ( inputState.guessing==0 ) {
				v = Op.META_LEQ;
			}
			break;
		}
		case META_GEQ:
		{
			match(META_GEQ);
			if ( inputState.guessing==0 ) {
				v = Op.META_GEQ;
			}
			break;
		}
		case META_EQ:
		{
			match(META_EQ);
			if ( inputState.guessing==0 ) {
				v = Op.META_EQ;
			}
			break;
		}
		case META_INT_ADD:
		{
			match(META_INT_ADD);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_ADD;
			}
			break;
		}
		case META_INT_SUB:
		{
			match(META_INT_SUB);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_SUB;
			}
			break;
		}
		case META_INT_MUL:
		{
			match(META_INT_MUL);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_MUL;
			}
			break;
		}
		case META_INT_DIV:
		{
			match(META_INT_DIV);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_DIV;
			}
			break;
		}
		case META_INT_MOD:
		{
			match(META_INT_MOD);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_MOD;
			}
			break;
		}
		case META_INT_AND:
		{
			match(META_INT_AND);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_AND;
			}
			break;
		}
		case META_INT_OR:
		{
			match(META_INT_OR);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_OR;
			}
			break;
		}
		case META_INT_XOR:
		{
			match(META_INT_XOR);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_XOR;
			}
			break;
		}
		case META_INT_COMPLEMENT:
		{
			match(META_INT_COMPLEMENT);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_COMPLEMENT;
			}
			break;
		}
		case META_INT_SHIFTRIGHT:
		{
			match(META_INT_SHIFTRIGHT);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_SHIFTRIGHT;
			}
			break;
		}
		case META_INT_SHIFTLEFT:
		{
			match(META_INT_SHIFTLEFT);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_SHIFTLEFT;
			}
			break;
		}
		case META_INT_UNSIGNEDSHIFTRIGHT:
		{
			match(META_INT_UNSIGNEDSHIFTRIGHT);
			if ( inputState.guessing==0 ) {
				v = Op.META_INT_UNSIGNEDSHIFTRIGHT;
			}
			break;
		}
		case META_LONG_ADD:
		{
			match(META_LONG_ADD);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_ADD;
			}
			break;
		}
		case META_LONG_SUB:
		{
			match(META_LONG_SUB);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_SUB;
			}
			break;
		}
		case META_LONG_MUL:
		{
			match(META_LONG_MUL);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_MUL;
			}
			break;
		}
		case META_LONG_DIV:
		{
			match(META_LONG_DIV);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_DIV;
			}
			break;
		}
		case META_LONG_MOD:
		{
			match(META_LONG_MOD);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_MOD;
			}
			break;
		}
		case META_LONG_AND:
		{
			match(META_LONG_AND);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_AND;
			}
			break;
		}
		case META_LONG_OR:
		{
			match(META_LONG_OR);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_OR;
			}
			break;
		}
		case META_LONG_XOR:
		{
			match(META_LONG_XOR);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_XOR;
			}
			break;
		}
		case META_LONG_COMPLEMENT:
		{
			match(META_LONG_COMPLEMENT);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_COMPLEMENT;
			}
			break;
		}
		case META_LONG_SHIFTRIGHT:
		{
			match(META_LONG_SHIFTRIGHT);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_SHIFTRIGHT;
			}
			break;
		}
		case META_LONG_SHIFTLEFT:
		{
			match(META_LONG_SHIFTLEFT);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_SHIFTLEFT;
			}
			break;
		}
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		{
			match(META_LONG_UNSIGNEDSHIFTRIGHT);
			if ( inputState.guessing==0 ) {
				v = Op.META_LONG_UNSIGNEDSHIFTRIGHT;
			}
			break;
		}
		case META_IN_LONG:
		{
			match(META_IN_LONG);
			if ( inputState.guessing==0 ) {
				v = Op.META_IN_LONG;
			}
			break;
		}
		case META_IN_INT:
		{
			match(META_IN_INT);
			if ( inputState.guessing==0 ) {
				v = Op.META_IN_INT;
			}
			break;
		}
		case META_IN_CHAR:
		{
			match(META_IN_CHAR);
			if ( inputState.guessing==0 ) {
				v = Op.META_IN_CHAR;
			}
			break;
		}
		case META_IN_SHORT:
		{
			match(META_IN_SHORT);
			if ( inputState.guessing==0 ) {
				v = Op.META_IN_SHORT;
			}
			break;
		}
		case META_IN_BYTE:
		{
			match(META_IN_BYTE);
			if ( inputState.guessing==0 ) {
				v = Op.META_IN_BYTE;
			}
			break;
		}
		case META_MODULO_BYTE:
		{
			match(META_MODULO_BYTE);
			if ( inputState.guessing==0 ) {
				v = Op.META_MODULO_BYTE;
			}
			break;
		}
		case META_MODULO_SHORT:
		{
			match(META_MODULO_SHORT);
			if ( inputState.guessing==0 ) {
				v = Op.META_MODULO_SHORT;
			}
			break;
		}
		case META_MODULO_INTEGER:
		{
			match(META_MODULO_INTEGER);
			if ( inputState.guessing==0 ) {
				v = Op.META_MODULO_INTEGER;
			}
			break;
		}
		case META_MODULO_LONG:
		{
			match(META_MODULO_LONG);
			if ( inputState.guessing==0 ) {
				v = Op.META_MODULO_LONG;
			}
			break;
		}
		case STATIC_ANALYSIS:
		{
			match(STATIC_ANALYSIS);
			if ( inputState.guessing==0 ) {
				v = Op.STATIC_ANALYSIS;
			}
			break;
		}
		case RESOLVE_QUERY:
		{
			match(RESOLVE_QUERY);
			if ( inputState.guessing==0 ) {
				v = Op.RESOLVE_QUERY;
			}
			break;
		}
		case META_LENGTH:
		{
			match(META_LENGTH);
			if ( inputState.guessing==0 ) {
				v = Op.META_LENGTH;
			}
			break;
		}
		case META_TRANSIENT:
		{
			match(META_TRANSIENT);
			if ( inputState.guessing==0 ) {
				v = Op.META_TRANSIENT;
			}
			break;
		}
		case META_ALLSUBTYPES:
		{
			match(META_ALLSUBTYPES);
			if ( inputState.guessing==0 ) {
				v = Op.METAALLSUBTYPES;
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return v;
	}
	
	public final Term  metaTerm() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		LinkedList al = new LinkedList();
		Operator vf = null;
		Term t = null;
		
		
		{
		vf=metaId();
		match(LPAREN);
		t=term();
		if ( inputState.guessing==0 ) {
			
			al.add(t);
			
		}
		{
		_loop81:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				t=term();
				if ( inputState.guessing==0 ) {
					
					al.add(t);
					
				}
			}
			else {
				break _loop81;
			}
			
		} while (true);
		}
		match(RPAREN);
		if ( inputState.guessing==0 ) {
			
			result = tf.createMetaTerm((MetaOperator)vf,
			(Term[])al.toArray(AN_ARRAY_OF_TERMS));
			
		}
		}
		return result;
	}
	
	public final Term  logicTermReEntry() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		{
		switch ( LA(1)) {
		case TRUE:
		case FALSE:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case LPAREN:
		case AT:
		case ARITH_OP:
		case CHAR_LITERAL:
		case PIDENT:
		case IDENT:
		{
			result=accessterm();
			break;
		}
		case LBRACE:
		{
			result=update_or_substitution();
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			if (result.sort() == Sort.FORMULA) {
			throw 
			new KeYSemanticException
			("Only terms may stand here.", getLine(), getColumn(), 
			getFilename());
			}
			
		}
		return result;
	}
	
	public final Term  substitutionterm() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		Token  v1 = null;
		Token  v2 = null;
		Term a1, a2 = null, a3;
		Sort s = null;
		TermSymbol ts;
		QuantifiableVariable v = null, vv = null;
		boolean bound = false;
		boolean waryAll = false;
		boolean waryEx = false;
		
		
		match(LBRACE);
		v1 = LT(1);
		match(IDENT);
		{
		if ((LA(1)==COLON)) {
			match(COLON);
			s=sortId();
			if ( inputState.guessing==0 ) {
				
				bound = true;
				
			}
		}
		else if (((_tokenSet_0.member(LA(1))))&&( functions().lookup(new Name(v1.getText())) == null )) {
			if ( inputState.guessing==0 ) {
				
				ts = (TermSymbol) variables().lookup(new Name(v1.getText()));
				if (   ts == null
				|| ts instanceof LogicVariable ) {
				throw new SemanticException("Substituted variables need a sort.");
				}
				if ( ! (ts instanceof SchemaVariable &&
				((SchemaVariable)ts).isVariableSV()) ) {
				throw new SemanticException(
				"   "+ts+" is not allowed as bound variable "
				+"in a substitution term.");
				}
				v = (SortedSchemaVariable) ts;
				bound = false;
				
			}
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		}
		{
		switch ( LA(1)) {
		case TRUE:
		case FALSE:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case LPAREN:
		case LBRACE:
		case AT:
		case ARITH_OP:
		case CHAR_LITERAL:
		case PIDENT:
		case IDENT:
		{
			a1=logicTermReEntry();
			{
			switch ( LA(1)) {
			case WARY_ALL:
			{
				match(WARY_ALL);
				if ( inputState.guessing==0 ) {
					waryAll = true;
				}
				break;
			}
			case WARY_EX:
			{
				match(WARY_EX);
				if ( inputState.guessing==0 ) {
					waryEx  = true;
				}
				break;
			}
			case RBRACE:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			match(RBRACE);
			if ( inputState.guessing==0 ) {
				
				if (bound) {v = bindVar(v1,s);} 
				
			}
			a2=term60();
			if ( inputState.guessing==0 ) {
				
				if ( waryAll ) {
				result = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
				v, a1, a2 );
				} else if ( waryEx ) {
				result = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
				v, a1, a2 );
				} else {
				result = tf.createSubstitutionTerm ( Op.SUBST,
				v, a1, a2 );
				}
				if (bound) {unbindVars();}
				
			}
			break;
		}
		case SLASH:
		{
			match(SLASH);
			a1=logicTermReEntry();
			if ( inputState.guessing==0 ) {
				
				if (bound) {v = bindVar(v1,s);} 
				
			}
			{
			switch ( LA(1)) {
			case COMMA:
			{
				match(COMMA);
				v2 = LT(1);
				match(IDENT);
				{
				if ((LA(1)==COLON)) {
					match(COLON);
					s=sortId();
					if ( inputState.guessing==0 ) {
						bound = true;
					}
				}
				else if (((LA(1)==SLASH))&&( functions().lookup(new Name(v2.getText())) == null )) {
					if ( inputState.guessing==0 ) {
						
						ts = (TermSymbol) variables().lookup(new Name(v2.getText()));
						if (   ts == null
						|| ts instanceof LogicVariable ) {
						throw new SemanticException("Substituted variables need a sort.");
						}
						if ( ! (ts instanceof SchemaVariable &&
						((SchemaVariable)ts).isVariableSV()) ) {
						throw new SemanticException(
						"	"+ts+" is not allowed as bound variable "
						+"in a substitution term.");
						}
						vv = (SortedSchemaVariable) ts;
						bound = false;
						
					}
				}
				else {
					throw new NoViableAltException(LT(1), getFilename());
				}
				
				}
				match(SLASH);
				a2=logicTermReEntry();
				if ( inputState.guessing==0 ) {
					
					if (bound) {vv = bindVar(v2, s);} 
					
				}
				break;
			}
			case RBRACE:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			match(RBRACE);
			a3=term60();
			if ( inputState.guessing==0 ) {
				
				if (vv == null)
				{
				result = tf.createSubstitutionTerm (Op.SUBST, v, a1, a3);
				}
				else
				{
				result = tf.createTernarySubstTerm(Op.SUBST, v, vv, 
				new Term[]{a1, a2, a3});
				}
				
				if (bound) {unbindVars();}
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return result;
	}
	
	public final Term  updateterm() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		SingleUpdateData sud = null;
		List subs = new LinkedList();  
		Term a2 = null;
		
		
		match(LBRACE);
		sud=singleupdate();
		if ( inputState.guessing==0 ) {
			
			subs.add(sud.a0); 
			subs.add(sud.a1); 
			
		}
		{
		_loop146:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				sud=singleupdate();
				if ( inputState.guessing==0 ) {
					
					subs.add(sud.a0); 
					subs.add(sud.a1); 
					
				}
			}
			else {
				break _loop146;
			}
			
		} while (true);
		}
		match(RBRACE);
		a2=term60();
		if ( inputState.guessing==0 ) {
			
			subs.add(a2);
			
		}
		if ( inputState.guessing==0 ) {
			
			result = tf.createUpdateTerm((Term[])subs.toArray(new Term[subs.size()]));
			
		}
		return result;
	}
	
	public final Term  shortCond(
		Term cond1, Term cond2
	) throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		Term c_then = null;
		ParsableVariable v = null;
		
		
		match(DOT);
		v=varId();
		match(CONDUP);
		c_then=logicTermReEntry();
		if ( inputState.guessing==0 ) {
			
			if (v instanceof SortedSchemaVariable) {
			result = tf.createIfElseTerm((SortedSchemaVariable)v, cond1, 
			cond2, c_then);
			} else {
			result = tf.createIfElseTerm((ProgramVariable)v, cond1, 
			cond2, c_then);
			}
			
		}
		return result;
	}
	
	public final Term  specialTerm() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		Operator vf = null;
		
		
		switch ( LA(1)) {
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		{
			result=metaTerm();
			break;
		}
		case IN_TYPE:
		case IS_OBJECT:
		{
			vf=expr_op();
			if ( inputState.guessing==0 ) {
				
				result = tf.createFunctionTerm((Function)vf, AN_ARRAY_OF_TERMS);
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return result;
	}
	
	public final Operator  expr_op() throws RecognitionException, TokenStreamException {
		Operator v = null;
		
		Token  id = null;
		Token  id1 = null;
		Token  id2 = null;
		
		ParsableVariable left_var0 = null;
		ParsableVariable right_var0 = null;
		ParsableVariable unary_var0 = null;
		
		
		switch ( LA(1)) {
		case IN_TYPE:
		{
			match(IN_TYPE);
			match(LPAREN);
			{
			switch ( LA(1)) {
			case IDENT:
			{
				{
				left_var0=varId();
				{
				switch ( LA(1)) {
				case SLASH:
				case ARITH_OP:
				case IDENT:
				{
					{
					switch ( LA(1)) {
					case ARITH_OP:
					{
						id = LT(1);
						match(ARITH_OP);
						break;
					}
					case SLASH:
					{
						id1 = LT(1);
						match(SLASH);
						break;
					}
					case IDENT:
					{
						id2 = LT(1);
						match(IDENT);
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					right_var0=varId();
					break;
				}
				case RPAREN:
				{
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				}
				break;
			}
			case 188:
			{
				{
				match(188);
				unary_var0=varId();
				}
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			match(RPAREN);
			if ( inputState.guessing==0 ) {
				
				if (id==null) id=id1;   
				if (unary_var0!=null) {
				// can only be unary minus
				SortedSchemaVariable unary_var = (SortedSchemaVariable) unary_var0;
				Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
				return new InTypeOperator
				(type, new de.uka.ilkd.key.java.expression.operator.Negative(unary_var));
				}
				else {
				SortedSchemaVariable left_var = (SortedSchemaVariable) left_var0;
				Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
				if (right_var0 != null) {
				SortedSchemaVariable right_var = (SortedSchemaVariable) right_var0;
				String op_str = (id == null ? id2.getText() : id.getText());
				if ("+".equals(op_str)) {
				return new InTypeOperator
				(type, new de.uka.ilkd.key.java.expression.operator.Plus(left_var, right_var));
				} else if ("-".equals(op_str)) {
				return new InTypeOperator
				(type, new de.uka.ilkd.key.java.expression.operator.Minus(left_var, right_var));
				} else if ("*".equals(op_str)) {
				return new InTypeOperator
				(type, new de.uka.ilkd.key.java.expression.operator.Times(left_var, right_var));
				} else if ("/".equals(op_str)) {
				return new InTypeOperator
				(type, new de.uka.ilkd.key.java.expression.operator.Divide(left_var, right_var));
				} else if ("%".equals(op_str)) {
				return new InTypeOperator
				(type, new de.uka.ilkd.key.java.expression.operator.Modulo(left_var, right_var));
				} 
				}
				else {
				return new InTypeOperator(type, left_var);
				}
				}
				
			}
			break;
		}
		case IS_OBJECT:
		{
			match(IS_OBJECT);
			match(LPAREN);
			{
			left_var0=varId();
			}
			match(RPAREN);
			if ( inputState.guessing==0 ) {
				
				SortedSchemaVariable var = (SortedSchemaVariable) left_var0;
				return new IsObject(de.uka.ilkd.key.logic.sort.Sort.FORMULA, var);
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return v;
	}
	
	public final Term  formula() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		a=term();
		if ( inputState.guessing==0 ) {
			
			if ( a.sort() != Sort.FORMULA ) {
			throw new antlr.SemanticException
			("Just Parsed a Term where a Formula was expected");
			}
			
		}
		return a;
	}
	
	public final Term  term20() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		Term a1;
		
		
		a=term30();
		{
		switch ( LA(1)) {
		case EQV:
		{
			match(EQV);
			a1=term20();
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.EQV, new Term[]{a, a1});
			}
			break;
		}
		case EOF:
		case COMMA:
		case RPAREN:
		case SEQARROW:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return a;
	}
	
	public final Term  term30() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		Term a1;
		
		
		a=term40();
		{
		switch ( LA(1)) {
		case IMP:
		{
			match(IMP);
			a1=term30();
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.IMP, new Term[]{a, a1});
			}
			break;
		}
		case EOF:
		case COMMA:
		case RPAREN:
		case EQV:
		case SEQARROW:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return a;
	}
	
	public final Term  term40() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		Term a1;
		
		
		a=term50();
		{
		switch ( LA(1)) {
		case OR:
		{
			match(OR);
			a1=term40();
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.OR, new Term[]{a, a1});
			}
			break;
		}
		case EOF:
		case COMMA:
		case RPAREN:
		case IMP:
		case EQV:
		case SEQARROW:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return a;
	}
	
	public final Term  term50() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		Term a1;
		
		
		a=term60();
		{
		switch ( LA(1)) {
		case AND:
		{
			match(AND);
			a1=term50();
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.AND, new Term[]{a, a1});
			}
			break;
		}
		case EOF:
		case COMMA:
		case RPAREN:
		case OR:
		case IMP:
		case EQV:
		case SEQARROW:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return a;
	}
	
	public final Term  dlTermOrTerm70() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		if (((LA(1) >= DIAMOND && LA(1) <= THROUGHOUT))) {
			a=normal_dl_term();
		}
		else {
			boolean synPredMatched107 = false;
			if (((LA(1)==IDENT))) {
				int _m107 = mark();
				synPredMatched107 = true;
				inputState.guessing++;
				try {
					{
					match(IDENT);
					match(MODALITYARG);
					}
				}
				catch (RecognitionException pe) {
					synPredMatched107 = false;
				}
				rewind(_m107);
				inputState.guessing--;
			}
			if ( synPredMatched107 ) {
				a=named_dl_term();
			}
			else if ((_tokenSet_1.member(LA(1)))) {
				a=term70();
			}
			else {
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			return a;
		}
		
	public final Term  normal_dl_term() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		Token  d1 = null;
		Token  d2 = null;
		Token  d3 = null;
		JavaBlock jb=null;
		Term a1;
		
		
		switch ( LA(1)) {
		case DIAMOND:
		{
			d1 = LT(1);
			match(DIAMOND);
			if ( inputState.guessing==0 ) {
				
				jb=getJavaBlock(d1);
				HashSet progVars=progVars(jb);
				bindProgVars(progVars);
				
			}
			a1=term60();
			if ( inputState.guessing==0 ) {
				
				a = tf.createDiamondTerm(jb, a1);
				unbindProgVars();
				
			}
			break;
		}
		case BOX:
		{
			d2 = LT(1);
			match(BOX);
			if ( inputState.guessing==0 ) {
				
				jb=getJavaBlock(d2);
				HashSet progVars=progVars(jb);
				bindProgVars(progVars);
				
			}
			a1=term60();
			if ( inputState.guessing==0 ) {
				
				a = tf.createBoxTerm(jb, a1);
				unbindProgVars();
				
			}
			break;
		}
		case THROUGHOUT:
		{
			d3 = LT(1);
			match(THROUGHOUT);
			if ( inputState.guessing==0 ) {
				
				jb=getJavaBlock(d3);
				HashSet progVars=progVars(jb);
				bindProgVars(progVars);
				
			}
			a1=term60();
			if ( inputState.guessing==0 ) {
				
				a = tf.createProgramTerm(Op.TOUT,jb, a1);
				unbindProgVars();
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return a;
	}
	
	public final Term  named_dl_term() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		Token  id = null;
		Token  bl = null;
		
		TermWithBoundVars[] argsWithBoundVars = new TermWithBoundVars[0];
		Term[] terms = null;
		Term term = null;
		JavaBlock jb = null;
		
		
		id = LT(1);
		match(IDENT);
		bl = LT(1);
		match(MODALITYARG);
		{
		if ( inputState.guessing==0 ) {
			
			jb = getJavaBlock(bl);               
			bindProgVars(progVars(jb));
			
		}
		argsWithBoundVars=argument_list();
		if ( inputState.guessing==0 ) {
			
			unbindProgVars();
			terms = new Term[argsWithBoundVars.length];
			for (int i=0; i<argsWithBoundVars.length; i++) {
			terms[i] = argsWithBoundVars[i].term;
			}
			Operator op = null;
			if (id.getText().charAt(0)=='#') {
			if (!inSchemaMode()) {
			throw new KeYSemanticException
			("No schema elements allowed outside taclet declarations.",
			id, getFilename());
			}
			op = (SchemaVariable)variables().lookup(new Name(id.getText()));
			} else {
			op = Op.getModality(id.getText());
			}
			if(op == null) {
			throw new KeYSemanticException("Unknown modal operator.", id, getFilename());
			}
			if(op.arity() != terms.length) {
			throw new KeYSemanticException
			("The arity of " + op.name() +
			" is "+op.arity() + ", but you provided " + terms.length + " arguments",
			getLine(), getColumn(), getFilename());
			}
			a = tf.createProgramTerm(op, jb, terms);
			
		}
		}
		return a;
	}
	
	public final Term  term70() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		Term a1;
		
		
		a=accessterm();
		{
		boolean synPredMatched111 = false;
		if (((LA(1)==EQUALS))) {
			int _m111 = mark();
			synPredMatched111 = true;
			inputState.guessing++;
			try {
				{
				match(EQUALS);
				}
			}
			catch (RecognitionException pe) {
				synPredMatched111 = false;
			}
			rewind(_m111);
			inputState.guessing--;
		}
		if ( synPredMatched111 ) {
			match(EQUALS);
			a1=logicTermReEntry();
			if ( inputState.guessing==0 ) {
				a = tf.createEqualityTerm(a, a1);
			}
		}
		else if ((_tokenSet_2.member(LA(1)))) {
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		}
		return a;
	}
	
	public final Term  accessterm() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		Token  neg = null;
		Token  rb = null;
		
		KeYJavaType kjt = null;
		Sort arraySort = null;
		Term indexTerm = null;
		int shadowPrimes = 0;
		String className = null;
		
		
		boolean synPredMatched129 = false;
		if (((LA(1)==LPAREN))) {
			int _m129 = mark();
			synPredMatched129 = true;
			inputState.guessing++;
			try {
				{
				match(LPAREN);
				logicTermReEntry();
				{
				switch ( LA(1)) {
				case COMMA:
				{
					match(COMMA);
					break;
				}
				case GUARD:
				{
					match(GUARD);
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				}
			}
			catch (RecognitionException pe) {
				synPredMatched129 = false;
			}
			rewind(_m129);
			inputState.guessing--;
		}
		if ( synPredMatched129 ) {
			result=conditional();
		}
		else if (((LA(1)==ARITH_OP))&&("-".equals(LT(1).getText()) && 
	!TermParser.isNumber(LT(2).getText()))) {
			neg = LT(1);
			match(ARITH_OP);
			result=logicTermReEntry();
			if ( inputState.guessing==0 ) {
				
				if (result.sort() != Sort.FORMULA) {
				result = tf.createFunctionTerm
				((Function) functions().lookup(new Name("~m")), result);
				} else {
				throw new KeYSemanticException
				("Formula cannot be prefixed with: ", neg, getFilename());
				}
				
			}
		}
		else if ((_tokenSet_1.member(LA(1)))) {
			{
			if (((LA(1)==IDENT))&&(LA(2) == DOT && ( isPackage ( LT(1).getText() ) || 
                ( kjt = getTypeByClassName(LT(1).getText()))!=null ))) {
				className=classReference();
				result=static_attribute_suffix(className);
			}
			else if ((_tokenSet_1.member(LA(1)))) {
				result=term80();
			}
			else {
				throw new NoViableAltException(LT(1), getFilename());
			}
			
			}
			{
			_loop133:
			do {
				switch ( LA(1)) {
				case LBRACKET:
				{
					match(LBRACKET);
					indexTerm=logicTermReEntry();
					{
					switch ( LA(1)) {
					case RBRACKET:
					{
						match(RBRACKET);
						break;
					}
					case RBRACKETPRIMES:
					{
						rb = LT(1);
						match(RBRACKETPRIMES);
						if ( inputState.guessing==0 ) {
							
							shadowPrimes = rb.getText().length()-1;                 
							
						}
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						
						if (arraySort != null && 
						arraySort != ProgramSVSort.VARIABLE) {
						if (!(arraySort instanceof ArraySort)){
						throw new NotDeclException
						("array "+result+" declared with a different"+
						" dimension than required here", ""+result, 
						getFilename(), getLine(), getColumn());
						} else {
						arraySort = ((ArraySort) arraySort).elementSort();
						} 
						} else if (arraySort == ProgramSVSort.VARIABLE && !inSchemaMode()) {
						throw new antlr.SemanticException
						("Schemavariable are not allowed in this context.");
						}
						if (shadowPrimes>0) {
						result = tf.createArrayTerm
						(Op.SHADOW_ARRAY, result, indexTerm, 
						TermParser.toZNotation("" + shadowPrimes, functions(), tf));
						} else {
						result = tf.createArrayTerm(Op.ARRAY, result, indexTerm);
						}   
						
					}
					break;
				}
				case DOT:
				{
					result=attribute_or_query_suffix(result);
					break;
				}
				default:
				{
					break _loop133;
				}
				}
			} while (true);
			}
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		return result;
	}
	
	public final String  classReference() throws RecognitionException, TokenStreamException {
		String classReference = "";
		
		Token  id = null;
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			
			classReference = id.getText(); 
			while (isPackage(classReference)) {
			match(DOT);
			classReference += "." + LT(1).getText();
			match(IDENT);
			}      
			
			if (getJavaInfo().getTypeByClassName(classReference) == null) {
			throw new NotDeclException
			("class " + classReference + " is unknown.", 
			classReference, getFilename(), getLine(), 
			getColumn());
			}
			
		}
		return classReference;
	}
	
	public final Term  static_attribute_suffix(
		String className
	) throws RecognitionException, TokenStreamException {
		Term result = null;
		
		Token  ps = null;
		
		TermSymbol v = null;
		int shadowPrimes = 0;    
		String attributeName = "";
		
		
		match(DOT);
		attributeName=attrid();
		if ( inputState.guessing==0 ) {
			v = getAttribute(getTypeByClassName(className).
			getSort(), attributeName);
		}
		{
		{
		switch ( LA(1)) {
		case PRIMES:
		{
			ps = LT(1);
			match(PRIMES);
			if ( inputState.guessing==0 ) {
				shadowPrimes = ps.getText().length();
			}
			break;
		}
		case EOF:
		case WARY_ALL:
		case WARY_EX:
		case COLON:
		case GUARD:
		case ASSIGN:
		case DOT:
		case COMMA:
		case RPAREN:
		case RBRACE:
		case LBRACKET:
		case RBRACKET:
		case OR:
		case AND:
		case IMP:
		case EQV:
		case EQUALS:
		case RBRACKETPRIMES:
		case SEQARROW:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			result = createAttributeTerm(null, v, shadowPrimes);
			
		}
		}
		return result;
	}
	
	public final String  attrid() throws RecognitionException, TokenStreamException {
		String attr = "";;
		
		Token  attrID = null;
		
		{
		if (((LA(1)==IDENT))&&( isPackage(LT(1).getText()) || (LA(2) == COLON && LA(3) == COLON) )) {
			attr=classReference();
			match(COLON);
			match(COLON);
			if ( inputState.guessing==0 ) {
				attr += "::";
			}
		}
		else if ((LA(1)==IDENT)) {
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		}
		attrID = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			attr += attrID.getText();
		}
		return attr;
	}
	
	public final Term  attribute_or_query_suffix(
		Term prefix
	) throws RecognitionException, TokenStreamException {
		Term result = null;
		
		Token  ps = null;
		
		TermSymbol v = null;
		int shadowPrimes = 0;
		result = prefix;
		String attributeName = "";    
		
		
		match(DOT);
		{
		boolean synPredMatched121 = false;
		if (((LA(1)==IDENT))) {
			int _m121 = mark();
			synPredMatched121 = true;
			inputState.guessing++;
			try {
				{
				id();
				match(LPAREN);
				}
			}
			catch (RecognitionException pe) {
				synPredMatched121 = false;
			}
			rewind(_m121);
			inputState.guessing--;
		}
		if ( synPredMatched121 ) {
			{
			result=query(prefix);
			}
		}
		else if ((LA(1)==IDENT)) {
			attributeName=attrid();
			if ( inputState.guessing==0 ) {
				v = getAttribute(prefix.sort(), attributeName);
			}
			{
			{
			switch ( LA(1)) {
			case PRIMES:
			{
				ps = LT(1);
				match(PRIMES);
				if ( inputState.guessing==0 ) {
					shadowPrimes = ps.getText().length();
				}
				break;
			}
			case EOF:
			case WARY_ALL:
			case WARY_EX:
			case COLON:
			case GUARD:
			case ASSIGN:
			case DOT:
			case COMMA:
			case RPAREN:
			case RBRACE:
			case LBRACKET:
			case RBRACKET:
			case OR:
			case AND:
			case IMP:
			case EQV:
			case EQUALS:
			case RBRACKETPRIMES:
			case SEQARROW:
			{
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			if ( inputState.guessing==0 ) {
				
				result = createAttributeTerm(prefix, v, shadowPrimes);
				
			}
			}
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		}
		return result;
	}
	
	public final String  id() throws RecognitionException, TokenStreamException {
		String name = null;
		
		Token  prefix = null;
		Token  pprefix = null;
		Token  nameID = null;
		
		{
		switch ( LA(1)) {
		case IDENT:
		{
			prefix = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				name = prefix.getText();
			}
			break;
		}
		case PIDENT:
		{
			pprefix = LT(1);
			match(PIDENT);
			if ( inputState.guessing==0 ) {
				name = pprefix.getText();
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		{
		boolean synPredMatched281 = false;
		if (((LA(1)==COLON))) {
			int _m281 = mark();
			synPredMatched281 = true;
			inputState.guessing++;
			try {
				{
				match(COLON);
				match(COLON);
				}
			}
			catch (RecognitionException pe) {
				synPredMatched281 = false;
			}
			rewind(_m281);
			inputState.guessing--;
		}
		if ( synPredMatched281 ) {
			match(COLON);
			match(COLON);
			nameID = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				
				name += "::" + nameID.getText();
				
			}
		}
		else if ((_tokenSet_3.member(LA(1)))) {
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		}
		return name;
	}
	
	public final Term  query(
		Term prefix
	) throws RecognitionException, TokenStreamException {
		Term result = null;
		
		Token  mid = null;
		
		String classRef = "";
		Term[] args;
		TermWithBoundVars[] argsWithBoundVars = new TermWithBoundVars[0];
		
		
		classRef=classReference();
		match(COLON);
		match(COLON);
		mid = LT(1);
		match(IDENT);
		argsWithBoundVars=argument_list();
		if ( inputState.guessing==0 ) {
			
			args = new Term[argsWithBoundVars.length];
			for (int i=0; i<argsWithBoundVars.length; i++) {
			args[i] = argsWithBoundVars[i].term;
			}
			try {
			result = getServices().getJavaInfo().getProgramMethodTerm
			(prefix, mid.getText(), args, classRef);
			} catch (IllegalArgumentException e) {
			throw new antlr.SemanticException(e.getMessage());
			}
			
		}
		return result;
	}
	
	public final TermWithBoundVars[]  argument_list() throws RecognitionException, TokenStreamException {
		TermWithBoundVars[] ts = null;
		
		
		List args = new LinkedList();
		TermWithBoundVars p1, p2;
		
		
		match(LPAREN);
		{
		switch ( LA(1)) {
		case FORALL:
		case EXISTS:
		case TRUE:
		case FALSE:
		case NOT:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case BACKSLASH:
		case LPAREN:
		case LBRACE:
		case AT:
		case CTRL:
		case ARITH_OP:
		case CHAR_LITERAL:
		case DIAMOND:
		case BOX:
		case THROUGHOUT:
		case PIDENT:
		case IDENT:
		{
			p1=argument();
			if ( inputState.guessing==0 ) {
				args.add(p1);
			}
			{
			_loop158:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					p2=argument();
					if ( inputState.guessing==0 ) {
						args.add(p2);
					}
				}
				else {
					break _loop158;
				}
				
			} while (true);
			}
			break;
		}
		case RPAREN:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		match(RPAREN);
		if ( inputState.guessing==0 ) {
			
			ts = (TermWithBoundVars[]) args.toArray(new TermWithBoundVars[args.size()]);
			
		}
		return ts;
	}
	
	public final Term  conditional() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		Term comparator = null;
		
		
		match(LPAREN);
		comparator=logicTermReEntry();
		{
		switch ( LA(1)) {
		case GUARD:
		{
			result=normalConditional(comparator);
			break;
		}
		case COMMA:
		{
			result=conjunctionConditional(comparator);
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return result;
	}
	
	public final Term  term80() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		switch ( LA(1)) {
		case AT:
		case ARITH_OP:
		case CHAR_LITERAL:
		case PIDENT:
		case IDENT:
		{
			a=funcpredvarterm();
			break;
		}
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		{
			a=specialTerm();
			break;
		}
		case LPAREN:
		{
			match(LPAREN);
			a=term();
			match(RPAREN);
			break;
		}
		case TRUE:
		{
			match(TRUE);
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.TRUE);
			}
			break;
		}
		case FALSE:
		{
			match(FALSE);
			if ( inputState.guessing==0 ) {
				a = tf.createJunctorTerm(Op.FALSE);
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return a;
	}
	
	public final Term  funcpredvarterm() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		Token  ch = null;
		Token  number = null;
		
		Term[] arguments = null;
		TermWithBoundVars[] argsWithBoundVars = null;
		String varfuncid;
		String neg = "";
		
		
		switch ( LA(1)) {
		case CHAR_LITERAL:
		{
			ch = LT(1);
			match(CHAR_LITERAL);
			if ( inputState.guessing==0 ) {
				
				String s = ch.getText();
				int intVal = 0;
				if (s.length()==3) {
				intVal = (int)s.charAt(1);
				} else {
				try {
				intVal = Integer.parseInt(s.substring(3,s.length()-1),16);
				} catch (NumberFormatException ex) {
				throw new KeYSemanticException
				("Expected '"+s+"' to be a character", ch, getFilename());
				}       
				}
				a = tf.createFunctionTerm((Function) functions().lookup(new Name("C")), 
				TermParser.toZNotation(""+intVal, functions(), tf).sub(0));
				
			}
			break;
		}
		case AT:
		{
			match(AT);
			a=abbreviation();
			break;
		}
		default:
			if (((LA(1)==ARITH_OP||LA(1)==IDENT))&&( ("-".equals(LT(1).getText()) && TermParser.isNumber(LT(2).getText())) 
	|| 
        (TermParser.isNumber(LT(1).getText()) && !"(".equals(LT(2).getText())))) {
				{
				if (((LA(1)==ARITH_OP))&&("-".equals(LT(1).getText()))) {
					match(ARITH_OP);
					if ( inputState.guessing==0 ) {
						neg = "-";
					}
				}
				else if ((LA(1)==IDENT)) {
				}
				else {
					throw new NoViableAltException(LT(1), getFilename());
				}
				
				}
				number = LT(1);
				match(IDENT);
				if ( inputState.guessing==0 ) {
					a = TermParser.toZNotation(neg+number.getText(), functions(), tf);
				}
			}
			else if ((LA(1)==PIDENT||LA(1)==IDENT)) {
				varfuncid=id();
				{
				switch ( LA(1)) {
				case LPAREN:
				{
					argsWithBoundVars=argument_list();
					break;
				}
				case EOF:
				case WARY_ALL:
				case WARY_EX:
				case COLON:
				case GUARD:
				case ASSIGN:
				case DOT:
				case COMMA:
				case RPAREN:
				case RBRACE:
				case LBRACKET:
				case RBRACKET:
				case OR:
				case AND:
				case IMP:
				case EQV:
				case EQUALS:
				case RBRACKETPRIMES:
				case SEQARROW:
				{
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				if ( inputState.guessing==0 ) {
					
					ArrayOfQuantifiableVariable vars = null;
					if (argsWithBoundVars != null) {
					arguments = new Term[argsWithBoundVars.length];           
					for (int i=0; i<argsWithBoundVars.length; i++) {
					arguments[i] = argsWithBoundVars[i].term;
					
					//For OCL simplification
					if (argsWithBoundVars[i].boundVars != null) {
					if ((i == 1 && argsWithBoundVars[i].boundVars.size() == 1)
					|| (i == 2 && argsWithBoundVars[i].boundVars.size() == 2)) {
					vars = argsWithBoundVars[i].boundVars;
					} else {
					throw new antlr.SemanticException("Parsing error: The " + varfuncid + 
					" expression does not match any OCL collection operation!");
					}
					}
					//
					}
					}
					Operator op = lookupVarfuncId(varfuncid, arguments);
					if (op instanceof ParsableVariable) {
					a = termForParsedVariable((ParsableVariable)op);
					} else  if (op instanceof TermSymbol) {
					if (arguments==null) {
					arguments = AN_ARRAY_OF_TERMS;
					}
					if (vars == null) {
					a = tf.createFunctionTerm((TermSymbol)op, arguments);
					}
					
					//For OCL simplification
					else {
					if (vars.size() == 2) {
					a = tf.createTernaryBindingTerm(
					(TermSymbol)op, 
					vars.getQuantifiableVariable(0), 
					vars.getQuantifiableVariable(1), 
					arguments);
					} else {
					a = tf.createBinaryBindingTerm(
					(TermSymbol)op, 
					vars.getQuantifiableVariable(0),  
					arguments);
					}
					} 
					//
					}
					
				}
			}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return a;
	}
	
	public final Term  abbreviation() throws RecognitionException, TokenStreamException {
		Term a=null;
		
		
		String sc = null;
		
		
		{
		sc=id();
		if ( inputState.guessing==0 ) {
			
			a =  scm.getTerm(sc);
			if(a==null){
			throw new NotDeclException
			("abbreviation", sc, 
			getFilename(), getLine(), getColumn());
			}                                
			
		}
		}
		return a;
	}
	
	public final Term  normalConditional(
		Term cond1
	) throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		Term cond2 = null;
		Term c_then = null;
		Term c_else = null;
		
		
		match(GUARD);
		cond2=logicTermReEntry();
		match(RPAREN);
		{
		switch ( LA(1)) {
		case DOT:
		{
			match(DOT);
			result=shortCond(cond1, cond2);
			break;
		}
		case TRUE:
		case FALSE:
		case IN_TYPE:
		case IS_OBJECT:
		case META_IN_LONG:
		case META_IN_INT:
		case META_IN_CHAR:
		case META_IN_SHORT:
		case META_IN_BYTE:
		case META_ADD:
		case META_SUB:
		case META_MUL:
		case META_DIV:
		case META_JDIV:
		case META_MOD:
		case META_JMOD:
		case META_LESS:
		case META_GREATER:
		case META_LEQ:
		case META_GEQ:
		case META_EQ:
		case META_INT_UNARY_MINUS:
		case META_INT_ADD:
		case META_INT_SUB:
		case META_INT_MUL:
		case META_INT_DIV:
		case META_INT_MOD:
		case META_INT_AND:
		case META_INT_OR:
		case META_INT_XOR:
		case META_INT_COMPLEMENT:
		case META_INT_SHIFTRIGHT:
		case META_INT_SHIFTLEFT:
		case META_INT_UNSIGNEDSHIFTRIGHT:
		case META_LONG_UNARY_MINUS:
		case META_LONG_ADD:
		case META_LONG_SUB:
		case META_LONG_MUL:
		case META_LONG_DIV:
		case META_LONG_MOD:
		case META_LONG_AND:
		case META_LONG_OR:
		case META_LONG_XOR:
		case META_LONG_COMPLEMENT:
		case META_LONG_SHIFTRIGHT:
		case META_LONG_SHIFTLEFT:
		case META_LONG_UNSIGNEDSHIFTRIGHT:
		case META_MODULO_BYTE:
		case META_MODULO_SHORT:
		case META_MODULO_INTEGER:
		case META_MODULO_LONG:
		case STATIC_ANALYSIS:
		case RESOLVE_QUERY:
		case CONSTANT_VALUE:
		case META_LENGTH:
		case META_TRANSIENT:
		case META_SHADOW:
		case META_ALLSUBTYPES:
		case LPAREN:
		case LBRACE:
		case AT:
		case ARITH_OP:
		case CHAR_LITERAL:
		case PIDENT:
		case IDENT:
		{
			c_then=logicTermReEntry();
			match(COLON);
			c_else=logicTermReEntry();
			if ( inputState.guessing==0 ) {
				
				result = tf.createIfElseTerm(cond1, cond2, c_then, c_else);
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		return result;
	}
	
	public final Term  conjunctionConditional(
		Term cond1_1
	) throws RecognitionException, TokenStreamException {
		Term a = null;
		
		
		Term cond1_2 = null;
		Term cond2_1 = null;
		Term cond2_2 = null;
		Term c_then = null;
		Term c_else = null;
		
		
		match(COMMA);
		cond1_2=logicTermReEntry();
		match(COLON);
		c_else=logicTermReEntry();
		match(GUARD);
		cond2_1=logicTermReEntry();
		match(COMMA);
		cond2_2=logicTermReEntry();
		match(COLON);
		c_then=logicTermReEntry();
		match(RPAREN);
		if ( inputState.guessing==0 ) {
			
			a = tf.createConjCondTerm(new Term[]{cond1_1, cond1_2, c_else, 
			cond2_1, cond2_2, c_then});
			
		}
		return a;
	}
	
	public final Term  local_bindingterm() throws RecognitionException, TokenStreamException {
		Term a = null;
		
		Token  v0 = null;
		
		Operator op = null;
		LogicVariable v = null;
		Term a1;
		Sort s;
		
		
		switch ( LA(1)) {
		case FORALL:
		case EXISTS:
		{
			{
			switch ( LA(1)) {
			case FORALL:
			{
				match(FORALL);
				if ( inputState.guessing==0 ) {
					op = Op.ALL;
				}
				break;
			}
			case EXISTS:
			{
				match(EXISTS);
				if ( inputState.guessing==0 ) {
					op = Op.EX;
				}
				break;
			}
			default:
			{
				throw new NoViableAltException(LT(1), getFilename());
			}
			}
			}
			v0 = LT(1);
			match(IDENT);
			match(COLON);
			s=sortId();
			match(DOT);
			if ( inputState.guessing==0 ) {
				v = bindVar(v0,s);
			}
			a1=term60();
			if ( inputState.guessing==0 ) {
				
				a = tf.createQuantifierTerm((Quantifier)op, v,a1);
				unbindVars();
				
			}
			break;
		}
		case LBRACE:
		{
			a=update_or_substitution();
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		return a;
	}
	
	public final SingleUpdateData  singleupdate() throws RecognitionException, TokenStreamException {
		SingleUpdateData sud=null;
		
		
		Term a0 = null;
		Term a1 = null;
		ArraySort as;
		TermSymbol pv;
		LhsUpdateData lud = null;
		boolean shadow_l = false;
		boolean shadow_r = false;
		
		
		{
		a0=lhsSingle();
		}
		match(ASSIGN);
		{
		a1=rhsSingle();
		}
		if ( inputState.guessing==0 ) {
			
			sud = new SingleUpdateData();
			sud.a0 = a0;
			sud.a1 = a1;
			
		}
		return sud;
	}
	
	public final Term  lhsSingle() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		Term a1 = null;
		
		
		result=logicTermReEntry();
		if ( inputState.guessing==0 ) {
			
			if (!(result.sort() instanceof ProgramSVSort) &&
			result.op() != Op.META_SHADOW &&
			!(result.op() instanceof ProgramVariable || 
			result.op() instanceof AccessOp)) {
			throw new KeYSemanticException
			("Term of sort " + result.sort() + 
			" cannot stand for a location.",
			getLine(), getColumn(), getFilename());
			}
			
		}
		return result;
	}
	
	public final Term  rhsSingle() throws RecognitionException, TokenStreamException {
		Term result = null;
		
		
		result=logicTermReEntry();
		return result;
	}
	
	public final void decls() throws RecognitionException, TokenStreamException {
		
		
		{
		_loop179:
		do {
			if ((LA(1)==INCLUDE)) {
				match(INCLUDE);
				{
				switch ( LA(1)) {
				case LDTS:
				{
					match(LDTS);
					match(COLON);
					match(LBRACE);
					include(true);
					{
					_loop178:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							include(true);
						}
						else {
							break _loop178;
						}
						
					} while (true);
					}
					match(RBRACE);
					break;
				}
				case STRING_LITERAL:
				case IDENT:
				{
					include(false);
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				match(SEMI);
			}
			else {
				break _loop179;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			
			if(parse_includes) return;
			activatedChoices = SetAsListOfChoice.EMPTY_SET;  
			
		}
		{
		switch ( LA(1)) {
		case WITH:
		{
			match(WITH);
			match(LPAREN);
			activated_choice();
			{
			_loop182:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					activated_choice();
				}
				else {
					break _loop182;
				}
				
			} while (true);
			}
			match(RPAREN);
			break;
		}
		case EOF:
		case PROGRAM:
		case HEURISTICS:
		case SORTS:
		case OPTIONS:
		case SCHEMA:
		case PREDICATES:
		case FUNCTIONS:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			if(onlyWith) return;
		}
		{
		_loop184:
		do {
			switch ( LA(1)) {
			case OPTIONS:
			{
				option_decls();
				break;
			}
			case SORTS:
			{
				sort_decls();
				break;
			}
			case PROGRAM:
			{
				prog_var_decls();
				break;
			}
			case SCHEMA:
			{
				schema_var_decls();
				break;
			}
			case PREDICATES:
			{
				pred_decls();
				break;
			}
			case FUNCTIONS:
			{
				func_decls();
				break;
			}
			case HEURISTICS:
			{
				heur_decls();
				break;
			}
			default:
			{
				break _loop184;
			}
			}
		} while (true);
		}
	}
	
	public final void include(
		boolean ldt
	) throws RecognitionException, TokenStreamException {
		
		Token  absfile = null;
		Token  relfile = null;
		
		{
		switch ( LA(1)) {
		case IDENT:
		{
			absfile = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				
				if(parse_includes){
				addInclude(absfile.getText(),false,ldt);
				}
				
			}
			break;
		}
		case STRING_LITERAL:
		{
			relfile = LT(1);
			match(STRING_LITERAL);
			if ( inputState.guessing==0 ) {
				
				if(parse_includes){
				addInclude(relfile.getText(),true,ldt);
				}
				
			}
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
	}
	
	public final void activated_choice() throws RecognitionException, TokenStreamException {
		
		Token  cat = null;
		Token  choice = null;
		
		String name;
		Choice c;
		
		
		cat = LT(1);
		match(IDENT);
		match(COLON);
		choice = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			if(usedChoiceCategories.contains(cat.getText())){
			throw new IllegalArgumentException("You have already chosen a different option for "+cat.getText());
			}
			usedChoiceCategories.add(cat.getText());
			name = cat.getText()+":"+choice.getText();
			c = (Choice) choices().lookup(new Name(name));
			if(c==null){
			throw new NotDeclException("Option", choice,
			getFilename());
			}else{
			activatedChoices=activatedChoices.add(c);
			namespaces().setFunctions(functions().extended(c.funcNS().allElements()));
			}
			
		}
	}
	
	public final void option_decls() throws RecognitionException, TokenStreamException {
		
		
		
		
		match(OPTIONS);
		match(LBRACE);
		choice();
		{
		_loop190:
		do {
			if ((LA(1)==SEMI)) {
				match(SEMI);
				choice();
			}
			else {
				break _loop190;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void sort_decls() throws RecognitionException, TokenStreamException {
		
		
		boolean isObjectSort=false;
		boolean isGenericSort=false;
		
		//For OCL simplification
		boolean isOclSort = false;
		
		Sort[] sortExt=new Sort [0];
		Sort[] sortOneOf=new Sort [0];
		ListOfName sortIds;
		
		
		match(SORTS);
		match(LBRACE);
		{
		_loop201:
		do {
			if ((_tokenSet_4.member(LA(1)))) {
				if ( inputState.guessing==0 ) {
					
					isObjectSort=false; 
					isGenericSort=false; 
					isOclSort = false;
					sortExt = new Sort[0];
					sortOneOf = new Sort[0];
					
				}
				{
				switch ( LA(1)) {
				case OBJECT:
				{
					match(OBJECT);
					if ( inputState.guessing==0 ) {
						isObjectSort=true;
					}
					sortIds=sortIdentifiers();
					break;
				}
				case GENERIC:
				{
					match(GENERIC);
					if ( inputState.guessing==0 ) {
						isGenericSort=true;
					}
					sortIds=sortIdentifiers();
					{
					switch ( LA(1)) {
					case ONEOF:
					{
						match(ONEOF);
						sortOneOf=oneof_sorts();
						break;
					}
					case EXTENDS:
					case SEMI:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					{
					switch ( LA(1)) {
					case EXTENDS:
					{
						match(EXTENDS);
						sortExt=extends_sorts();
						break;
					}
					case SEMI:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					break;
				}
				case OCL:
				{
					match(OCL);
					if ( inputState.guessing==0 ) {
						isOclSort=true;
					}
					sortIds=sortIdentifiers();
					break;
				}
				case IDENT:
				{
					sortIds=sortIdentifiers();
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				match(SEMI);
				if ( inputState.guessing==0 ) {
					
					if (!skip_sorts) {
					Name sort_name;
					IteratorOfName it = sortIds.iterator ();
					while ( it.hasNext () ) {
					sort_name = it.next ();
					
					if (sorts().lookup(sort_name) == null) {
					Sort s;
					if (isObjectSort) {
					if ("Object".equals(sort_name)) {
					if (sorts().lookup(new Name("Object")) == null) {
					s = new ClassInstanceSortImpl(sort_name);
					}
					s=(Sort)sorts().lookup(new Name("Object"));
					} else {
					s = new ClassInstanceSortImpl(sort_name, 
					(Sort)sorts().lookup
					(new Name("Object")));
					}	
					}                            
					//Just constant for now
					else if (isOclSort) {
					s = Sort.OCL_ANY;
					}
					//
					else if (isGenericSort) {
					int        i;
					SetOfSort  ext   = SetAsListOfSort.EMPTY_SET;
					SetOfSort  oneOf = SetAsListOfSort.EMPTY_SET;
					
					for ( i = 0; i != sortExt.length; ++i )
					ext = ext.add ( sortExt[i] );
					
					for ( i = 0; i != sortOneOf.length; ++i )
					oneOf = oneOf.add ( sortOneOf[i] );
					
					try {
					s = new GenericSort ( sort_name,
					ext,
					oneOf );
					} catch ( GenericSupersortException e ) {
					throw new GenericSortException ( "sort",
					"Illegal sort given", e.getIllegalSort (),
					getFilename (), getLine (), getColumn () );
					}
					} else {
					s = new PrimitiveSort(sort_name);
					}
					sorts().add(s);
					if (s instanceof NonCollectionSort) {
					NonCollectionSort ns
					= (NonCollectionSort)s;
					sorts().add(ns.getSetSort());
					sorts().add(ns.getSequenceSort());
					sorts().add(ns.getBagSort());
					}
					if ( s instanceof SortDefiningSymbols )
					((SortDefiningSymbols)s).createSymbols ( defaultChoice.funcNS(),
					sorts() );
					}
					
					}
					}
					
				}
			}
			else {
				break _loop201;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void prog_var_decls() throws RecognitionException, TokenStreamException {
		
		Token  v0 = null;
		Token  v1 = null;
		Token  pid = null;
		
		String var_name;
		String type = null;
		
		
		if ( inputState.guessing==0 ) {
			switchToNormalMode();
		}
		match(PROGRAM);
		match(VARIABLES);
		match(LBRACE);
		{
		_loop219:
		do {
			if ((LA(1)==PIDENT||LA(1)==IDENT)) {
				{
				switch ( LA(1)) {
				case IDENT:
				{
					{
					v0 = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						type= "" + v0.getText();
					}
					{
					_loop216:
					do {
						if ((LA(1)==DOT)) {
							match(DOT);
							v1 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								type= type + "." + v1.getText();
							}
						}
						else {
							break _loop216;
						}
						
					} while (true);
					}
					}
					break;
				}
				case PIDENT:
				{
					{
					pid = LT(1);
					match(PIDENT);
					if ( inputState.guessing==0 ) {
						type = pid.getText();
					}
					}
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				{
				switch ( LA(1)) {
				case LBRACKET:
				{
					match(LBRACKET);
					match(RBRACKET);
					if ( inputState.guessing==0 ) {
						
						type = type + "[]";
						
					}
					break;
				}
				case PIDENT:
				case IDENT:
				{
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				var_name=id();
				if ( inputState.guessing==0 ) {
					
					KeYJavaType kjt = getJavaInfo().getKeYJavaType(type);
							ProgramElementName pvName = new ProgramElementName(var_name);
					if (kjt == null) {
					throw new antlr.SemanticException
					("Program Variable " + var_name + 
					" declared with unknown type " + 
					type);
					
					} else if (lookup(pvName)!=null) {
							  // commented out as pv do not have unique name (at the moment)
							    //throw new AmbigiousDeclException
							//	(var_name, getFilename(), getLine(), getColumn()); 
							}
					namespaces().programVariables().add(new ProgramVariable
					(new ProgramElementName(var_name), kjt));
					
					
				}
				match(SEMI);
			}
			else {
				break _loop219;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void schema_var_decls() throws RecognitionException, TokenStreamException {
		
		Token  id1 = null;
		Token  id2 = null;
		Token  id3 = null;
		Token  id5 = null;
		Token  id6 = null;
		Token  id7 = null;
		Token  id8 = null;
		
		Sort s = null;
		boolean makeVariableSV  = false;
		boolean makeDependingSV = false; // skolem functions
		boolean rigid           = false;
		boolean listModifier    = false;
		String type = "";    
		
		
		match(SCHEMA);
		match(VARIABLES);
		match(LBRACE);
		if ( inputState.guessing==0 ) {
			switchToSchemaMode();
		}
		{
		_loop244:
		do {
			switch ( LA(1)) {
			case MODAL:
			{
				{
				match(MODAL);
				match(OPERATOR);
				one_schema_modal_decl();
				{
				_loop224:
				do {
					if ((LA(1)==COMMA)) {
						match(COMMA);
						one_schema_modal_decl();
					}
					else {
						break _loop224;
					}
					
				} while (true);
				}
				match(SEMI);
				}
				break;
			}
			case PROGRAM:
			case FORMULA:
			case PIDENT:
			case IDENT:
			{
				if ( inputState.guessing==0 ) {
					makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;
				}
				{
				switch ( LA(1)) {
				case PROGRAM:
				{
					{
					match(PROGRAM);
					{
					switch ( LA(1)) {
					case SVLIST:
					{
						match(SVLIST);
						if ( inputState.guessing==0 ) {
							listModifier=true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						type = "";
					}
					{
					int _cnt231=0;
					_loop231:
					do {
						boolean synPredMatched230 = false;
						if (((LA(1)==IDENT))) {
							int _m230 = mark();
							synPredMatched230 = true;
							inputState.guessing++;
							try {
								{
								match(IDENT);
								match(IDENT);
								}
							}
							catch (RecognitionException pe) {
								synPredMatched230 = false;
							}
							rewind(_m230);
							inputState.guessing--;
						}
						if ( synPredMatched230 ) {
							id1 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								type += " "+id1.getText();
							}
						}
						else if ((LA(1)==IDENT)) {
							id2 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								type = type.substring(1,type.length());
								s = (Sort) DeclParser.name2sort.get(type);
								if (s == null) {
								throw new antlr.SemanticException
								("SchemaVariable of sort "+type+ " not found.");
								}
								schema_var_decl(id2.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
						}
						else {
							if ( _cnt231>=1 ) { break _loop231; } else {throw new NoViableAltException(LT(1), getFilename());}
						}
						
						_cnt231++;
					} while (true);
					}
					{
					_loop233:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							id3 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								schema_var_decl(id3.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
						}
						else {
							break _loop233;
						}
						
					} while (true);
					}
					match(SEMI);
					}
					break;
				}
				case FORMULA:
				{
					match(FORMULA);
					{
					switch ( LA(1)) {
					case RIGID:
					{
						match(RIGID);
						if ( inputState.guessing==0 ) {
							rigid = true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						s = Sort.FORMULA;type = "";
					}
					{
					id5 = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						schema_var_decl(id5.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
					}
					}
					{
					_loop238:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							{
							id6 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								schema_var_decl(id6.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
							}
						}
						else {
							break _loop238;
						}
						
					} while (true);
					}
					match(SEMI);
					break;
				}
				case PIDENT:
				case IDENT:
				{
					s=any_sortId_check(true);
					{
					switch ( LA(1)) {
					case VARIABLES:
					{
						match(VARIABLES);
						if ( inputState.guessing==0 ) {
							makeVariableSV = true;
						}
						break;
					}
					case DEPENDING:
					{
						match(DEPENDING);
						if ( inputState.guessing==0 ) {
							makeDependingSV = true;
						}
						break;
					}
					case RIGID:
					{
						match(RIGID);
						if ( inputState.guessing==0 ) {
							rigid = true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						type = "";
					}
					{
					id7 = LT(1);
					match(IDENT);
					if ( inputState.guessing==0 ) {
						schema_var_decl(id7.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
					}
					}
					{
					_loop243:
					do {
						if ((LA(1)==COMMA)) {
							match(COMMA);
							{
							id8 = LT(1);
							match(IDENT);
							if ( inputState.guessing==0 ) {
								schema_var_decl(id8.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);
							}
							}
						}
						else {
							break _loop243;
						}
						
					} while (true);
					}
					match(SEMI);
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				break;
			}
			default:
			{
				break _loop244;
			}
			}
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			switchToNormalMode();
		}
		match(RBRACE);
	}
	
	public final void pred_decls() throws RecognitionException, TokenStreamException {
		
		Token  cat = null;
		Token  choice = null;
		
		Sort[] argSorts;
		String pred_name;
		Namespace oldFNS=null;
		String ctext = null;
		Choice c=null;
		
		
		match(PREDICATES);
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			cat = LT(1);
			match(IDENT);
			match(COLON);
			choice = LT(1);
			match(IDENT);
			match(RPAREN);
			break;
		}
		case LBRACE:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			if(cat!=null){
			ctext = cat.getText()+":"+choice.getText();
			c = (Choice) choices().lookup(new Name(ctext));
			if(c==null){
			throw new NotDeclException("Option", choice,
			getFilename());
			}else{
			namespaces().setFunctions(c.funcNS());
			}
			}else{
			namespaces().setFunctions(defaultChoice.funcNS());
			}
			cat=null;
			
		}
		match(LBRACE);
		{
		_loop253:
		do {
			if ((LA(1)==PIDENT||LA(1)==IDENT)) {
				pred_name=id();
				argSorts=arg_sorts(!skip_predicates);
				if ( inputState.guessing==0 ) {
					
					if (!skip_predicates) {
					Name predicate = new Name(pred_name);		    		    
					if (lookup(predicate) != null) {
					throw new AmbigiousDeclException
					(pred_name, getFilename(), getLine(), getColumn());
					} else {
					Function p = new Function(predicate,
					Sort.FORMULA, 
					argSorts);
					functions().add(p);         
					}
					} 
					
				}
				match(SEMI);
			}
			else {
				break _loop253;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void func_decls() throws RecognitionException, TokenStreamException {
		
		Token  cat = null;
		Token  choice = null;
		
		Sort[] argSorts;
		Sort retSort;
		String func_name;
		Namespace oldFNS=null;
		String ctext = null;
		Choice c=null;
		
		
		match(FUNCTIONS);
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			cat = LT(1);
			match(IDENT);
			match(COLON);
			choice = LT(1);
			match(IDENT);
			match(RPAREN);
			break;
		}
		case LBRACE:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			if(cat!=null){
			ctext = cat.getText()+":"+choice.getText();
			c = (Choice) choices().lookup(new Name(ctext));
			if(c==null){
			throw new NotDeclException("Option", choice,
			getFilename());
			}
			}else{
			c=defaultChoice;
			}
			namespaces().setFunctions(c.funcNS());
			cat=null;
			
		}
		match(LBRACE);
		{
		_loop257:
		do {
			if ((LA(1)==PIDENT||LA(1)==IDENT)) {
				retSort=sortId_check(!skip_functions);
				func_name=id();
				argSorts=arg_sorts(!skip_functions);
				if ( inputState.guessing==0 ) {
					
					if (!skip_functions) {
					Name fct_name = new Name(func_name);
					if (lookup(fct_name) != null) {
					throw new AmbigiousDeclException
					(func_name, getFilename(), getLine(), getColumn());
					} else {
					functions().add(new Function(fct_name,
						                                             retSort, 
						                                     argSorts));
					}
					} 
					
				}
				match(SEMI);
			}
			else {
				break _loop257;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void heur_decls() throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		int     t_priority = 0;
		Integer priority   = null;
		
		
		match(HEURISTICS);
		match(LBRACE);
		{
		_loop266:
		do {
			if ((LA(1)==IDENT)) {
				if ( inputState.guessing==0 ) {
					priority = null;
				}
				id = LT(1);
				match(IDENT);
				{
				switch ( LA(1)) {
				case LPAREN:
				{
					match(LPAREN);
					t_priority=integer_num();
					match(RPAREN);
					if ( inputState.guessing==0 ) {
						priority = new Integer ( t_priority );
					}
					break;
				}
				case SEMI:
				{
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				match(SEMI);
				if ( inputState.guessing==0 ) {
					
					if (!skip_heuristics) {
					Heuristic h = new Heuristic(new Name(id.getText()));
					if(heuristics().lookup(new Name(id.getText()))==null){
					heuristics().add(h);
					}
					if ( priority != null )
					heur_prioritization.setPriority ( h, priority.intValue () );
					}
					
				}
			}
			else {
				break _loop266;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void choice() throws RecognitionException, TokenStreamException {
		
		Token  category = null;
		
		String cat=null;
		
		
		category = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			cat=category.getText();
		}
		{
		switch ( LA(1)) {
		case COLON:
		{
			match(COLON);
			match(LBRACE);
			choice_option(cat);
			{
			_loop194:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					choice_option(cat);
				}
				else {
					break _loop194;
				}
				
			} while (true);
			}
			match(RBRACE);
			break;
		}
		case SEMI:
		case RBRACE:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			if(!category2Default.containsKey(cat)){
			choices().add(new Choice("On",cat));
			choices().add(new Choice("Off",cat)); 
			category2Default.put(cat, cat+":On");               
			}
			
		}
	}
	
	public final void choice_option(
		String cat
	) throws RecognitionException, TokenStreamException {
		
		Token  choice = null;
		
		String name;
		
		
		choice = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			name=cat+":"+choice.getText();
			Choice c = (Choice) choices().lookup(new Name(name));
			if(c==null){
			c = new Choice(choice.getText(),cat);
			choices().add(c);
			}
			if(!category2Default.containsKey(cat)){
			category2Default.put(cat, name);
			}
			
		}
	}
	
	public final ListOfName  sortIdentifiers() throws RecognitionException, TokenStreamException {
		ListOfName ids = SLListOfName.EMPTY_LIST;
		
		Token  id = null;
		Token  id2 = null;
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			ids = ids.append ( new Name(id.getText()) );
		}
		{
		_loop204:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				id2 = LT(1);
				match(IDENT);
				if ( inputState.guessing==0 ) {
					ids = ids.append ( new Name(id2.getText()) );
				}
			}
			else {
				break _loop204;
			}
			
		} while (true);
		}
		return ids;
	}
	
	public final Sort[]  oneof_sorts() throws RecognitionException, TokenStreamException {
		Sort[] oneOfSorts = null;
		
		
		List args = new LinkedList();
		Sort s;
		
		
		match(LBRACE);
		s=sortId_check(true);
		if ( inputState.guessing==0 ) {
			args.add(s);
		}
		{
		_loop210:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s=sortId_check(true);
				if ( inputState.guessing==0 ) {
					args.add(s);
				}
			}
			else {
				break _loop210;
			}
			
		} while (true);
		}
		match(RBRACE);
		if ( inputState.guessing==0 ) {
			
			oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
			
		}
		return oneOfSorts;
	}
	
	public final Sort[]  extends_sorts() throws RecognitionException, TokenStreamException {
		Sort[] extendsSorts = null;
		
		
		List args = new LinkedList();
		Sort s;
		
		
		s=any_sortId_check(true);
		if ( inputState.guessing==0 ) {
			args.add(s);
		}
		{
		_loop207:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s=any_sortId_check(true);
				if ( inputState.guessing==0 ) {
					args.add(s);
				}
			}
			else {
				break _loop207;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			
			extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
			
		}
		return extendsSorts;
	}
	
	public final void one_schema_modal_decl() throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		Modality[] mods;
		
		
		id = LT(1);
		match(IDENT);
		match(EQUALS);
		mods=modalities_decl();
		if ( inputState.guessing==0 ) {
			
			SchemaVariable msv = (SchemaVariable)variables().lookup(new Name(id.getText()));
			if(msv != null)
			throw new antlr.SemanticException("Schema variable "+id.getText()+ " already defined");
			int arity = mods[0].arity();
			for(int i=1;i<mods.length;i++)
			if(arity != mods[i].arity())
			throw new antlr.SemanticException("Arity mismatch for schema variable "+id.getText());
			msv = SchemaVariableFactory.createModalOperatorSV(new Name(id.getText()), arity, mods);
			if (inSchemaMode()) {
			variables().add(msv);
			}
			
		}
	}
	
	public final Modality[]  modalities_decl() throws RecognitionException, TokenStreamException {
		Modality[] modalities = null;
		
		Token  s1 = null;
		Token  s2 = null;
		
		List args = new LinkedList();
		
		
		{
		match(LBRACE);
		s1 = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			
			if(s1.getText().charAt(0) == '#') {
			ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s1.getText()));
			if(msv == null)
			throw new antlr.SemanticException("Schema variable "+s1.getText()+ " not defined");
			for(int i=0; i<msv.modalities().length; i++)
			args.add(msv.modalities()[i]);
			}else{
			Modality op = Op.getModality(s1.getText());
			if(op == null)
			throw new antlr.SemanticException("Unrecognised modal operator "+s1.getText());
			args.add(op);
			}
			
		}
		{
		_loop249:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s2 = LT(1);
				match(IDENT);
				if ( inputState.guessing==0 ) {
					
					if(s2.getText().charAt(0) == '#') {
					ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s2.getText()));
					if(msv == null)
					throw new antlr.SemanticException("Schema variable "+s2.getText()+ " not defined");
					for(int i=0; i<msv.modalities().length; i++)
					args.add(msv.modalities()[i]);
					} else {
					Modality op = Op.getModality(s2.getText());
					if(op == null)
					throw new antlr.SemanticException("Unrecognised modal operator "+s2.getText());
					args.add(op);
					}
					
				}
			}
			else {
				break _loop249;
			}
			
		} while (true);
		}
		match(RBRACE);
		}
		if ( inputState.guessing==0 ) {
			
			modalities = (Modality[])args.toArray(new Modality[args.size()]);
			
		}
		return modalities;
	}
	
	public final Sort[]  arg_sorts(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort[] argSorts = null;
		
		
		List args = new LinkedList();
		Sort s;
		
		
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			s=sortId_check(checkSort);
			if ( inputState.guessing==0 ) {
				args.add(s);
			}
			{
			_loop261:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					s=sortId_check(checkSort);
					if ( inputState.guessing==0 ) {
						args.add(s);
					}
				}
				else {
					break _loop261;
				}
				
			} while (true);
			}
			match(RPAREN);
			break;
		}
		case SEMI:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
			
		}
		return argSorts;
	}
	
	public final int  integer_num() throws RecognitionException, TokenStreamException {
		 int num = 0 ;
		
		
		if ( inputState.guessing==0 ) {
			
			try {
			Token t = LT(1);
			num = Integer.decode ( t.getText () ).intValue ();
			consume ();
			} catch ( TokenStreamException e ) {
			throw new antlr.SemanticException ( "Expected a number",
			getFilename (), getLine (), getColumn () );
			} catch ( NumberFormatException e ) {
			throw new antlr.SemanticException ( "Expected a number",
			getFilename (), getLine (), getColumn () );
			} catch ( NullPointerException e ) {
			throw new antlr.SemanticException ( "Expected a number",
			getFilename (), getLine (), getColumn () );
			}
			
		}
		return num;
	}
	
	public final Sort  sortId_check_help(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		s=any_sortId_check_help(checkSort);
		if ( inputState.guessing==0 ) {
			
			// don't allow generic sorts or collection sorts of
			// generic sorts at this point
			Sort t = s;
			while ( t instanceof CollectionSort )
			t = ((CollectionSort)t).elementSort ();
			if ( t instanceof GenericSort )
			throw new GenericSortException ( "sort",
			"Non-generic sort expected", t,
			getFilename (), getLine (), getColumn () );
			
		}
		return s;
	}
	
	public final Sort  array_decls(
		Sort p
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		int  n=0;
		
		
		{
		_loop274:
		do {
			if ((LA(1)==LBRACKET)) {
				match(LBRACKET);
				match(RBRACKET);
				if ( inputState.guessing==0 ) {
					n++;
				}
			}
			else {
				break _loop274;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			
			if(n != 0){
			s = ArraySortImpl.getArraySortForDim(
			p, n, getJavaInfo().getObjectSort(),
			getJavaInfo().getCloneableSort());
			}else{
			s = p;
			}
			
		}
		return s;
	}
	
	public final Sort  any_sortId_check_help(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		String name;
		
		
		name=id();
		if ( inputState.guessing==0 ) {
			s= (Sort) sorts().lookup(new Name(name));
			if (checkSort && s == null) {
			throw new NotDeclException("sort", name, 
			getFilename(), 
			getLine(), getColumn());
			}
			
		}
		return s;
	}
	
	public final IdDeclaration  id_declaration() throws RecognitionException, TokenStreamException {
		 IdDeclaration idd = null ;
		
		Token  id = null;
		
		Sort s = null;
		
		
		id = LT(1);
		match(IDENT);
		{
		switch ( LA(1)) {
		case COLON:
		{
			match(COLON);
			s=sortId_check(true);
			break;
		}
		case EOF:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			idd = new IdDeclaration ( id.getText (), s );
			
		}
		return idd;
	}
	
	
	public static final String[] _tokenNames = {
		"<0>",
		"EOF",
		"<2>",
		"NULL_TREE_LOOKAHEAD",
		"\"new\"",
		"\"branch\"",
		"\"notsimple\"",
		"\"include\"",
		"\"LDTs\"",
		"\"nostandardrules\"",
		"\"typeof\"",
		"\"object\"",
		"\"number\"",
		"\"intliteral\"",
		"\"longliteral\"",
		"\"generic\"",
		"\"extends\"",
		"\"oneof\"",
		"\"program\"",
		"\"svlist\"",
		"\"all\"",
		"\"ex\"",
		"\"true\"",
		"\"false\"",
		"\"find\"",
		"\"if\"",
		"\"varcond\"",
		"\"containsquery\"",
		"\"noninteractive\"",
		"\"recursive\"",
		"\"displayname\"",
		"\"helptext\"",
		"\"not\"",
		"\"free\"",
		"\"in\"",
		"\"depending\"",
		"\"on\"",
		"\"close\"",
		"\"goal\"",
		"\"replacewith\"",
		"\"add\"",
		"\"addrules\"",
		"\"addprogvars\"",
		"\"heuristics\"",
		"\"hasSort\"",
		"\"sorts\"",
		"\"options\"",
		"\"with\"",
		"\"schema\"",
		"\"variables\"",
		"\"local\"",
		"\"rigid\"",
		"\"predicates\"",
		"\"functions\"",
		"\"formula\"",
		"\"modal\"",
		"\"operator\"",
		"\"rules\"",
		"\"proof\"",
		"\"rule\"",
		"\"term\"",
		"\"inst\"",
		"\"ifseqformula\"",
		"\"interactive\"",
		"\"heur\"",
		"\"modifies\"",
		"\"contracts\"",
		"\"problem\"",
		"\"#inType\"",
		"\"#isObject\"",
		"\"#inLong\"",
		"\"#inInt\"",
		"\"#inChar\"",
		"\"#inShort\"",
		"\"#inByte\"",
		"\"#add\"",
		"\"#sub\"",
		"\"#mul\"",
		"\"#div\"",
		"\"#jdiv\"",
		"\"#mod\"",
		"\"#jmod\"",
		"\"#less\"",
		"\"#greater\"",
		"\"#leq\"",
		"\"#geq\"",
		"\"#eq\"",
		"\"#JavaIntUnaryMinus\"",
		"\"#JavaIntAdd\"",
		"\"#JavaIntSub\"",
		"\"#JavaIntMul\"",
		"\"#JavaIntDiv\"",
		"\"#JavaIntMod\"",
		"\"#JavaIntAnd\"",
		"\"#JavaIntOr\"",
		"\"#JavaIntXor\"",
		"\"#JavaIntComplement\"",
		"\"#JavaIntShiftRight\"",
		"\"#JavaIntShiftLeft\"",
		"\"#JavaIntUnsignedShiftRight\"",
		"\"#JavaLongUnaryMinus\"",
		"\"#JavaLongAdd\"",
		"\"#JavaLongSub\"",
		"\"#JavaLongMul\"",
		"\"#JavaLongDiv\"",
		"\"#JavaLongMod\"",
		"\"#JavaLongAnd\"",
		"\"#JavaLongOr\"",
		"\"#JavaLongXor\"",
		"\"#JavaLongComplement\"",
		"\"#JavaLongShiftRight\"",
		"\"#JavaLongShiftLeft\"",
		"\"#JavaLongUnsignedShiftRight\"",
		"\"#moduloByte\"",
		"\"#moduloShort\"",
		"\"#moduloInteger\"",
		"\"#moduloLong\"",
		"\"#staticanalysis\"",
		"\"#ResolveQuery\"",
		"\"#constantvalue\"",
		"\"#lengthReference\"",
		"\"#transient\"",
		"\"#shadowed\"",
		"\"#concat\"",
		"\"#introduce\"",
		"\"waryAll\"",
		"\"waryEx\"",
		"\"same\"",
		"\"compatible\"",
		"\"sameUpdateLevel\"",
		"\"smaller\"",
		"\"than\"",
		"\"quotes\"",
		"\"ThisReference\"",
		"\"keyLog\"",
		"\"keyUser\"",
		"\"keyVersion\"",
		"\"keySettings\"",
		"\"ocl\"",
		"\"#allSubtypes\"",
		"VOCAB",
		"`;'",
		"`/'",
		"`\\'",
		"`:'",
		"`?='",
		"`/->'",
		"`:='",
		"`.'",
		"`,'",
		"`('",
		"`)'",
		"`{'",
		"`}'",
		"`['",
		"']'",
		"`at'",
		"`|'",
		"`&'",
		"`->'",
		"`<->'",
		"`='",
		"] followed by a sequence of primes (at least one)",
		"`==>'",
		"'^'",
		"java operator",
		"white space",
		"a string in double quotes",
		"PRIMES_OR_CHARLITERAL",
		"Sequence of primes (at least one)",
		"a char in single quotes",
		"ESC",
		"DIAMOND",
		"BOX",
		"THROUGHOUT",
		"MODALITYARG",
		"BLOCK",
		"a string with double quotes",
		"a comment",
		"a comment",
		"a digit",
		"a hexadeciamal number",
		"a letter",
		"an admissible character for identifiers",
		"an admissible character for identifiers protected with ``",
		"PIDENT",
		"an identifer",
		"\"elemtypeof\"",
		"\"-\""
	};
	
	private static final long[] mk_tokenSet_0() {
		long[] data = { 12582912L, 576460752303423472L, 432350100002457600L, 0L, 0L, 0L};
		return data;
	}
	public static final BitSet _tokenSet_0 = new BitSet(mk_tokenSet_0());
	private static final long[] mk_tokenSet_1() {
		long[] data = { 12582912L, 576460752303423472L, 432350099985664000L, 0L, 0L, 0L};
		return data;
	}
	public static final BitSet _tokenSet_1 = new BitSet(mk_tokenSet_1());
	private static final long[] mk_tokenSet_2() {
		long[] data = { 2L, 6917529027641081856L, 68361584640L, 0L, 0L, 0L};
		return data;
	}
	public static final BitSet _tokenSet_2 = new BitSet(mk_tokenSet_2());
	private static final long[] mk_tokenSet_3() {
		long[] data = { 2814784139427842L, 7493989779944505328L, 432350168432240640L, 0L, 0L, 0L};
		return data;
	}
	public static final BitSet _tokenSet_3 = new BitSet(mk_tokenSet_3());
	private static final long[] mk_tokenSet_4() {
		long[] data = { 34816L, 0L, 288230376151712768L, 0L, 0L, 0L};
		return data;
	}
	public static final BitSet _tokenSet_4 = new BitSet(mk_tokenSet_4());
	
	}
