header {

package de.uka.ilkd.key.parser;

import java.io.*;
import java.util.List;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import antlr.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.metaconstruct.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.ListOfKeYJavaType;
import de.uka.ilkd.key.java.abstraction.SLListOfKeYJavaType;
import de.uka.ilkd.key.java.JavaReader;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.SchemaRecoder2KeY;
import de.uka.ilkd.key.java.SchemaJavaReader;
import de.uka.ilkd.key.java.visitor.DeclarationProgramVariableCollector;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.java.visitor.*;
import de.uka.ilkd.key.gui.AbbrevMap;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.util.*;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;

}
class ProblemParser extends Parser;

options {
	importVocab=KeYLexer;
	k= 1;
	defaultErrorHandler= false;
}

{
    
    
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];
    protected AbbrevMap scm = new AbbrevMap();
    private Prioritization.StaticPrioritization heur_prioritization
        = new Prioritization.StaticPrioritization ();

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_heuristics;
    protected boolean skip_taclets;
    protected boolean onlyWith=false;
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected UseInsteadMap useInstead=new UseInsteadMap(); 

    // if this is used then we can capture parts of the input for later use
    protected DeclPicker capturer = null;

    protected int lineOffset=0;
    protected int colOffset=0;

    protected SetOfTaclet taclets = SetAsListOfTaclet.EMPTY_SET; 
    protected HashMap taclet2Builder;   
    protected Choice defaultChoice;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();
    protected HashMap category2Default = new HashMap();

    private static ListOfRulevariable NO_RULE_VARIABLES =
        SLListOfRulevariable.EMPTY_LIST;
    private Term a=null;

    protected TermFactory tf = TermFactory.DEFAULT;
    boolean schemaMode; 

    public boolean errors = false;
    
    private ParserConfig schemaConfig;
    private ParserConfig normalConfig;
    
    // the current active config
    private ParserConfig parserConfig;

    private ContractSet contracts = new ContractSet();

    public ProblemParser(TokenStream lexer, 
                         String filename,
                         ParserConfig schemaConfig,
                         ParserConfig normalConfig,
                         HashMap taclet2Builder,
                         SetOfTaclet taclets) { 
        this(lexer);
        setFilename(filename);
        if (lexer instanceof DeclPicker) {
            this.capturer = (DeclPicker) lexer;
        }
        if (normalConfig!=null) defaultChoice = normalConfig.getDefaultChoice(); 
        this.schemaConfig = schemaConfig;
        this.normalConfig = normalConfig;       
        switchToNormalMode();
        this.taclet2Builder=taclet2Builder;
        this.taclets = taclets;
    }

    protected void setSkips(boolean svars, boolean funcs, 
                            boolean preds, boolean sorts,
                            boolean heurs, boolean tacs) {

        skip_schemavariables = svars;
        skip_functions       = funcs;
        skip_predicates      = preds;
        skip_sorts           = sorts;
        skip_heuristics      = heurs;
        skip_taclets         = tacs;
    }
    

    protected void skipFuncs() {
        skip_functions = true;
    }
    
    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }

    protected boolean isPackage(String name){
        try {
            return getJavaInfo().isPackage(name);
        } catch(RuntimeException e){        
            // may be thrown in cases of invalid java identifiers
            return false;
        } 
    }

    protected void skipPreds() {
        skip_predicates = true;
    }

    protected void skipTaclets() {
        skip_taclets = true;
    }

    protected void skipVars() {
        skip_schemavariables = true;
    }

    protected void skipSorts() {
        skip_sorts = true;
    }

    protected void skipHeur() {
        skip_heuristics = true;
    }
    
    /** returns the java information object */
    public JavaInfo getJavaInfo() {
        return parserConfig.javaInfo();
    }


    /** parses a problem but without reading the declarations of
     * sorts, functions and predicates. These have to be given
     * explicitly.
     * @param functions() the Namespace with the function and predicate
     * symbols to be used
     * @param sorts() the Namespace with the sorts to be used
     * @param heuristics() the Namespace with the heuristics to be used
     * the heuristics of the current problem file will be added 
     */ 
    public Term parseTacletsAndProblem() 
        throws antlr.RecognitionException, antlr.TokenStreamException
        
        {
            // restore known state
            setSkips(false, false, false, false, false, false);
            // skip decls
            skipSorts(); 
            skipFuncs();
            skipPreds();    
            
            return problem();
        }

    /** parses a problem but without reading the declarations of
     * sorts, functions and predicates. These have to be given
     * explicitly.
     * @param functions() the Namespace with the function and predicate
     * symbols to be used
     * @param sorts() the Namespace with the sorts to be used
     * @param heuristics() the Namespace with the heuristics to be used
     * the heuristics of the current problem file will be added 
     */ 
    public Term parseProblem() 
        throws antlr.RecognitionException, antlr.TokenStreamException
        
        {
            // restore known state
            setSkips(false, false, false, false, false, false);
            // skip decls
            skipSorts(); 
            skipFuncs();
            skipPreds();    
            skipTaclets();
            return problem();
        }

    public void parseIncludes()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            parse_includes=true;
            problem();
        }

    public void parseWith()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            onlyWith=true;
            problem();
        }

    public void addSort(Sort s) {
        sorts().add(s);
    }

    public void addSorts(Namespace ns) {
        IteratorOfNamed it=ns.elements().iterator();
        while (it.hasNext()) {
            addSort((Sort)it.next());
        }
    }

    public void addFunction(Function f) {
        functions().add(f);
    }


    public void addFunctions(Namespace ns) {
        IteratorOfNamed it=ns.elements().iterator();
        while (it.hasNext()) {
            addFunction((Function)it.next());
        }
    }

    public Services getServices() {
        return parserConfig.services();
    }

    public NamespaceSet namespaces() {
        return parserConfig.namespaces();
    }

    public Namespace sorts() {
        return namespaces().sorts();
    }

    public Namespace variables() {
        return namespaces().variables();
    }

    public Namespace functions() {
        return namespaces().functions();
    }

    public Namespace heuristics() {
        return namespaces().heuristics();
    }

    public Namespace programVariables() {
        return namespaces().programVariables();
    }

    public Namespace choices(){
        return namespaces().choices();
    }

    public Named lookup(Name n) {
        final Namespace[] lookups = {
            normalConfig.namespaces().functions(), 
            schemaConfig.namespaces().functions(), 
            normalConfig.namespaces().variables(), 
            schemaConfig.namespaces().variables(), 
            schemaConfig.namespaces().programVariables()
        };

        for (int i = 0; i<lookups.length; i++) {
            if (lookups[i].lookup(n) != null) {
                return lookups[i].lookup(n);
            }
        }
        return null;    
    }

    public SetOfTaclet getTaclets(){
        return taclets;
    }

    public ContractSet getContracts(){
        return contracts;
    }

    public HashMap getCategory2Default(){
        return category2Default;
    }

    public void setOnlyWith(boolean onlyWith){
        this.onlyWith = onlyWith;
        parse_includes = false;
    }

    public SetOfChoice getActivatedChoices(){
        return activatedChoices;
    }

    public Prioritization getHeuristicPrioritization () {
        return heur_prioritization;
    }


    private void switchToNormalMode() {
        schemaMode   = false;
        parserConfig = normalConfig;
    }

    private boolean inSchemaMode() {
        return schemaMode;
    }

    private void switchToSchemaMode() {
        schemaMode   = true;
        parserConfig = schemaConfig;    
    }

    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine()+lineOffset;
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    protected int getColumn() {
        int     col = -1;
        try {
            col = LT(0).getColumn()+colOffset;
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }       

    protected TermSymbol progVar(Token id,Sort s) throws NotDeclException{
        TermSymbol v = (TermSymbol)
            variables().lookup(new Name(id.getText()));

        if (v == null) {
            throw new NotDeclException("program variable", id,
                                       getFilename());
        }
        return v;
    }

    protected static HashSet progVars(JavaBlock jb) {
        if (jb == JavaBlock.EMPTY_JAVABLOCK) {
            return new HashSet();
        }   
        DeclarationProgramVariableCollector pvc
            = new DeclarationProgramVariableCollector(jb.program());
        pvc.start();
        return pvc.result();
    }

    protected void bindProgVars(HashSet progVars) {
        namespaces().setVariables(new Namespace(variables()));
        Iterator it = progVars.iterator();
        while (it.hasNext()) {
            variables().add((IProgramVariable)it.next());
        }
    }

    public Choice getDefaultChoice(){
        return defaultChoice;
    }

    public Includes getIncludes(){
        return includes;
    }

    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }   
    
    protected void unbindProgVars() {
        namespaces().setVariables(variables().parent());
    }


    protected LogicVariable bindVar(Token id,Sort s) {
        LogicVariable v=new LogicVariable(new Name(id.getText()), s);
        namespaces().setVariables(variables().extended(v));
        return v;
    }
    
    protected void unbindVars() {
        namespaces().setVariables(variables().parent());
    }

    public Term createAttributeTerm(Term prefix, 
                                    TermSymbol attribute,
                                    int shadowPrimes) {
        Term result = prefix;
        if (!inSchemaMode()) {
            if (((ProgramVariable)attribute).isStatic()){
                result = tf.createVariableTerm((ProgramVariable)attribute);
            } else {
                if (shadowPrimes > 0) {
                    Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
                    result = tf.createShadowAttributeTerm((ProgramVariable)attribute, 
                                                          result, shadownum);
                } else {
                    result = tf.createAttributeTerm((ProgramVariable)attribute, result);
                }
            }
        } else {
            if(shadowPrimes>0){
                Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
                result = tf.createShadowAttributeTerm((SchemaVariable)attribute, result, shadownum);         
            } else
                result = tf.createAttributeTerm((SchemaVariable)attribute, result);         
        }

        return result;
    }

    public TermSymbol getAttribute(Sort prefixSort, 
                                   String attributeName)
        throws antlr.SemanticException  {
        final JavaInfo javaInfo = getJavaInfo();
        TermSymbol v = null;
        if (!inSchemaMode()) {
            if (attributeName.indexOf(':') != -1) {
                v = javaInfo.getAttribute(attributeName, 
                                          javaInfo.getKeYJavaType
                                          (attributeName.substring(0, attributeName.indexOf(':'))));
            } else {
                v = javaInfo.getAttribute(attributeName, (ObjectSort)prefixSort);
            }
        } else {
            v = (SortedSchemaVariable)variables().lookup(new Name(attributeName));
        }
        if ( v == null && !"length".equals(attributeName) ) {
            throw new NotDeclException
                ("Attribute ", attributeName, getFilename(), 
                 getLine(), getColumn());

        }

        return v;
    }


    protected Term termForParsedVariable(ParsableVariable v) 
        throws antlr.SemanticException 
        {
            if ( v instanceof LogicVariable ) {
                return tf.createVariableTerm((LogicVariable)v);
            } else if ( inSchemaMode()
                        && v instanceof SchemaVariable ) {
                return tf.createVariableTerm((SchemaVariable)v);
            } else if ( v instanceof ProgramVariable ) {
                return tf.createVariableTerm((ProgramVariable)v);
            }
            if ( inSchemaMode() ) {
                throw new antlr.SemanticException(
                                                  ""+ v +" is not a program or logic or schema variable");
            } else {
                throw new antlr.SemanticException(
                                                  ""+ v +" is not a logic variable "
                                                  +"or program variable");
            }
        }



    private TacletBuilder createTacletBuilderFor
        (Object find, boolean sameUpdatePrefix) 
        throws InvalidFindException {
        if ( sameUpdatePrefix && !( find instanceof Term ) )
            throw new InvalidFindException
                ("\"sameUpdateLevel\" may only be used for rewrite taclets:"+find,
                 getFilename(), getLine(), getColumn());
        if ( find == null ) {
            return new NoFindTacletBuilder();
        } else if ( find instanceof Term ) {
            return new RewriteTacletBuilder().setFind((Term)find)
                .setSameUpdatePrefix(sameUpdatePrefix);
        } else if ( find instanceof Sequent ) {
            Sequent findSeq = (Sequent) find;
            if ( findSeq == Sequent.EMPTY_SEQUENT ) {
                return new NoFindTacletBuilder();
            } else if (findSeq.antecedent().size() == 1
                       && findSeq.succedent().size() == 0 ) {
                return new AntecTacletBuilder().
                    setFind(findSeq.antecedent().get(0).formula());
            } else if (   findSeq.antecedent().size() == 0
                          && findSeq.succedent().size() == 1 ) {
                return new SuccTacletBuilder().
                    setFind(findSeq.succedent().get(0).formula());
            } else {
                throw new InvalidFindException
                    ("Unknown find-sequent (perhaps null?):"+findSeq,
                     getFilename(), getLine(), getColumn());    
            }
        } else {
            throw new IllegalArgumentException("createTacletBuilderFor called with"
                                               + find.getClass().getName() + "object");
        }
    }       



    private void addGoalTemplate(TacletBuilder b,
                                 String id,
                                 Object rwObj,
                                 Sequent addSeq,
                                 ListOfTaclet addRList,
                                 SetOfSchemaVariable pvs,
                                 SetOfChoice soc) 
        throws SemanticException
        {
            TacletGoalTemplate gt = null;
            if ( rwObj == null ) {
                // there is no replacewith, so we take
                gt = new TacletGoalTemplate(addSeq,
                                            addRList,
                                            NO_RULE_VARIABLES,
                                            pvs);
            } else {
                if ( b instanceof NoFindTacletBuilder ) {
                    // there is a replacewith without a find.
                    throw
                        new UnfittingReplacewithException
                        ("Replacewith without find", getFilename(),
                         getLine(), getColumn());
                } else if ( b instanceof SuccTacletBuilder
                            || b instanceof AntecTacletBuilder ) {
                    if ( rwObj instanceof Sequent ) {
                        gt = new AntecSuccTacletGoalTemplate(addSeq,
                                                             addRList,
                                                             NO_RULE_VARIABLES,
                                                             (Sequent)rwObj, pvs); 
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a Antec-or SuccTaclet has "+
                             "to contain a sequent (not a term)", 
                             getFilename(), getLine(), getColumn());
                    }
                } else if ( b instanceof RewriteTacletBuilder ) {
                    if ( rwObj instanceof Term ) {
                        gt = new RewriteTacletGoalTemplate(addSeq,
                                                           addRList,
                                                           NO_RULE_VARIABLES,
                                                           (Term)rwObj, pvs);  
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a RewriteTaclet has "+
                             "to contain a term (not a sequent)", 
                             getFilename(), getLine(), getColumn());
                    }
                }
            }
            gt.setName(id);
            b.addTacletGoalTemplate(gt);
            if(soc != null) b.addGoal2ChoicesMapping(gt,soc);
        }
    

    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
        String s=t.getText();
        de.uka.ilkd.key.java.JavaReader jr=null;
        try {
            if (inSchemaMode()) {
                jr = new SchemaRecoder2KeY(parserConfig.services(), 
                                           parserConfig.namespaces());
                ((SchemaJavaReader)jr).setSVNamespace(variables());
                return jr.readBlockWithEmptyContext(s);
            } else {
                jr = new Recoder2KeY(parserConfig.services(), 
                                     parserConfig.namespaces());
                return jr.readBlockWithProgramVariables(variables(), s);    
            }
        } catch (de.uka.ilkd.key.java.PosConvertException e) {
            lineOffset=e.getLine()-1;
            colOffset=e.getColumn();
            throw new JavaParserException(e.getMessage(), t, 
                                          getFilename(), lineOffset, colOffset);            
            
        } catch (de.uka.ilkd.key.java.ConvertException e) {
            e.printStackTrace(); ///
                if (s.length()>10) {
                    s=s.substring(0,10)+"...";
                }
                e.printStackTrace(); ///
                    if (e.parseException()!=null
                        &&  e.parseException().currentToken != null
                        &&  e.parseException().currentToken.next != null) {       
                        lineOffset=e.parseException().currentToken.next.beginLine;
                        colOffset=e.parseException().currentToken.next.beginColumn;
                        e.parseException().currentToken.next.beginLine=getLine()-1;
                        e.parseException().currentToken.next.beginColumn=getColumn();
                        throw new JavaParserException(e.parseException().getMessage(), t, 
                                                      getFilename(), -1, -1);  // row/columns already in text
                    }       

                    throw new JavaParserException(e.getMessage(), t, getFilename());
        }
    }


    
    public void schema_var_decl(String name,Sort s,
                                boolean makeVariableSV,
                                boolean makeDependingSV,
                                boolean rigid,
                                boolean listModifier)
        {
            if (!skip_schemavariables) {
                SchemaVariable v;
                if ( s == Sort.FORMULA ) {
                    v = SchemaVariableFactory.createFormulaSV
                        (new Name(name), listModifier, rigid);
                } else if ( s instanceof ProgramSVSort ) {
                    v = SchemaVariableFactory.createProgramSV
                        (new ProgramElementName(name),(ProgramSVSort) s, listModifier);
                } else {
                    if ( makeVariableSV ) {
                        v=SchemaVariableFactory.createVariableSV
                            (new Name(name), s, listModifier);
                    } else if ( makeDependingSV ) {
                        v = SchemaVariableFactory.createDependingSV
                            (new Name(name), s, listModifier);
                    } else { v = SchemaVariableFactory.createTermSV
                            (new Name(name), s, listModifier, rigid);
                    }
                }
                if (inSchemaMode()) {
                    variables().add(v);
                }
            }

        }
    


    public void testLiteral(String l1, String l2)
        throws KeYSemanticException
        {
            if (!l1.equals(l2)){
                throw new KeYSemanticException
                    ("Expecting '"+l1+"', found '"+l2+"'.",getLine(),getColumn(),getFilename());
            };
        }

    /** looks up a function, (program) variable or static query of the 
     * given name varfunc_id and the argument terms args in the namespaces 
     * and java info. 
     * @param args is null iff no argument list is given, for instance `f', 
     * and is an array of size zero, if an empty argument list was given,
     * for instance `f()'.
     */
    protected Operator lookupVarfuncId(String varfunc_name, Term[] args) 
        throws NotDeclException{
        // case 1: variable
        Operator v = (TermSymbol) variables().lookup(new Name(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        // case 2: function
        v = (Function) functions().lookup(new Name(varfunc_name));
        if (v != null) { // we allow both args==null (e.g. `c')
                         // and args.length=0 (e.g. 'c())' here 
            return v;
        }
        // case 3: static query
        if (varfunc_name.lastIndexOf("::")>0 && args!=null) {
            String classname = varfunc_name.substring
                (0, varfunc_name.lastIndexOf("::"));
            varfunc_name = varfunc_name.substring
                (varfunc_name.lastIndexOf("::")+2);
            v = getServices().getJavaInfo().getProgramMethodTerm
                (null, varfunc_name, args, classname).op();
            if (v != null) {
                return v;
            }
        }
        // case 4: program variable
        v = (ProgramVariable) programVariables().lookup
            (new ProgramElementName(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        String[] argTypeStrings = null;
        if (args!=null) {
            argTypeStrings = new String[args.length];
            for (int i=0; i<args.length; i++) {
                argTypeStrings[i] = args[i].sort().toString();
            }
        }
        if (args==null) {
            throw new NotDeclException
                ("(program) variable or constant", varfunc_name,
                 getFilename(), getLine(), getColumn());
        } else {
            throw new NotDeclException
                ("function or static query", varfunc_name,
                 getFilename(), getLine(), getColumn());
        }
    }
    
}
top :{} a=problem;

problem returns [ Term a = null ]{
    Taclet s = null;
    SetOfChoice choices=SetAsListOfChoice.EMPTY_SET;
    Choice c = null;
    String ctext = null;
    String st = null;
    String[] varDecls = null;
    if (capturer != null) capturer.mark();
    Namespace funcNSForActivatedChoices = new Namespace();
    Term fma = null;
    Term modifTerm = null;
    SetOfTerm termSet = null;
    String contractName = null;
}
:( st = java )
        decls
        { 
            if(parse_includes || onlyWith) return null;
            switchToNormalMode(); 
            variables().add(parserConfig.namespaces().programVariables()); 
             IteratorOfChoice it = activatedChoices.iterator(); 
             while(it.hasNext()){
                 funcNSForActivatedChoices=(funcNSForActivatedChoices.
                                            extended(it.next().funcNS().allElements()));
             } 
             funcNSForActivatedChoices=funcNSForActivatedChoices.
                 extended(defaultChoice.
                          funcNS().allElements()); 
        }
        ( "contracts"
            LBRACE {
                switchToNormalMode();
                 IteratorOfChoice it = choices.iterator();
                 Namespace funcNSForRules = funcNSForActivatedChoices;
                 while(it.hasNext()){
                     c=it.next();
                     funcNSForRules = 
                         funcNSForRules.extended(c.funcNS().allElements());
                 }
                 namespaces().setFunctions(funcNSForRules); 
            }
            (   {contractName = null;}
                (name:STRING_LITERAL {contractName = name.getText();})?
                fma = formula ("modifies" {termSet = SetAsListOfTerm.EMPTY_SET;}
                    (modifTerm=accessterm 
                        {termSet=termSet.add(modifTerm);}
                    (COMMA modifTerm=accessterm 
                        {termSet=termSet.add(modifTerm);})*)? )?
                { //Syntax check fma
                    contracts.add(new DLHoareTripleContract(fma, termSet,
                        heuristics(), contractName, getServices()));
                } SEMI )*
            RBRACE 
        )*
        (  "rules" 
            (LPAREN choices=option[choices]
                (COMMA choices=option[choices])* RPAREN )? LBRACE
            { 
                switchToSchemaMode(); 
                 IteratorOfChoice it = choices.iterator();
                 Namespace funcNSForRules = funcNSForActivatedChoices;
                 while(it.hasNext()){
                     c=it.next();
                     funcNSForRules = 
                         funcNSForRules.extended(c.funcNS().allElements());
                 }
                 namespaces().setFunctions(funcNSForRules); 
            }
            ( 
                s = taclet[choices] SEMI
                {
                    try {
                        if (!skip_taclets) {
                            taclets = taclets.addUnique(s);
                        }
                    } catch(de.uka.ilkd.key.collection.NotUniqueException e) {
                        throw new IllegalArgumentException(e.toString());
                    }
                }
            )*
            RBRACE {choices=SetAsListOfChoice.EMPTY_SET;}
        ) *
        (PROBLEM LBRACE 
            {switchToNormalMode(); if (capturer != null) capturer.capture();}
            a = formula RBRACE)?
    ;

java returns [String s = null]{StringBuffer b=new StringBuffer(); }
:((javaId:IDENT{testLiteral("java",javaId.getText());} LBRACE (l:STRING_LITERAL {b.append(l.getText());} 
                | SLASH {b.append("/");})+
                {s=b.toString();} RBRACE))?
    ;

proof[ProblemLoader prl] :( PROOF LBRACE
            ( pseudosexpr[prl] )+ 
            RBRACE
        )?
    ;

pseudosexpr[ProblemLoader prl] { char eid='0'; }
:LPAREN (eid=expreid
            (str:STRING_LITERAL { prl.beginExpr(eid,str.getText()); } )? 
            ( pseudosexpr[prl] )* ) ?
        { prl.endExpr(eid); }
        RPAREN   
    ;

expreid returns [ char eid = '0' ]:FORMULA { eid='f'; }
   |
   TERM    { eid='t'; }
   |
   RULE    { eid= 'r'; }
   |
   BRANCH  { eid='b'; }
   |
   INST    { eid='i'; }
   |
   IFSEQFORMULA { eid='q'; }
   |
   INTERACTIVE  { eid='a'; }
   |
   HEUR     { eid='h'; }
   |
   IDENT
   |
   KEYLOG { eid='l'; }
   |
   KEYLOGUSER { eid='u'; }
   |
   KEYLOGVERSION {eid='v'; }
   |
   KEYLOGSETTINGS {eid='s'; }
;

// inherited from grammar TacletParser
taclet[SetOfChoice choices] returns [Taclet r]{ 
    Sequent ifSeq = Sequent.EMPTY_SEQUENT;
    Object  find = null;
    r = null;
    TacletBuilder b = null;
    boolean sameUpdatePrefix = false;
}
:name:IDENT 
        (LPAREN choices=option[choices](COMMA choices=option[choices])* RPAREN)? 
        LBRACE
        ( "if" LPAREN ifSeq=seq RPAREN ) ?
        ( "find" LPAREN find = termorseq RPAREN
            ( "sameUpdateLevel" { sameUpdatePrefix = true; } ) ? ) ?
        { 
            b = createTacletBuilderFor(find, sameUpdatePrefix);
            b.setName(new Name(name.getText()));
            b.setIfSequent(ifSeq);
        }
        ( "varcond" LPAREN varexplist[b] RPAREN ) ?
        goalspecs[b]
        modifiers[b]
        RBRACE
        { 
            b.setChoices(choices);
            r = b.getTaclet(); 
            taclet2Builder.put(r,b);
        }
    ;

// inherited from grammar TacletParser
modifiers[TacletBuilder b] :( heuristics[b] 
        | "noninteractive" { 
                //      b.setNoninteractive(true);  
                // "noninteractive" (as it is now) is confusing
                // dropped this completely until a better solution (->Uwe)
            }
        | "recursive"      { b.setRecursive(true); }
        | DISPLAYNAME dname:STRING_LITERAL 
            {b.setDisplayName(dname.getText());}
        | HELPTEXT htext:STRING_LITERAL
            {b.setHelpText(htext.getText());}
        ) *
    ;

// inherited from grammar TacletParser
seq returns [Sequent s]{Semisequent ant,suc; s = null; }
:ant=semisequent SEQARROW suc=semisequent 
        { s = Sequent.createSequent(ant, suc); }
    ;

// inherited from grammar TacletParser
termorseq returns [Object o]{
    Term head = null;
    Sequent s = null;
    Semisequent ss = null;
    o = null; 
}
:head=term ( COMMA s=seq | SEQARROW ss=semisequent ) ?
        {
            if ( s == null ) {
                if ( ss == null ) {
                    // Just a term
                    o = head;
                } else {
                    // A sequent with only head in the antecedent.
                    Semisequent ant = Semisequent.EMPTY_SEMISEQUENT;
                    ant = ant.insertFirst(
                                          new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
                    o = Sequent.createSequent(ant,ss);
                }
            } else {
                // A sequent.  Prepend head to the antecedent.
                Semisequent newAnt = s.antecedent();
                newAnt = newAnt .insertFirst(
                                             new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
                o = Sequent.createSequent(newAnt,s.succedent());
            }
        }
    |
        SEQARROW ss=semisequent
        {
            o = Sequent.createSequent(Semisequent.EMPTY_SEMISEQUENT,ss);
        }
    ;

// inherited from grammar TacletParser
semisequent returns [Semisequent ss]{ 
    Term head = null;
    ss = Semisequent.EMPTY_SEMISEQUENT; 
}
:/* empty */ | 
        head=term ( COMMA ss=semisequent ) ? 
        { ss = ss.insertFirst(new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent(); }
    ;

// inherited from grammar TacletParser
varexplist[TacletBuilder b] :varexp[b] ( COMMA varexp[b] ) * ;

// inherited from grammar TacletParser
varexp[TacletBuilder b] {
    ParsableVariable x;
    ParsableVariable y; 
    ParsableVariable z;
    Sort s = null;
    boolean negation = false;
}
:(
            "typeof" LPAREN y=varId RPAREN  
		(x=varId "new" 
            		{ b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, false); }
		| "not" (
			"same" "typeof" LPAREN x=varId RPAREN
            		{ b.addVariableCondition
			   (new TypeComparisionCondition((SchemaVariable) y, (SchemaVariable) x, 
				TypeComparisionCondition.NOT_SAME)); }
			|
			"compatible" "typeof" LPAREN x=varId RPAREN 
            		{ b.addVariableCondition
			   (new TypeComparisionCondition((SchemaVariable) y, (SchemaVariable) x, 
				TypeComparisionCondition.NOT_COMPATIBLE)); }
			)
		)
        | "elemtypeof" LPAREN y=varId  RPAREN  x=varId "new"
            { b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, true); }
        | (sortId_check[true] varId "new") => s=sortId_check[true]  x=varId "new"
            {b.addVarsNew((SchemaVariable) x, s); }

                
	| (IDENT varId varId varId) => 
	   (mId:IDENT { testLiteral("methodreference",mId.getText());} 
	    x=varId y=varId z=varId ("not" { negation = true; } )? sId1:IDENT {
                    testLiteral("static",sId1.getText());
		    b.addVariableCondition(new StaticMethodCondition
                                           (negation, (SchemaVariable)x, (SchemaVariable)y,
                                            (SchemaVariable)z));
                    negation = false; 
                })
	| x = varId
            ( "not" (
                    "free" "in" y=varId 
                    { b.addVarsNotFreeIn((SchemaVariable)x,(SchemaVariable)y); }
                | ( {"static".equals(LT(1).getText())}?
                    sId2:IDENT
                    { testLiteral("static",sId2.getText());
		      b.addVariableCondition(new StaticReferenceCondition
                                             ((SchemaVariable) x, true));}		
		    | 
                    sId4:IDENT y=varId 
                    { testLiteral("sameLiteral",sId4.getText());
		      b.addVariableCondition(new TestLiteral
                                             ((SchemaVariable) x, (SchemaVariable) y));
		    }
                ) 
	    )
            | sId3:IDENT
                { testLiteral("static",sId3.getText());
		   b.addVariableCondition
                  (new StaticReferenceCondition((SchemaVariable) x,
                                                false));}
            | "new" "depending" "on" y=varId
                { b.addVarsNewDependingOn((SchemaVariable)x,(SchemaVariable)y); }
            | "smaller" "than" y=varId
                { b.addVariableCondition(
                        new OrderingCondition ((SchemaVariable)x,(SchemaVariable)y)); }
            | "containsquery" 
                { b.addVariableCondition(new ContainsQueryCondition
                    ((SchemaVariable)x));}
            | "hasSort" s=any_sortId_check[true]
               { if ( !( s instanceof GenericSort ) )
                   throw new GenericSortException ( "sort",
                                                    "Generic sort expected", s,
                                                    getFilename (), getLine (), getColumn () );
                 if ( !( x instanceof SortedSchemaVariable
                         && ( (SortedSchemaVariable)x ).sort () ==
                                ProgramSVSort.EXPRESSION ) )
                   throw new antlr.SemanticException
                                  ("Expected schema variable of kind EXPRESSION, got " + x);
                 b.addVariableCondition(new JavaTypeToSortCondition ((SchemaVariable)x, (GenericSort)s)); }

            
	    )        
        )
    ;

// inherited from grammar TacletParser
goalspecs[TacletBuilder b] :"close" "goal"
    | goalspecwithoption[b] ( SEMI goalspecwithoption[b] )* ;

// inherited from grammar TacletParser
goalspecwithoption[TacletBuilder b] {
    SetOfChoice soc = SetAsListOfChoice.EMPTY_SET;
}
:((LPAREN soc=option[soc](COMMA soc=option[soc])* RPAREN 
                LBRACE
                goalspec[b,soc] 
                RBRACE)
        |  
            goalspec[b,null] 
        )
    ;

// inherited from grammar TacletParser
option[SetOfChoice soc] returns [SetOfChoice result=null]{
    Choice c = null;
}
:cat:IDENT COLON choice:IDENT
        {
            c = (Choice) choices().lookup(new Name(cat.getText()+":"+choice.getText()));
            if(c==null){
                throw new NotDeclException("Option", choice,
                                           getFilename());
            }else{
                result = soc.add(c);
            }
        }
    ;

// inherited from grammar TacletParser
goalspec[TacletBuilder b, SetOfChoice soc] {
    Object rwObj = null;
    Sequent addSeq = Sequent.EMPTY_SEQUENT;
    ListOfTaclet addRList = SLListOfTaclet.EMPTY_LIST;
    SetOfSchemaVariable addpv = SetAsListOfSchemaVariable.EMPTY_SET;
    String name = null;
}
:(id:STRING_LITERAL COLON {name = id.getText();})?
        (   ( rwObj = replacewith
                (addSeq=add)? 
                (addRList=addrules)? 
                (addpv=addprogvar)?
            )
        | ( addSeq=add (addRList=addrules)? )
        | ( addRList=addrules )
        )
        {
            addGoalTemplate(b,name,rwObj,addSeq,addRList,addpv,soc);
        }
        
    ;

// inherited from grammar TacletParser
replacewith returns [Object o]{ o = null; }
:"replacewith" LPAREN o=termorseq RPAREN;

// inherited from grammar TacletParser
add returns [Sequent s]{ s = null;}
:"add" LPAREN s=seq RPAREN;

// inherited from grammar TacletParser
addrules returns [ListOfTaclet lor]{ lor = null; }
:"addrules" LPAREN lor=tacletlist RPAREN;

// inherited from grammar TacletParser
addprogvar returns [SetOfSchemaVariable pvs]{pvs = null; }
:"addprogvars" LPAREN pvs=pvset RPAREN;

// inherited from grammar TacletParser
tacletlist returns [ListOfTaclet lor]{ 
    Taclet head = null;
    lor = SLListOfTaclet.EMPTY_LIST; 
}
:head=taclet[SetAsListOfChoice.EMPTY_SET]   
        ( /*empty*/ | COMMA lor=tacletlist ) { lor = lor.prepend(head); }
    ;

// inherited from grammar TacletParser
pvset returns [SetOfSchemaVariable pvs]{
    ParsableVariable pv = null;
    pvs = SetAsListOfSchemaVariable.EMPTY_SET;
}
:pv=varId
        ( /*empty*/ | COMMA pvs=pvset ) { pvs = pvs.add
                                          ((SchemaVariable)pv); };

// inherited from grammar TacletParser
heuristics[TacletBuilder b] :"heuristics" LPAREN heuristic[b] ( COMMA heuristic[b] ) * RPAREN ;

// inherited from grammar TacletParser
heuristic[TacletBuilder b] :id:IDENT
        {   
            Heuristic h = (Heuristic) heuristics().lookup(new Name(id.getText()));
            if (h == null) {
                throw new NotDeclException("heuristic", id,
                                           getFilename());
            }
            b.addHeuristic(h);
        }
    ;

// inherited from grammar TacletParser
bindingterm returns [Term a = null]{}
:a = sv_or_local_bindingterm
    ;

// inherited from grammar TacletParser
sv_or_local_bindingterm returns [Term a = null]{
    Operator op = null;
    QuantifiableVariable v = null;
    TermSymbol ts = null;
    Term a1;
    Sort s = null;
    boolean bound = false;
}
:(( 
                "all" { op = Op.ALL; }
            |   "ex"  { op = Op.EX;  }
            )
            v0:IDENT
            (   
                COLON s=sortId 
                {
                    v = bindVar(v0,s);
                    bound = true;
                }
            |
                /*empty*/
                {
                    ts = (TermSymbol) variables().lookup(new Name(v0.getText()));   
                    if (   ts == null 
                           || ts instanceof LogicVariable ) {
                        throw new SemanticException(
                                                    "Quantified variables need a sort.");
                    }
                    if ( ! (ts instanceof SchemaVariable && 
                            ((SchemaVariable)ts).isVariableSV())) {
                        throw new SemanticException(
                                                    ""+ts+" is not allowed in a quantifier.");
                    }
                    v = (SortedSchemaVariable) ts;
                    bound = false;
                }
            )
            DOT a1 = term60 
            { 
                a = tf.createQuantifierTerm((Quantifier)op, v, a1);
                if (bound) {unbindVars();}
            })
    |
        a = update_or_substitution
    ;

// inherited from grammar TacletParser
argument returns [TermWithBoundVars p = null]{
    List listVars = new LinkedList();
    SortedSchemaVariable var = null;
    ArrayOfQuantifiableVariable vars = null;
    SortedSchemaVariable[] arrayVars = null;
    Term a1;

    LogicVariable v = null;
    LogicVariable[] logArrayVars = null;
    Sort s = null;
}
:(
            BACKSLASH v0:IDENT 
            (
                //Takes care of parsing the "problem" part, i.e. interprets
                //variables as logical variables.
                COLON s=sortId
                {
                    v = bindVar(v0, s);
                    listVars.add(v);
                }
                (
                    BACKSLASH v1:IDENT COLON s=sortId
                    {
                        v = bindVar(v1, s);
                        listVars.add(v);
                    }
                )* SLASH a1=term60
                {
                    logArrayVars = (LogicVariable[])listVars.toArray(new LogicVariable[listVars.size()]);
                    vars = new ArrayOfQuantifiableVariable(logArrayVars);
                    p = new TermWithBoundVars(vars, a1);
                    unbindVars();
                }
            |
                //Takes care of parsing the "rules" part, i.e. interprets variables
                //as schema variables.
                {
                    var = (SortedSchemaVariable)variables().lookup(new Name(v0.getText())); 
                    listVars.add(var);
                }
                (
                    BACKSLASH v2:IDENT
                    {
                        var = (SortedSchemaVariable)variables().lookup(new Name(v2.getText())); 
                        listVars.add(var);
                    }
                )* SLASH a1=term60
                {
                    arrayVars = (SortedSchemaVariable[])listVars.toArray(new SortedSchemaVariable[listVars.size()]);
                    vars = new ArrayOfQuantifiableVariable(arrayVars);
                    p = new TermWithBoundVars(vars, a1);
                }
            )
	    | a1=term60
            {
                p = new TermWithBoundVars(vars, a1);
            }
        )
	;

// inherited from grammar TacletParser
metaId returns [Operator v = null]:META_SHADOW 
        { v = Op.META_SHADOW ; }
    | CONSTANT_VALUE
        { v = Op.CONSTANT_VALUE; }
    | META_INT_UNARY_MINUS 
        { v = Op.META_INT_UNARY_MINUS ; }
    | META_LONG_UNARY_MINUS 
        { v = Op.META_LONG_UNARY_MINUS ; }
    | META_ADD 
        { v = Op.META_ADD; }
    | META_SUB
        { v = Op.META_SUB; }
    | META_MUL
        { v = Op.META_MUL; }
    | META_DIV 
        { v = Op.META_DIV; }
    | META_JDIV 
        { v = Op.META_JDIV; }
    | META_MOD
        { v = Op.META_MOD; }
    | META_JMOD
        { v = Op.META_JMOD; }
    | META_LESS
        { v = Op.META_LESS; }
    | META_GREATER
        { v = Op.META_GREATER; }
    | META_LEQ
        { v = Op.META_LEQ; }
    | META_GEQ
        { v = Op.META_GEQ; }
    | META_EQ
        { v = Op.META_EQ; }
    | META_INT_ADD 
        { v = Op.META_INT_ADD; }
    | META_INT_SUB 
        { v = Op.META_INT_SUB; }
    | META_INT_MUL 
        { v = Op.META_INT_MUL; }
    | META_INT_DIV 
        { v = Op.META_INT_DIV; }
    | META_INT_MOD 
        { v = Op.META_INT_MOD; }
    | META_INT_AND 
        { v = Op.META_INT_AND; }
    | META_INT_OR
        { v = Op.META_INT_OR; }
    | META_INT_XOR
        { v = Op.META_INT_XOR; }
    | META_INT_COMPLEMENT
        { v = Op.META_INT_COMPLEMENT; }
    | META_INT_SHIFTRIGHT
        { v = Op.META_INT_SHIFTRIGHT; }
    | META_INT_SHIFTLEFT
        { v = Op.META_INT_SHIFTLEFT; }
    | META_INT_UNSIGNEDSHIFTRIGHT
        { v = Op.META_INT_UNSIGNEDSHIFTRIGHT; }
    | META_LONG_ADD 
        { v = Op.META_LONG_ADD; }
    | META_LONG_SUB 
        { v = Op.META_LONG_SUB; }
    | META_LONG_MUL 
        { v = Op.META_LONG_MUL; }
    | META_LONG_DIV 
        { v = Op.META_LONG_DIV; }
    | META_LONG_MOD 
        { v = Op.META_LONG_MOD; }
    | META_LONG_AND 
        { v = Op.META_LONG_AND; }
    | META_LONG_OR
        { v = Op.META_LONG_OR; }
    | META_LONG_XOR
        { v = Op.META_LONG_XOR; }
    | META_LONG_COMPLEMENT
        { v = Op.META_LONG_COMPLEMENT; }
    | META_LONG_SHIFTRIGHT
        { v = Op.META_LONG_SHIFTRIGHT; }
    | META_LONG_SHIFTLEFT
        { v = Op.META_LONG_SHIFTLEFT; }
    | META_LONG_UNSIGNEDSHIFTRIGHT
        { v = Op.META_LONG_UNSIGNEDSHIFTRIGHT; }
    | META_IN_LONG 
        { v = Op.META_IN_LONG; }
    | META_IN_INT
        { v = Op.META_IN_INT; }
    | META_IN_CHAR
        { v = Op.META_IN_CHAR; }
    | META_IN_SHORT
        { v = Op.META_IN_SHORT; }
    | META_IN_BYTE 
        { v = Op.META_IN_BYTE; }
    | META_MODULO_BYTE
        { v = Op.META_MODULO_BYTE; }
    | META_MODULO_SHORT
        { v = Op.META_MODULO_SHORT; }
    | META_MODULO_INTEGER
        { v = Op.META_MODULO_INTEGER; }
    | META_MODULO_LONG 
        { v = Op.META_MODULO_LONG; }    
    | STATIC_ANALYSIS
        { v = Op.STATIC_ANALYSIS; }
    | RESOLVE_QUERY
        { v = Op.RESOLVE_QUERY; }
    | META_LENGTH
        { v = Op.META_LENGTH; }
    | META_TRANSIENT
        { v = Op.META_TRANSIENT; }

    //For OCL simplification
    | META_ALLSUBTYPES
        {v = Op.METAALLSUBTYPES;}
    ;

// inherited from grammar TacletParser
metaTerm returns [Term result = null]{
    LinkedList al = new LinkedList();
    Operator vf = null;
    Term t = null;
}
:(vf = metaId
            LPAREN 
            t = term
            {
                al.add(t);
            }
            ( COMMA 
                t = term
                {
                    al.add(t);
                }   
            )* RPAREN                      
            {   
                result = tf.createMetaTerm((MetaOperator)vf,
                                           (Term[])al.toArray(AN_ARRAY_OF_TERMS));
            }         
        ) 

   ;

// inherited from grammar TacletParser
update_or_substitution returns [Term result = null]{}
:(
             LBRACE IDENT 
             (
                 COLON sortId 
             | 
             //For OCL simplification
                 SLASH term60 
             | 
             //
                 logicTermReEntry
             )
         ) => result = substitutionterm 

    | result = updateterm 
    ;

// inherited from grammar TacletParser
substitutionterm returns [Term result = null]{ Term a1, a2 = null, a3;
Sort s = null;
TermSymbol ts;
QuantifiableVariable v = null, vv = null;
boolean bound = false;
boolean waryAll = false;
boolean waryEx = false;
}
:LBRACE v1:IDENT
        (COLON s=sortId {
                bound = true;
            } | 
            { functions().lookup(new Name(v1.getText())) == null }?             
            {
                ts = (TermSymbol) variables().lookup(new Name(v1.getText()));
                if (   ts == null
                       || ts instanceof LogicVariable ) {
                    throw new SemanticException("Substituted variables need a sort.");
                }
                if ( ! (ts instanceof SchemaVariable &&
                        ((SchemaVariable)ts).isVariableSV()) ) {
                    throw new SemanticException(
                                                "   "+ts+" is not allowed as bound variable "
                                                +"in a substitution term.");
                }
                v = (SortedSchemaVariable) ts;
                bound = false;
            })  
	(  
        a1=logicTermReEntry
        ( "waryAll" { waryAll = true; } |
            "waryEx"  { waryEx  = true; } )?
        RBRACE
        { 
            if (bound) {v = bindVar(v1,s);} 
        } 
        a2=term60
        { 
            if ( waryAll ) {
                result = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
                                                     v, a1, a2 );
            } else if ( waryEx ) {
                result = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
                                                     v, a1, a2 );
            } else {
                result = tf.createSubstitutionTerm ( Op.SUBST,
                                                     v, a1, a2 );
            }
            if (bound) {unbindVars();}
        }

        //For OCL simplification
        |
            SLASH a1=logicTermReEntry
            { 
                if (bound) {v = bindVar(v1,s);} 
            }
            (
                COMMA v2:IDENT
                (
                    COLON s=sortId 
                    {bound = true;} 

                | 

                    { functions().lookup(new Name(v2.getText())) == null }?				
                    {
                        ts = (TermSymbol) variables().lookup(new Name(v2.getText()));
                        if (   ts == null
                        || ts instanceof LogicVariable ) {
                            throw new SemanticException("Substituted variables need a sort.");
                        }
                        if ( ! (ts instanceof SchemaVariable &&
                                ((SchemaVariable)ts).isVariableSV()) ) {
                            throw new SemanticException(
                                                "	"+ts+" is not allowed as bound variable "
                                                +"in a substitution term.");
                        }
                        vv = (SortedSchemaVariable) ts;
                        bound = false;
                    }
                )
                SLASH a2=logicTermReEntry
                { 
                    if (bound) {vv = bindVar(v2, s);} 
                }
            )?
            RBRACE
            a3=term60
            {
                if (vv == null)
                {
                   result = tf.createSubstitutionTerm (Op.SUBST, v, a1, a3);
                }
                else
                {
                    result = tf.createTernarySubstTerm(Op.SUBST, v, vv, 
                        new Term[]{a1, a2, a3});
                }

                if (bound) {unbindVars();}
            }
        //
        )
    ;

// inherited from grammar TacletParser
shortCond[Term cond1, Term cond2] returns [Term result = null]{
  Term c_then = null;
  ParsableVariable v = null;
}
:DOT v = varId CONDUP c_then = logicTermReEntry {
            if (v instanceof SortedSchemaVariable) {
                result = tf.createIfElseTerm((SortedSchemaVariable)v, cond1, 
                                        cond2, c_then);
            } else {
                result = tf.createIfElseTerm((ProgramVariable)v, cond1, 
                                             cond2, c_then);
            }
        }
    ;

// inherited from grammar TacletParser
specialTerm returns [Term result = null]{
    Operator vf = null;
}
:result = metaTerm
    |
        vf = expr_op 
            {   
                result = tf.createFunctionTerm((Function)vf, AN_ARRAY_OF_TERMS);
            }                 
    ;

// inherited from grammar TermParser
formula returns [Term a = null]:a = term 
        {
            if ( a.sort() != Sort.FORMULA ) {
                throw new antlr.SemanticException
                ("Just Parsed a Term where a Formula was expected");
            }
        }
    ;

// inherited from grammar TermParser
term returns [Term a = null]:a=term20 
    ;

// inherited from grammar TermParser
term20 returns [Term a = null]{
    Term a1;
}
:a=term30 
        (EQV a1=term20 
            { a = tf.createJunctorTerm(Op.EQV, new Term[]{a, a1});} ) ?;

// inherited from grammar TermParser
term30 returns [Term a = null]{
    Term a1;
}
:a=term40 
        (IMP a1=term30 
            { a = tf.createJunctorTerm(Op.IMP, new Term[]{a, a1});} ) ?;

// inherited from grammar TermParser
term40 returns [Term a = null]{
    Term a1;
}
:a=term50 
        (OR a1=term40 
            { a = tf.createJunctorTerm(Op.OR, new Term[]{a, a1});} ) ?;

// inherited from grammar TermParser
term50 returns [Term a = null]{
    Term a1;
}
:a=term60 
        (AND a1=term50
            { a = tf.createJunctorTerm(Op.AND, new Term[]{a, a1});} ) ?
    ;

// inherited from grammar TermParser
term60 returns [Term a = null]{
    Term a1,a2;
    Sort s;
}
:NOT a1  = term60 { a = tf.createJunctorTerm(Op.NOT,new Term[]{a1}); }
    |   CTRL a1 = term60 { a = tf.createJunctorTerm(Op.CTRL,new Term[]{a1}); }
    |   a = bindingterm 
    |   a = dlTermOrTerm70
    ;

// inherited from grammar TermParser
dlTermOrTerm70 returns [Term a = null]{}
:a = normal_dl_term
    |  (IDENT MODALITYARG)=> a = named_dl_term
    |  a = term70
    ;

// inherited from grammar TermParser
term70 returns [Term a = null]{
    Term a1;
}
:a = accessterm
        // a term like {o:=u}x=y is parsed as {o:=u}(x=y)
        ((EQUALS)=>EQUALS a1 = logicTermReEntry
            { a = tf.createEqualityTerm(a, a1);}) ?

    ;

// inherited from grammar TermParser
logicTermReEntry returns [Term result = null]{}
:(
            result = accessterm |
            result = update_or_substitution
        ) 
        {
            if (result.sort() == Sort.FORMULA) {
                throw 
                    new KeYSemanticException
                    ("Only terms may stand here.", getLine(), getColumn(), 
                     getFilename());
            }
        }          
    ;

// inherited from grammar TermParser
classReference returns [String classReference = ""]{}
:id:IDENT {
            classReference = id.getText(); 
            while (isPackage(classReference)) {
                match(DOT);
                classReference += "." + LT(1).getText();
                match(IDENT);
            }      
            
            if (getJavaInfo().getTypeByClassName(classReference) == null) {
                throw new NotDeclException
                    ("class " + classReference + " is unknown.", 
                     classReference, getFilename(), getLine(), 
                     getColumn());
            }
        }  
    ;

// inherited from grammar TermParser
static_attribute_suffix[String className] returns [Term result = null]{
    TermSymbol v = null;
    int shadowPrimes = 0;    
    String attributeName = "";
}
:DOT attributeName = attrid
        {   v = getAttribute(getTypeByClassName(className).
                             getSort(), attributeName); }   

        ((ps:PRIMES 
                { shadowPrimes = ps.getText().length();})?
            {
                result = createAttributeTerm(null, v, shadowPrimes);
            }
        )
    ;

// inherited from grammar TermParser
attribute_or_query_suffix[Term prefix] returns [Term result = null]{
    TermSymbol v = null;
    int shadowPrimes = 0;
    result = prefix;
    String attributeName = "";    
}
:DOT 
        ((id LPAREN)=>( result = query[prefix])
        | 
            attributeName = attrid 
            {   v = getAttribute(prefix.sort(), attributeName); }   
            ((ps:PRIMES 
                    { shadowPrimes = ps.getText().length();})?

                {
                    result = createAttributeTerm(prefix, v, shadowPrimes);
                }        
            ))
    ;

// inherited from grammar TermParser
query[Term prefix] returns [Term result = null]{
    String classRef = "";
    Term[] args;
    TermWithBoundVars[] argsWithBoundVars = new TermWithBoundVars[0];
}
:classRef = classReference COLON COLON mid:IDENT argsWithBoundVars = argument_list
        { 
            args = new Term[argsWithBoundVars.length];
            for (int i=0; i<argsWithBoundVars.length; i++) {
                args[i] = argsWithBoundVars[i].term;
            }
            try {
                result = getServices().getJavaInfo().getProgramMethodTerm
                (prefix, mid.getText(), args, classRef);
            } catch (IllegalArgumentException e) {
                throw new antlr.SemanticException(e.getMessage());
            }
        };

// inherited from grammar TermParser
accessterm returns [Term result = null]{
    KeYJavaType kjt = null;
    Sort arraySort = null;
    Term indexTerm = null;
    int shadowPrimes = 0;
    String className = null;
}
:(LPAREN logicTermReEntry (COMMA | GUARD)) => result = conditional
    |
        {"-".equals(LT(1).getText()) && 
	!TermParser.isNumber(LT(2).getText())}? neg:ARITH_OP result = logicTermReEntry
        {
            if (result.sort() != Sort.FORMULA) {
                result = tf.createFunctionTerm
                ((Function) functions().lookup(new Name("~m")), result);
            } else {
                throw new KeYSemanticException
                ("Formula cannot be prefixed with: ", neg, getFilename());
            }
        } 
    | ({LA(2) == DOT && ( isPackage ( LT(1).getText() ) || 
                ( kjt = getTypeByClassName(LT(1).getText()))!=null )}?

        className = classReference
        result = static_attribute_suffix[className]
      | 	
        result = term80)
        (LBRACKET indexTerm = logicTermReEntry
            ( RBRACKET
            | rb:RBRACKETPRIMES {
                    shadowPrimes = rb.getText().length()-1;                 
                }
            )
            {
                if (arraySort != null && 
                    arraySort != ProgramSVSort.VARIABLE) {
                    if (!(arraySort instanceof ArraySort)){
                        throw new NotDeclException
                            ("array "+result+" declared with a different"+
                             " dimension than required here", ""+result, 
                             getFilename(), getLine(), getColumn());
                    } else {
                        arraySort = ((ArraySort) arraySort).elementSort();
                    } 
                } else if (arraySort == ProgramSVSort.VARIABLE && !inSchemaMode()) {
                    throw new antlr.SemanticException
                        ("Schemavariable are not allowed in this context.");
                }
                if (shadowPrimes>0) {
                    result = tf.createArrayTerm
                        (Op.SHADOW_ARRAY, result, indexTerm, 
                         TermParser.toZNotation("" + shadowPrimes, functions(), tf));
                } else {
                    result = tf.createArrayTerm(Op.ARRAY, result, indexTerm);
                }   
            }
        | result = attribute_or_query_suffix[result]
        )*
    ;

// inherited from grammar TermParser
term80 returns [Term a = null]:a = funcpredvarterm
    |   a = specialTerm
    |   LPAREN a = term RPAREN 
    |   "true"  { a = tf.createJunctorTerm(Op.TRUE); }
    |   "false" { a = tf.createJunctorTerm(Op.FALSE); }
    ;

// inherited from grammar TermParser
abbreviation returns [Term a=null]{ 
    String sc = null;
}
:(   sc = id
            {
                a =  scm.getTerm(sc);
                if(a==null){
                    throw new NotDeclException
                        ("abbreviation", sc, 
                         getFilename(), getLine(), getColumn());
                }                                
            }
        )
    ;

// inherited from grammar TermParser
conditional returns [Term result = null]{ 
    Term comparator = null;
}
:LPAREN comparator = logicTermReEntry 
        ( result=normalConditional[comparator] | result=conjunctionConditional[comparator])
    ;

// inherited from grammar TermParser
normalConditional[Term cond1] returns [Term result = null]{ 
    Term cond2 = null;
    Term c_then = null;
    Term c_else = null;
}
:GUARD cond2 = logicTermReEntry RPAREN
        (
            DOT result = shortCond[cond1, cond2]
        | 
            c_then = logicTermReEntry COLON c_else = logicTermReEntry {
                result = tf.createIfElseTerm(cond1, cond2, c_then, c_else);
            }

        ) 
    ;

// inherited from grammar TermParser
conjunctionConditional[Term cond1_1] returns [Term a = null]{ 
    Term cond1_2 = null;
    Term cond2_1 = null;
    Term cond2_2 = null;
    Term c_then = null;
    Term c_else = null;
}
:COMMA cond1_2=logicTermReEntry COLON c_else=logicTermReEntry 
        GUARD cond2_1 = logicTermReEntry 
        COMMA cond2_2=logicTermReEntry COLON c_then=logicTermReEntry 
        RPAREN
        {
            a = tf.createConjCondTerm(new Term[]{cond1_1, cond1_2, c_else, 
                                                 cond2_1, cond2_2, c_then});
        }        
    ;

// inherited from grammar TermParser
local_bindingterm returns [Term a = null]{
    Operator op = null;
    LogicVariable v = null;
    Term a1;
    Sort s;
}
:(
            "all" { op = Op.ALL; }
        |   "ex"  { op = Op.EX;  }
        )
        v0:IDENT COLON s=sortId DOT {v = bindVar(v0,s);} a1 = term60
        {
            a = tf.createQuantifierTerm((Quantifier)op, v,a1);
            unbindVars();
        }
    | a = update_or_substitution
    ;

// inherited from grammar TermParser
updateterm returns [Term result = null]{ 
    SingleUpdateData sud = null;
    List subs = new LinkedList();  
    Term a2 = null;
}
:LBRACE
        sud = singleupdate {
            subs.add(sud.a0); 
            subs.add(sud.a1); 
        } ( COMMA sud = singleupdate {
                subs.add(sud.a0); 
                subs.add(sud.a1); 
            } )* 
        RBRACE a2 = term60 {
            subs.add(a2);
        }
        { 
            result = tf.createUpdateTerm((Term[])subs.toArray(new Term[subs.size()]));
        }
    ;

// inherited from grammar TermParser
singleupdate returns [SingleUpdateData sud=null]{
    Term a0 = null;
    Term a1 = null;
    ArraySort as;
    TermSymbol pv;
    LhsUpdateData lud = null;
    boolean shadow_l = false;
    boolean shadow_r = false;
}
:(a0 = lhsSingle ) ASSIGN (a1 = rhsSingle) {
            sud = new SingleUpdateData();
            sud.a0 = a0;
            sud.a1 = a1;
        }
    ;

// inherited from grammar TermParser
rhsSingle returns [Term result = null]{}
:result = logicTermReEntry
    ;

// inherited from grammar TermParser
lhsSingle returns [Term result = null]{
    Term a1 = null;
}
:result = logicTermReEntry
        { 
            if (!(result.sort() instanceof ProgramSVSort) &&
                result.op() != Op.META_SHADOW &&
                !(result.op() instanceof ProgramVariable || 
                  result.op() instanceof AccessOp)) {
                throw new KeYSemanticException
                ("Term of sort " + result.sort() + 
                 " cannot stand for a location.",
                 getLine(), getColumn(), getFilename());
            }
        }
    ;

// inherited from grammar TermParser
normal_dl_term returns [Term a = null]{   JavaBlock jb=null;
    Term a1;
}
:d1:DIAMOND 
        {
            jb=getJavaBlock(d1);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60
        {
            a = tf.createDiamondTerm(jb, a1);
            unbindProgVars();
        }
    |
        d2:BOX
        {
            jb=getJavaBlock(d2);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60   
        {
            a = tf.createBoxTerm(jb, a1);
            unbindProgVars();
        }
    |
        d3:THROUGHOUT
        {
            jb=getJavaBlock(d3);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60
        {
            a = tf.createProgramTerm(Op.TOUT,jb, a1);
            unbindProgVars();
        }
    ;

// inherited from grammar TermParser
named_dl_term returns [Term a = null]{
    TermWithBoundVars[] argsWithBoundVars = new TermWithBoundVars[0];
    Term[] terms = null;
    Term term = null;
    JavaBlock jb = null;
}
:id:IDENT
        bl:MODALITYARG ( 
            {
                jb = getJavaBlock(bl);               
                bindProgVars(progVars(jb));
            }
            argsWithBoundVars = argument_list
            {   
                unbindProgVars();
                terms = new Term[argsWithBoundVars.length];
                for (int i=0; i<argsWithBoundVars.length; i++) {
                    terms[i] = argsWithBoundVars[i].term;
                }
                Operator op = null;
                                if (id.getText().charAt(0)=='#') {
                    if (!inSchemaMode()) {
                        throw new KeYSemanticException
                            ("No schema elements allowed outside taclet declarations.",
                             id, getFilename());
                    }
                    op = (SchemaVariable)variables().lookup(new Name(id.getText()));
                } else {
                    op = Op.getModality(id.getText());
                }
                if(op == null) {
                    throw new KeYSemanticException("Unknown modal operator.", id, getFilename());
                }
                if(op.arity() != terms.length) {
                    throw new KeYSemanticException
                        ("The arity of " + op.name() +
                         " is "+op.arity() + ", but you provided " + terms.length + " arguments",
                         getLine(), getColumn(), getFilename());
                }
                a = tf.createProgramTerm(op, jb, terms);
            }
        )
    ;

// inherited from grammar TermParser
argument_list returns [TermWithBoundVars[] ts = null]{
    List args = new LinkedList();
    TermWithBoundVars p1, p2;
}
:LPAREN 
        (p1 = argument { args.add(p1);  }

            (COMMA p2 = argument { args.add(p2); } )* )?

        RPAREN
        {
            ts = (TermWithBoundVars[]) args.toArray(new TermWithBoundVars[args.size()]);
        }

    ;

// inherited from grammar TermParser
funcpredvarterm returns [Term a = null]{
    Term[] arguments = null;
    TermWithBoundVars[] argsWithBoundVars = null;
    String varfuncid;
    String neg = "";
}
:ch:CHAR_LITERAL {
            String s = ch.getText();
            int intVal = 0;
            if (s.length()==3) {
                intVal = (int)s.charAt(1);
            } else {
                try {
                    intVal = Integer.parseInt(s.substring(3,s.length()-1),16);
                } catch (NumberFormatException ex) {
                    throw new KeYSemanticException
                        ("Expected '"+s+"' to be a character", ch, getFilename());
                }       
            }
            a = tf.createFunctionTerm((Function) functions().lookup(new Name("C")), 
                                      TermParser.toZNotation(""+intVal, functions(), tf).sub(0));
        }
    | { ("-".equals(LT(1).getText()) && TermParser.isNumber(LT(2).getText())) 
	|| 
        (TermParser.isNumber(LT(1).getText()) && !"(".equals(LT(2).getText()))}? 
	({"-".equals(LT(1).getText())}? ARITH_OP {neg = "-";})?
         number:IDENT     
        { a = TermParser.toZNotation(neg+number.getText(), functions(), tf);}    
    | AT a = abbreviation
    | varfuncid = id

        ((LPAREN)=>argsWithBoundVars = argument_list)? 
        //argsWithBoundVars==null indicates no argument list
        //argsWithBoundVars.length==0 indicates open-close-parens ()
        {   
            ArrayOfQuantifiableVariable vars = null;
            if (argsWithBoundVars != null) {
                arguments = new Term[argsWithBoundVars.length];           
                for (int i=0; i<argsWithBoundVars.length; i++) {
                    arguments[i] = argsWithBoundVars[i].term;

                    //For OCL simplification
                    if (argsWithBoundVars[i].boundVars != null) {
                        if ((i == 1 && argsWithBoundVars[i].boundVars.size() == 1)
                            || (i == 2 && argsWithBoundVars[i].boundVars.size() == 2)) {
                            vars = argsWithBoundVars[i].boundVars;
                        } else {
                            throw new antlr.SemanticException("Parsing error: The " + varfuncid + 
                                " expression does not match any OCL collection operation!");
                        }
                    }
                    //
                }
            }
            Operator op = lookupVarfuncId(varfuncid, arguments);
            if (op instanceof ParsableVariable) {
                a = termForParsedVariable((ParsableVariable)op);
            } else  if (op instanceof TermSymbol) {
                if (arguments==null) {
                    arguments = AN_ARRAY_OF_TERMS;
                }
                if (vars == null) {
                    a = tf.createFunctionTerm((TermSymbol)op, arguments);
                }

                //For OCL simplification
                else {
                    if (vars.size() == 2) {
                        a = tf.createTernaryBindingTerm(
                                                        (TermSymbol)op, 
                                                        vars.getQuantifiableVariable(0), 
                                                        vars.getQuantifiableVariable(1), 
                                                        arguments);
                    } else {
                        a = tf.createBinaryBindingTerm(
                                                       (TermSymbol)op, 
                                                       vars.getQuantifiableVariable(0),  
                                                       arguments);
                    }
                } 
                //
            }
        }
    ;

// inherited from grammar TermParser
attrid returns [String attr = "";]{}
:({ isPackage(LT(1).getText()) || (LA(2) == COLON && LA(3) == COLON) }? 
            attr = classReference COLON COLON {attr += "::";} )?
        attrID:IDENT {attr += attrID.getText();}
    ;

// inherited from grammar TermParser
expr_op returns [Operator v = null]{
    ParsableVariable left_var0 = null;
    ParsableVariable right_var0 = null;
    ParsableVariable unary_var0 = null;
}
:IN_TYPE LPAREN ((left_var0=varId ((id:ARITH_OP | id1:SLASH | id2:IDENT) right_var0 = varId)?) |
            ( "-" unary_var0 = varId) )  RPAREN
        {  
            if (id==null) id=id1;   
            if (unary_var0!=null) {
                // can only be unary minus
                SortedSchemaVariable unary_var = (SortedSchemaVariable) unary_var0;
                Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
                return new InTypeOperator
                    (type, new de.uka.ilkd.key.java.expression.operator.Negative(unary_var));
            }
            else {
                SortedSchemaVariable left_var = (SortedSchemaVariable) left_var0;
                Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
                if (right_var0 != null) {
                    SortedSchemaVariable right_var = (SortedSchemaVariable) right_var0;
                    String op_str = (id == null ? id2.getText() : id.getText());
                    if ("+".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Plus(left_var, right_var));
                    } else if ("-".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Minus(left_var, right_var));
                    } else if ("*".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Times(left_var, right_var));
                    } else if ("/".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Divide(left_var, right_var));
                    } else if ("%".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Modulo(left_var, right_var));
                    } 
                }
                else {
                    return new InTypeOperator(type, left_var);
                }
            }
        }
    | IS_OBJECT  LPAREN (left_var0=varId )  RPAREN
        {
            SortedSchemaVariable var = (SortedSchemaVariable) left_var0;
            return new IsObject(de.uka.ilkd.key.logic.sort.Sort.FORMULA, var);
        }
    ;

// inherited from grammar TermParser
varId returns [ParsableVariable v = null]:id:IDENT 
        {   
            v = (ParsableVariable) variables().lookup(new Name(id.getText()));
            if (v == null) {
                throw new NotDeclException("variable", id, 
                                           getFilename());
            }
        }
    ;

// inherited from grammar DeclParser
decls :(INCLUDE 
            ( LDTS COLON LBRACE include[true] (COMMA include[true])* RBRACE
            |
                include[false]
            )
            SEMI)*
        {
            if(parse_includes) return;
            activatedChoices = SetAsListOfChoice.EMPTY_SET;  
        }
        ("with" LPAREN activated_choice (COMMA activated_choice)* RPAREN)?
        {if(onlyWith) return;}
        (
            option_decls
        |    
            sort_decls
        |
            prog_var_decls
        |
            schema_var_decls
        |
            pred_decls
        |
            func_decls
        |
            heur_decls

        ) *
    ;

// inherited from grammar DeclParser
include[boolean ldt] :(absfile:IDENT{ 
                if(parse_includes){
                    addInclude(absfile.getText(),false,ldt);
                }
            }
        | relfile:STRING_LITERAL{ 
                if(parse_includes){
                    addInclude(relfile.getText(),true,ldt);
                }
            })
    ;

// inherited from grammar DeclParser
activated_choice {
    String name;
    Choice c;
}
:cat:IDENT COLON choice:IDENT
        {if(usedChoiceCategories.contains(cat.getText())){
            throw new IllegalArgumentException("You have already chosen a different option for "+cat.getText());
        }
        usedChoiceCategories.add(cat.getText());
        name = cat.getText()+":"+choice.getText();
        c = (Choice) choices().lookup(new Name(name));
        if(c==null){
            throw new NotDeclException("Option", choice,
                                       getFilename());
        }else{
            activatedChoices=activatedChoices.add(c);
            namespaces().setFunctions(functions().extended(c.funcNS().allElements()));
        }
        }
    ;

// inherited from grammar DeclParser
option_decls {
}
:"options" LBRACE choice (SEMI choice)* RBRACE 
    ;

// inherited from grammar DeclParser
choice {
    String cat=null;
}
:category:IDENT {cat=category.getText();} (COLON LBRACE choice_option[cat] (COMMA choice_option[cat])* RBRACE)? 
        {
            if(!category2Default.containsKey(cat)){
                choices().add(new Choice("On",cat));
                choices().add(new Choice("Off",cat)); 
                category2Default.put(cat, cat+":On");               
            }
        }
    ;

// inherited from grammar DeclParser
choice_option[String cat] {
    String name;
}
:choice:IDENT { name=cat+":"+choice.getText();
        Choice c = (Choice) choices().lookup(new Name(name));
        if(c==null){
            c = new Choice(choice.getText(),cat);
            choices().add(c);
        }
        if(!category2Default.containsKey(cat)){
            category2Default.put(cat, name);
        }
        }
    ;

// inherited from grammar DeclParser
sort_decls {
    boolean isObjectSort=false;
    boolean isGenericSort=false;

    //For OCL simplification
    boolean isOclSort = false;

    Sort[] sortExt=new Sort [0];
    Sort[] sortOneOf=new Sort [0];
    ListOfName sortIds;
}
:"sorts" 
        LBRACE ( 
            { 
                isObjectSort=false; 
                isGenericSort=false; 
                isOclSort = false;
                sortExt = new Sort[0];
                sortOneOf = new Sort[0];
            }
            ( "object" {isObjectSort=true;} sortIds = sortIdentifiers
            | "generic" {isGenericSort=true;} sortIds = sortIdentifiers
                ( "oneof" sortOneOf = oneof_sorts )? 
                ("extends" sortExt = extends_sorts )?

            | "ocl" {isOclSort=true;} sortIds = sortIdentifiers

            | sortIds = sortIdentifiers

            )
            SEMI
            {   
                if (!skip_sorts) {
                    Name sort_name;
                    IteratorOfName it = sortIds.iterator ();
                    while ( it.hasNext () ) {
                        sort_name = it.next ();

                        if (sorts().lookup(sort_name) == null) {
                            Sort s;
                            if (isObjectSort) {
                                if ("Object".equals(sort_name)) {
                                    if (sorts().lookup(new Name("Object")) == null) {
                                        s = new ClassInstanceSortImpl(sort_name);
                                    }
                                    s=(Sort)sorts().lookup(new Name("Object"));
                                } else {
                                    s = new ClassInstanceSortImpl(sort_name, 
                                                                  (Sort)sorts().lookup
                                                                  (new Name("Object")));
                                }	
                            }                            
                            //Just constant for now
                            else if (isOclSort) {
                               s = Sort.OCL_ANY;
                            }
                            //
                            else if (isGenericSort) {
                                int        i;
                                SetOfSort  ext   = SetAsListOfSort.EMPTY_SET;
                                SetOfSort  oneOf = SetAsListOfSort.EMPTY_SET;

                                for ( i = 0; i != sortExt.length; ++i )
                                    ext = ext.add ( sortExt[i] );

                                for ( i = 0; i != sortOneOf.length; ++i )
                                    oneOf = oneOf.add ( sortOneOf[i] );

                                try {
                                    s = new GenericSort ( sort_name,
                                                          ext,
                                                          oneOf );
                                } catch ( GenericSupersortException e ) {
                                    throw new GenericSortException ( "sort",
                                                                     "Illegal sort given", e.getIllegalSort (),
                                                                     getFilename (), getLine (), getColumn () );
                                }
                            } else {
                                s = new PrimitiveSort(sort_name);
                            }
                            sorts().add(s);
                            if (s instanceof NonCollectionSort) {
                                NonCollectionSort ns
                                    = (NonCollectionSort)s;
                                sorts().add(ns.getSetSort());
                                sorts().add(ns.getSequenceSort());
                                sorts().add(ns.getBagSort());
                            }
                            if ( s instanceof SortDefiningSymbols )
                                ((SortDefiningSymbols)s).createSymbols ( defaultChoice.funcNS(),
                                                                         sorts() );
                        }

                    }
                }
            }
            
        )* 
        RBRACE
    ;

// inherited from grammar DeclParser
sortIdentifiers returns [ListOfName ids = SLListOfName.EMPTY_LIST]:id:IDENT { ids = ids.append ( new Name(id.getText()) ); }
        (
            COMMA
            id2:IDENT { ids = ids.append ( new Name(id2.getText()) ); }
        ) *
    ;

// inherited from grammar DeclParser
extends_sorts returns [Sort[] extendsSorts = null]{
    List args = new LinkedList();
    Sort s;
}
:s = any_sortId_check[true] { args.add(s); }
        (
            COMMA s = any_sortId_check[true] {args.add(s);}
        ) *
        {
            extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

// inherited from grammar DeclParser
oneof_sorts returns [Sort[] oneOfSorts = null]{
    List args = new LinkedList();
    Sort s;
}
:LBRACE
        s = sortId_check[true] { args.add(s); }
        (
            COMMA s = sortId_check[true] {args.add(s);}
        ) *
        RBRACE
        {
            oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

// inherited from grammar DeclParser
prog_var_decls {
    String var_name;
    String type = null;
}
:{ switchToNormalMode(); }
        "program" "variables"
        LBRACE 
        (
            ((v0:IDENT {type= "" + v0.getText();} (DOT v1:IDENT {type= type + "." + v1.getText();})*)
              |
             (pid:PIDENT {type = pid.getText();} )
            )
            (LBRACKET RBRACKET {
               type = type + "[]";
            })?
            //sortId_check[true]
            var_name = id
            {
                KeYJavaType kjt = getJavaInfo().getKeYJavaType(type);
		ProgramElementName pvName = new ProgramElementName(var_name);
                if (kjt == null) {
                    throw new antlr.SemanticException
                        ("Program Variable " + var_name + 
                         " declared with unknown type " + 
                         type);

                } else if (lookup(pvName)!=null) {
		  // commented out as pv do not have unique name (at the moment)
		    //throw new AmbigiousDeclException
		//	(var_name, getFilename(), getLine(), getColumn()); 
		}
            namespaces().programVariables().add(new ProgramVariable
                (new ProgramElementName(var_name), kjt));

            }
            SEMI
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
schema_var_decls {
    Sort s = null;
    boolean makeVariableSV  = false;
    boolean makeDependingSV = false; // skolem functions
    boolean rigid           = false;
    boolean listModifier    = false;
    String type = "";    
}
:"schema" "variables"
        LBRACE  { switchToSchemaMode(); }
        (
            ( "modal" "operator" one_schema_modal_decl (COMMA one_schema_modal_decl)* SEMI )
        |
            {makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;}
            ( ( "program" ("svlist" {listModifier=true;})? {type = "";}
                ((IDENT IDENT) => id1:IDENT{type += " "+id1.getText();}
                 | id2:IDENT{ type = type.substring(1,type.length());
                              s = (Sort) DeclParser.name2sort.get(type);
                              if (s == null) {
                                  throw new antlr.SemanticException
                                  ("SchemaVariable of sort "+type+ " not found.");
                              }
                              schema_var_decl(id2.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);} 
                )+
                (COMMA id3:IDENT{schema_var_decl(id3.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);})* SEMI
               )
               | "formula"
                  ( "rigid" { rigid = true; } ) ? {s = Sort.FORMULA;type = ""; }
		  (id5:IDENT {schema_var_decl(id5.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);})
		  ( COMMA (id6:IDENT {schema_var_decl(id6.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);}))* SEMI
               | s = any_sortId_check[true]
                  ( "variables" { makeVariableSV = true; }
                    | "depending" { makeDependingSV = true; }
                    | "rigid" { rigid = true; } ) ? {type = "";}
                  (id7:IDENT {schema_var_decl(id7.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);})
	          ( COMMA (id8:IDENT {schema_var_decl(id8.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);}))* SEMI
             )
        ) *
        { switchToNormalMode(); }
        RBRACE
    ;

// inherited from grammar DeclParser
one_schema_modal_decl {
    Modality[] mods;
}
:id:IDENT EQUALS mods=modalities_decl {
            SchemaVariable msv = (SchemaVariable)variables().lookup(new Name(id.getText()));
            if(msv != null)
            throw new antlr.SemanticException("Schema variable "+id.getText()+ " already defined");
            int arity = mods[0].arity();
            for(int i=1;i<mods.length;i++)
            if(arity != mods[i].arity())
            throw new antlr.SemanticException("Arity mismatch for schema variable "+id.getText());
            msv = SchemaVariableFactory.createModalOperatorSV(new Name(id.getText()), arity, mods);
            if (inSchemaMode()) {
                variables().add(msv);
            }
        }
    ;

// inherited from grammar DeclParser
modalities_decl returns [Modality[] modalities = null]{
    List args = new LinkedList();
}
:(LBRACE 
            s1:IDENT {
                if(s1.getText().charAt(0) == '#') {
                    ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s1.getText()));
                    if(msv == null)
                        throw new antlr.SemanticException("Schema variable "+s1.getText()+ " not defined");
                    for(int i=0; i<msv.modalities().length; i++)
                        args.add(msv.modalities()[i]);
                }else{
                    Modality op = Op.getModality(s1.getText());
                    if(op == null)
                        throw new antlr.SemanticException("Unrecognised modal operator "+s1.getText());
                    args.add(op);
                }
            } (COMMA s2:IDENT {
                    if(s2.getText().charAt(0) == '#') {
                        ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s2.getText()));
                        if(msv == null)
                            throw new antlr.SemanticException("Schema variable "+s2.getText()+ " not defined");
                        for(int i=0; i<msv.modalities().length; i++)
                            args.add(msv.modalities()[i]);
                    } else {
                        Modality op = Op.getModality(s2.getText());
                        if(op == null)
                            throw new antlr.SemanticException("Unrecognised modal operator "+s2.getText());
                        args.add(op);
                    }
                })* RBRACE
        ) {
            modalities = (Modality[])args.toArray(new Modality[args.size()]);
        }
    ;

// inherited from grammar DeclParser
pred_decls {
    Sort[] argSorts;
    String pred_name;
    Namespace oldFNS=null;
    String ctext = null;
    Choice c=null;
}
:"predicates" (LPAREN cat:IDENT COLON choice:IDENT RPAREN)?
        { if(cat!=null){
            ctext = cat.getText()+":"+choice.getText();
            c = (Choice) choices().lookup(new Name(ctext));
            if(c==null){
                throw new NotDeclException("Option", choice,
                                           getFilename());
            }else{
                namespaces().setFunctions(c.funcNS());
            }
        }else{
            namespaces().setFunctions(defaultChoice.funcNS());
        }
        cat=null;
        }
        LBRACE
        (
            pred_name = id
            argSorts=arg_sorts[!skip_predicates]
            {
                if (!skip_predicates) {
                    Name predicate = new Name(pred_name);		    		    
                    if (lookup(predicate) != null) {
                        throw new AmbigiousDeclException
                        (pred_name, getFilename(), getLine(), getColumn());
                    } else {
                        Function p = new Function(predicate,
                                                  Sort.FORMULA, 
                                                  argSorts);
                        functions().add(p);         
                    }
                } 
            }
            SEMI
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
func_decls {
    Sort[] argSorts;
    Sort retSort;
    String func_name;
    Namespace oldFNS=null;
    String ctext = null;
    Choice c=null;
}
:"functions" (LPAREN cat:IDENT COLON choice:IDENT RPAREN)?
        { 
            if(cat!=null){
                ctext = cat.getText()+":"+choice.getText();
                c = (Choice) choices().lookup(new Name(ctext));
                if(c==null){
                    throw new NotDeclException("Option", choice,
                                               getFilename());
                }
            }else{
                c=defaultChoice;
            }
            namespaces().setFunctions(c.funcNS());
            cat=null;
        }
        LBRACE
        (
            retSort = sortId_check[!skip_functions]
            func_name = id 
            argSorts = arg_sorts[!skip_functions]
            {
                if (!skip_functions) {
                    Name fct_name = new Name(func_name);
                    if (lookup(fct_name) != null) {
                        throw new AmbigiousDeclException
                        (func_name, getFilename(), getLine(), getColumn());
                    } else {
                        functions().add(new Function(fct_name,
     	                                             retSort, 
        	                                     argSorts));
                    }
                } 
            }
            SEMI
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
arg_sorts[boolean checkSort] returns [Sort[] argSorts = null]{
    List args = new LinkedList();
    Sort s;
}
:(
            LPAREN
            s = sortId_check[checkSort] { args.add(s); }
            (
                COMMA s = sortId_check[checkSort] {args.add(s);}
            ) *
            RPAREN
        ) ?
        {
            argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

// inherited from grammar DeclParser
integer_num returns [ int num = 0 ]
options {
	defaultErrorHandler=false;
}
:{
            try {
                Token t = LT(1);
                num = Integer.decode ( t.getText () ).intValue ();
                consume ();
            } catch ( TokenStreamException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            } catch ( NumberFormatException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            } catch ( NullPointerException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            }
        }
    ;

// inherited from grammar DeclParser
heur_decls {
    int     t_priority = 0;
    Integer priority   = null;
}
:"heuristics"
        LBRACE
        (
            { priority = null; }
            id:IDENT
            (
                LPAREN t_priority = integer_num RPAREN 
                { priority = new Integer ( t_priority ); }
            ) ?
            SEMI
            { 
                if (!skip_heuristics) {
                    Heuristic h = new Heuristic(new Name(id.getText()));
                    if(heuristics().lookup(new Name(id.getText()))==null){
                        heuristics().add(h);
                    }
                    if ( priority != null )
                        heur_prioritization.setPriority ( h, priority.intValue () );
                }
            }
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
sortId returns [Sort s = null]:s = sortId_check[true]
    ;

// inherited from grammar DeclParser
sortId_check[boolean checkSort] returns [Sort s = null]{
    Sort t;
}
:t = sortId_check_help[checkSort]
        s = array_decls[t]
    ;

// inherited from grammar DeclParser
any_sortId_check[boolean checkSort] returns [Sort s = null]{
    Sort t;
}
:t = any_sortId_check_help[checkSort]
        s = array_decls[t]
    ;

// inherited from grammar DeclParser
any_sortId_check_help[boolean checkSort] returns [Sort s = null]{
    String name;
}
:name = id
        {   s= (Sort) sorts().lookup(new Name(name));
            if (checkSort && s == null) {
                throw new NotDeclException("sort", name, 
                                           getFilename(), 
                                           getLine(), getColumn());
            }
        }
    ;

// inherited from grammar DeclParser
sortId_check_help[boolean checkSort] returns [Sort s = null]:s = any_sortId_check_help[checkSort]
        {
            // don't allow generic sorts or collection sorts of
            // generic sorts at this point
            Sort t = s;
            while ( t instanceof CollectionSort )
            t = ((CollectionSort)t).elementSort ();
            if ( t instanceof GenericSort )
            throw new GenericSortException ( "sort",
                                             "Non-generic sort expected", t,
                                             getFilename (), getLine (), getColumn () );
        }
    ;

// inherited from grammar DeclParser
array_decls[Sort p] returns [Sort s = null]{
    int  n=0;
}
:(LBRACKET RBRACKET {n++;})*

        { 
            if(n != 0){
                s = ArraySortImpl.getArraySortForDim(
                    p, n, getJavaInfo().getObjectSort(),
                    getJavaInfo().getCloneableSort());
            }else{
                s = p;
            }
        }
            
    ;

// inherited from grammar DeclParser
id_declaration returns [ IdDeclaration idd = null ]{
    Sort s = null;
}
:id:IDENT
        ( COLON s = sortId_check[true] ) ?
        {
            idd = new IdDeclaration ( id.getText (), s );
        }
    ;

// inherited from grammar DeclParser
id returns [String name = null]:(prefix:IDENT 
	{ name = prefix.getText(); }
        | pprefix : PIDENT { name = pprefix.getText(); })
        ((COLON COLON)=>COLON COLON nameID:IDENT 
            { 
                name += "::" + nameID.getText();
            }
        )?
    ;


