header {

package de.uka.ilkd.key.parser;

import java.io.*;
import antlr.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.gui.AbbrevMap;

import de.uka.ilkd.key.rule.metaconstruct.*;
import de.uka.ilkd.key.java.JavaReader;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;

import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.ListOfKeYJavaType;
import de.uka.ilkd.key.java.abstraction.SLListOfKeYJavaType;
import java.util.List;
import java.util.LinkedList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.HashMap;
import de.uka.ilkd.key.java.visitor.DeclarationProgramVariableCollector;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
}
class TermParser extends Parser;

options {
	importVocab=KeYLexer;
	k= 1;
	defaultErrorHandler=false;
}

{
    
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected Choice defaultChoice = null;
    protected HashMap category2Default = new HashMap();
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();
    protected NamespaceSet nss;

    private Prioritization.StaticPrioritization heur_prioritization
        = new Prioritization.StaticPrioritization ();


    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    private boolean skip_schemavariables;
    private boolean skip_functions;
    private boolean skip_predicates;
    private boolean skip_sorts;
    private boolean skip_heuristics;

    private Services services;
    private AbbrevMap scm;

    private int lineOffset=0;
    private int colOffset=0;

    private TermFactory tf;
    private JavaReader javaReader;
    
    private boolean errors = false;

    public TermParser(TokenStream  lexer,                   
                      String       filename,
                      TermFactory  tf,
                      JavaReader   jr,
                      Services     services,
                      NamespaceSet nss,
                      AbbrevMap    scm) {
        this(lexer,2);
        setFilename(filename);
        this.javaReader = jr;
        this.nss = nss;
        this.tf  = tf;
        this.services = services;
        this.scm = scm;
        switchToNormalMode();
    }

    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine()+lineOffset;
        } catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    public JavaInfo getJavaInfo() {
        return getServices().getJavaInfo();
    }

    public Services getServices() {
        return services;
    }

    private Namespace sorts() {
        return nss.sorts();
    }

    private Namespace variables() {
        return nss.variables();
    }

    private Namespace programVariables() {
        return nss.programVariables();
    }

    private Namespace functions() {
        return nss.functions();
    }

    private Namespace heuristics() {
        return nss.heuristics();
    }

    private Namespace choices(){
        return nss.choices();
    }

    private NamespaceSet namespaces() {
        return nss;
    }

    public Named lookup(Name n) {
        final Namespace[] lookups = {
            functions(), variables(), 
            programVariables()
        };
        for (int i = 0; i<lookups.length; i++) {
            if (lookups[i].lookup(n) != null) {
                return lookups[i].lookup(n);
            }
        }
        return null;    
    }

    // a pure term parser is never in schemamode
    // only grammers that extend this grammar may 
    // activate the schema mode

    public void switchToSchemaMode() {
    }

    private boolean inSchemaMode() {
        return false;
    }

    public void switchToNormalMode() {
    }


    public static Term toZNotation(String number, Namespace functions, TermFactory tf){
	String s = number;
        final boolean negative = (s.charAt(0) == '-');
	if (negative) {
	    s = number.substring(1, s.length());
	}

        Term result = tf.createFunctionTerm((Function) functions.lookup(new Name("#")));

        for(int i = 0; i<s.length(); i++){
            result = tf.createFunctionTerm
                ((Function)functions.lookup
                 (new Name(s.substring(i,i+1))), result);
        }

       	if (negative) {
  	    result = tf.createFunctionTerm
		((Function) functions.lookup(new Name("neg")), result);
        }
	return tf.createFunctionTerm
            ((Function) functions.lookup(new Name("Z")), result); 
    }


    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }  

    protected int getColumn() {
        int     col = -1;
        try {
            col = LT(0).getColumn()+colOffset;
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }   

    protected LogicVariable bindVar(Token id,Sort s) {
        LogicVariable v=new LogicVariable(new Name(id.getText()), s);
        namespaces().setVariables(variables().extended(v));
        return v;
    }

    protected TermSymbol progVar(Token id, Sort s) throws NotDeclException{
        TermSymbol v = (TermSymbol)
            variables().lookup(new Name(id.getText()));

        if (v == null) {
            throw new NotDeclException("program variable", id,
                                       getFilename());
        }
        return v;
    }

    protected static boolean isNumber(String s){
        if (s==null){
           return false; 
        }
	
        for (int i=0; i<s.length(); i++){
           if ((s.charAt(i)<'0' || s.charAt(i)>'9') && 
		!(i==0 && s.charAt(i)=='-')) {
             return false;
           }
        }        
        return true;
    }

    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }

    protected boolean isPackage(String name){
        try {   
            return getJavaInfo().isPackage(name);
        } catch(RuntimeException e){        
            // may be thrown in cases of invalid java identifiers
            return false;
        } 
    }

    
    protected void unbindVars() {
        namespaces().setVariables(variables().parent());
    }

    protected void bindProgVars(HashSet progVars) {
        namespaces().setVariables(new Namespace(variables()));
        Iterator it=progVars.iterator();
        while (it.hasNext()) {
            variables().add((IProgramVariable)it.next());
        }
    }

    protected void unbindProgVars() {
        namespaces().setVariables(variables().parent());
    }

    protected static HashSet progVars(JavaBlock jb) {
        if (jb==JavaBlock.EMPTY_JAVABLOCK) {
            return new HashSet();
        }   
        DeclarationProgramVariableCollector pvc
            = new DeclarationProgramVariableCollector(jb.program());
        pvc.start();
        return pvc.result();
    }


    protected Term termForParsedVariable(ParsableVariable v) 
        throws antlr.SemanticException {
        if ( v instanceof LogicVariable ) {
            return tf.createVariableTerm((LogicVariable)v);
        } else if ( v instanceof ProgramVariable ) {
            return tf.createVariableTerm((ProgramVariable)v);
        } else if ( v instanceof Metavariable ) {
            return tf.createFunctionTerm((Metavariable)v);
        } else {
            throw new antlr.SemanticException(
                                              ""+ v +" is not a logic variable");
        }
    }

    public Term createAttributeTerm(Term prefix, 
                                    TermSymbol attribute,
                                    int shadowPrimes) {
        Term result = prefix;
        if (((ProgramVariable)attribute).isStatic()){
            result = tf.createVariableTerm((ProgramVariable)attribute);
        } else {
            if (shadowPrimes > 0) {
                Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
                result = tf.createShadowAttributeTerm((ProgramVariable)attribute, 
                                                      result, shadownum);
            } else {
                result = tf.createAttributeTerm((ProgramVariable)attribute, result);
            }
        }
        return result;
    }



    public TermSymbol getAttribute(Sort prefixSort, 
                                   String attributeName) 
        throws antlr.SemanticException{
        final JavaInfo javaInfo = getJavaInfo();
        TermSymbol result = null;

        if (attributeName.indexOf(':') != -1) {     
            result = javaInfo.getAttribute
                (attributeName, 
                 javaInfo.getKeYJavaType
                 (attributeName.substring(0, 
                                          attributeName.indexOf(':'))));
        } else {
            result = javaInfo.getAttribute(attributeName,(ObjectSort) prefixSort);
        }
        if ( result == null && !("length".equals(attributeName)) ) {
            throw new NotDeclException
                ("Attribute ", attributeName, getFilename(), 
                 getLine(), getColumn());
        }

        return result;
    }
    
    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
        String s=t.getText();
        try {
            return javaReader.readBlockWithProgramVariables(variables(), s);
        } catch (de.uka.ilkd.key.java.PosConvertException e) {
            lineOffset=e.getLine()-1;
            colOffset=e.getColumn()+1;
            throw new JavaParserException(e.getMessage(), t, 
                                          getFilename(), lineOffset, colOffset);                        
        } catch (de.uka.ilkd.key.java.ConvertException e) { 
            if (e.parseException()!=null) {               
                lineOffset=e.parseException().currentToken.next.beginLine;               
                colOffset=e.parseException().currentToken.next.beginColumn;
                e.parseException().currentToken.next.beginLine=getLine()-1;
                e.parseException().currentToken.next.beginColumn=getColumn();
                throw new JavaParserException(e.parseException().getMessage(), t, 
                                              getFilename(), -1, -1);  // row/columns already in text
            }       
            throw new JavaParserException(e.getMessage(), t, getFilename(), 0, 1);
        }
    }

    
    public void schema_var_decl(String name,Sort s,
                                boolean makeVariableSV,
                                boolean makeDependingSV,
                                boolean rigid,
                                boolean listModifier)
        {
            if (!skip_schemavariables) {
                SchemaVariable v;
                if ( s == Sort.FORMULA ) {
                    v = SchemaVariableFactory.createFormulaSV
                        (new Name(name), listModifier, rigid);
                } else if ( s instanceof ProgramSVSort ) {
                    v = SchemaVariableFactory.createProgramSV
                        (new ProgramElementName(name),(ProgramSVSort) s, listModifier);
                } else {
                    if ( makeVariableSV ) {
                        v=SchemaVariableFactory.createVariableSV
                            (new Name(name), s, listModifier);
                    } else if ( makeDependingSV ) {
                        v = SchemaVariableFactory.createDependingSV
                            (new Name(name), s, listModifier);
                    } else { v = SchemaVariableFactory.createTermSV
                            (new Name(name), s, listModifier, rigid);
                    }
                }
                if (inSchemaMode()) {
                    variables().add(v);
                }
            }

        }

    /** looks up a function, (program) variable or static query of the 
     * given name varfunc_id and the argument terms args in the namespaces 
     * and java info. 
     * @param args is null iff no argument list is given, for instance `f', 
     * and is an array of size zero, if an empty argument list was given,
     * for instance `f()'.
     */
    protected Operator lookupVarfuncId(String varfunc_name, Term[] args) 
        throws NotDeclException{
        // case 1: variable
        Operator v = (TermSymbol) variables().lookup(new Name(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        // case 2: function
        v = (Function) functions().lookup(new Name(varfunc_name));
        if (v != null) { // we allow both args==null (e.g. `c')
                         // and args.length=0 (e.g. 'c())' here 
            return v;
        }
        // case 3: static query
        if (varfunc_name.lastIndexOf("::")>0 && args!=null) {
            String classname = varfunc_name.substring
                (0, varfunc_name.lastIndexOf("::"));
            varfunc_name = varfunc_name.substring
                (varfunc_name.lastIndexOf("::")+2);
            v = getServices().getJavaInfo().getProgramMethodTerm
                (null, varfunc_name, args, classname).op();
            if (v != null) {
                return v;
            }
        }
        // case 4: program variable
        v = (ProgramVariable) programVariables().lookup
            (new ProgramElementName(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        if (args==null) {
            throw new NotDeclException
                ("(program) variable or constant", varfunc_name,
                 getFilename(), getLine(), getColumn());
        } else {
            throw new NotDeclException
                ("function or static query", varfunc_name,
                 getFilename(), getLine(), getColumn());
        }
    }

}
top {Term t;}
:t=formula;

formula returns [Term a = null]:a = term 
        {
            if ( a.sort() != Sort.FORMULA ) {
                throw new antlr.SemanticException
                ("Just Parsed a Term where a Formula was expected");
            }
        }
    ;

term returns [Term a = null]:a=term20 
    ;

term20 returns [Term a = null]{
    Term a1;
}
:a=term30 
        (EQV a1=term20 
            { a = tf.createJunctorTerm(Op.EQV, new Term[]{a, a1});} ) ?;

term30 returns [Term a = null]{
    Term a1;
}
:a=term40 
        (IMP a1=term30 
            { a = tf.createJunctorTerm(Op.IMP, new Term[]{a, a1});} ) ?;

term40 returns [Term a = null]{
    Term a1;
}
:a=term50 
        (OR a1=term40 
            { a = tf.createJunctorTerm(Op.OR, new Term[]{a, a1});} ) ?;

term50 returns [Term a = null]{
    Term a1;
}
:a=term60 
        (AND a1=term50
            { a = tf.createJunctorTerm(Op.AND, new Term[]{a, a1});} ) ?
    ;

term60 returns [Term a = null]{
    Term a1,a2;
    Sort s;
}
:NOT a1  = term60 { a = tf.createJunctorTerm(Op.NOT,new Term[]{a1}); }
    |   CTRL a1 = term60 { a = tf.createJunctorTerm(Op.CTRL,new Term[]{a1}); }
    |   a = bindingterm 
    |   a = dlTermOrTerm70
    ;

dlTermOrTerm70 returns [Term a = null]{}
:a = normal_dl_term
    |  (IDENT MODALITYARG)=> a = named_dl_term
    |  a = term70
    ;

term70 returns [Term a = null]{
    Term a1;
}
:a = accessterm
        // a term like {o:=u}x=y is parsed as {o:=u}(x=y)
        ((EQUALS)=>EQUALS a1 = logicTermReEntry
            { a = tf.createEqualityTerm(a, a1);}) ?

    ;

logicTermReEntry returns [Term result = null]{}
:(
            result = accessterm |
            result = update_or_substitution
        ) 
        {
            if (result.sort() == Sort.FORMULA) {
                throw 
                    new KeYSemanticException
                    ("Only terms may stand here.", getLine(), getColumn(), 
                     getFilename());
            }
        }          
    ;

classReference returns [String classReference = ""]{}
:id:IDENT {
            classReference = id.getText(); 
            while (isPackage(classReference)) {
                match(DOT);
                classReference += "." + LT(1).getText();
                match(IDENT);
            }      
            
            if (getJavaInfo().getTypeByClassName(classReference) == null) {
                throw new NotDeclException
                    ("class " + classReference + " is unknown.", 
                     classReference, getFilename(), getLine(), 
                     getColumn());
            }
        }  
    ;

static_attribute_suffix[String className] returns [Term result = null]{
    TermSymbol v = null;
    int shadowPrimes = 0;    
    String attributeName = "";
}
:DOT attributeName = attrid
        {   v = getAttribute(getTypeByClassName(className).
                             getSort(), attributeName); }   

        ((ps:PRIMES 
                { shadowPrimes = ps.getText().length();})?
            {
                result = createAttributeTerm(null, v, shadowPrimes);
            }
        )
    ;

attribute_or_query_suffix[Term prefix] returns [Term result = null]{
    TermSymbol v = null;
    int shadowPrimes = 0;
    result = prefix;
    String attributeName = "";    
}
:DOT 
        ((id LPAREN)=>( result = query[prefix])
        | 
            attributeName = attrid 
            {   v = getAttribute(prefix.sort(), attributeName); }   
            ((ps:PRIMES 
                    { shadowPrimes = ps.getText().length();})?

                {
                    result = createAttributeTerm(prefix, v, shadowPrimes);
                }        
            ))
    ;

query[Term prefix] returns [Term result = null]{
    String classRef = "";
    Term[] args;
    TermWithBoundVars[] argsWithBoundVars = new TermWithBoundVars[0];
}
:classRef = classReference COLON COLON mid:IDENT argsWithBoundVars = argument_list
        { 
            args = new Term[argsWithBoundVars.length];
            for (int i=0; i<argsWithBoundVars.length; i++) {
                args[i] = argsWithBoundVars[i].term;
            }
            try {
                result = getServices().getJavaInfo().getProgramMethodTerm
                (prefix, mid.getText(), args, classRef);
            } catch (IllegalArgumentException e) {
                throw new antlr.SemanticException(e.getMessage());
            }
        };

accessterm returns [Term result = null]{
    KeYJavaType kjt = null;
    Sort arraySort = null;
    Term indexTerm = null;
    int shadowPrimes = 0;
    String className = null;
}
:(LPAREN logicTermReEntry (COMMA | GUARD)) => result = conditional
    |
        {"-".equals(LT(1).getText()) && 
	!TermParser.isNumber(LT(2).getText())}? neg:ARITH_OP result = logicTermReEntry
        {
            if (result.sort() != Sort.FORMULA) {
                result = tf.createFunctionTerm
                ((Function) functions().lookup(new Name("~m")), result);
            } else {
                throw new KeYSemanticException
                ("Formula cannot be prefixed with: ", neg, getFilename());
            }
        } 
    | ({LA(2) == DOT && ( isPackage ( LT(1).getText() ) || 
                ( kjt = getTypeByClassName(LT(1).getText()))!=null )}?

        className = classReference
        result = static_attribute_suffix[className]
      | 	
        result = term80)
        (LBRACKET indexTerm = logicTermReEntry
            ( RBRACKET
            | rb:RBRACKETPRIMES {
                    shadowPrimes = rb.getText().length()-1;                 
                }
            )
            {
                if (arraySort != null && 
                    arraySort != ProgramSVSort.VARIABLE) {
                    if (!(arraySort instanceof ArraySort)){
                        throw new NotDeclException
                            ("array "+result+" declared with a different"+
                             " dimension than required here", ""+result, 
                             getFilename(), getLine(), getColumn());
                    } else {
                        arraySort = ((ArraySort) arraySort).elementSort();
                    } 
                } else if (arraySort == ProgramSVSort.VARIABLE && !inSchemaMode()) {
                    throw new antlr.SemanticException
                        ("Schemavariable are not allowed in this context.");
                }
                if (shadowPrimes>0) {
                    result = tf.createArrayTerm
                        (Op.SHADOW_ARRAY, result, indexTerm, 
                         TermParser.toZNotation("" + shadowPrimes, functions(), tf));
                } else {
                    result = tf.createArrayTerm(Op.ARRAY, result, indexTerm);
                }   
            }
        | result = attribute_or_query_suffix[result]
        )*
    ;

term80 returns [Term a = null]:a = funcpredvarterm
    |   a = specialTerm
    |   LPAREN a = term RPAREN 
    |   "true"  { a = tf.createJunctorTerm(Op.TRUE); }
    |   "false" { a = tf.createJunctorTerm(Op.FALSE); }
    ;

abbreviation returns [Term a=null]{ 
    String sc = null;
}
:(   sc = id
            {
                a =  scm.getTerm(sc);
                if(a==null){
                    throw new NotDeclException
                        ("abbreviation", sc, 
                         getFilename(), getLine(), getColumn());
                }                                
            }
        )
    ;

conditional returns [Term result = null]{ 
    Term comparator = null;
}
:LPAREN comparator = logicTermReEntry 
        ( result=normalConditional[comparator] | result=conjunctionConditional[comparator])
    ;

normalConditional[Term cond1] returns [Term result = null]{ 
    Term cond2 = null;
    Term c_then = null;
    Term c_else = null;
}
:GUARD cond2 = logicTermReEntry RPAREN
        (
            DOT result = shortCond[cond1, cond2]
        | 
            c_then = logicTermReEntry COLON c_else = logicTermReEntry {
                result = tf.createIfElseTerm(cond1, cond2, c_then, c_else);
            }

        ) 
    ;

shortCond[Term cond1, Term cond2] returns [Term result = null]{
    ParsableVariable v=null;
    Term c_then = null;
}
:DOT v = varId CONDUP c_then = logicTermReEntry {
            result = tf.createIfElseTerm((ProgramVariable)v, cond1, 
                                         cond2, c_then);
        }
    ;

conjunctionConditional[Term cond1_1] returns [Term a = null]{ 
    Term cond1_2 = null;
    Term cond2_1 = null;
    Term cond2_2 = null;
    Term c_then = null;
    Term c_else = null;
}
:COMMA cond1_2=logicTermReEntry COLON c_else=logicTermReEntry 
        GUARD cond2_1 = logicTermReEntry 
        COMMA cond2_2=logicTermReEntry COLON c_then=logicTermReEntry 
        RPAREN
        {
            a = tf.createConjCondTerm(new Term[]{cond1_1, cond1_2, c_else, 
                                                 cond2_1, cond2_2, c_then});
        }        
    ;

bindingterm returns [Term a = null]:a = local_bindingterm             
    ;

argument returns [TermWithBoundVars p = null]{
    Sort s = null;
    List listVars = new LinkedList();
    LogicVariable v = null;
    LogicVariable[] arrayVars = null;
    ArrayOfQuantifiableVariable vars = null;
    Term a1;
}
:(
            BACKSLASH v0:IDENT COLON s=sortId
            {
                v = bindVar(v0, s); 
                listVars.add(v);
            }
        )* 
        SLASH a1=term
        {
            arrayVars = (LogicVariable[])listVars.toArray(new LogicVariable[listVars.size()]);
            vars = new ArrayOfQuantifiableVariable(arrayVars);
            p = new TermWithBoundVars(vars, a1);
            unbindVars();
        }
    | a1=term
        {
            p = new TermWithBoundVars(vars, a1);
        }
    ;

local_bindingterm returns [Term a = null]{
    Operator op = null;
    LogicVariable v = null;
    Term a1;
    Sort s;
}
:(
            "all" { op = Op.ALL; }
        |   "ex"  { op = Op.EX;  }
        )
        v0:IDENT COLON s=sortId DOT {v = bindVar(v0,s);} a1 = term60
        {
            a = tf.createQuantifierTerm((Quantifier)op, v,a1);
            unbindVars();
        }
    | a = update_or_substitution
    ;

update_or_substitution returns [Term result = null]{
}
:( LBRACE IDENT COLON ) => result = substitutionterm |
        result = updateterm
    ;

substitutionterm returns [Term result = null]{ Sort s;
    LogicVariable v = null;
    Term a1, a2;
    boolean waryAll = false;
    boolean waryEx = false;
}
:LBRACE v1:IDENT COLON s=sortId a1=logicTermReEntry
        ( "waryAll" { waryAll = true; } |
            "waryEx"  { waryEx  = true; } )?
        RBRACE 
        {v = bindVar(v1, s);} 
        a2 = term60 { 
            if ( waryAll )
                result = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
                                                     v, a1, a2 );
            else if ( waryEx )
                result = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
                                                     v, a1, a2 );
            else
                result = tf.createSubstitutionTerm ( Op.SUBST,
                                                     v, a1, a2 );
            unbindVars();
        }
    ;

updateterm returns [Term result = null]{ 
    SingleUpdateData sud = null;
    List subs = new LinkedList();  
    Term a2 = null;
}
:LBRACE
        sud = singleupdate {
            subs.add(sud.a0); 
            subs.add(sud.a1); 
        } ( COMMA sud = singleupdate {
                subs.add(sud.a0); 
                subs.add(sud.a1); 
            } )* 
        RBRACE a2 = term60 {
            subs.add(a2);
        }
        { 
            result = tf.createUpdateTerm((Term[])subs.toArray(new Term[subs.size()]));
        }
    ;

singleupdate returns [SingleUpdateData sud=null]{
    Term a0 = null;
    Term a1 = null;
    ArraySort as;
    TermSymbol pv;
    LhsUpdateData lud = null;
    boolean shadow_l = false;
    boolean shadow_r = false;
}
:(a0 = lhsSingle ) ASSIGN (a1 = rhsSingle) {
            sud = new SingleUpdateData();
            sud.a0 = a0;
            sud.a1 = a1;
        }
    ;

rhsSingle returns [Term result = null]{}
:result = logicTermReEntry
    ;

lhsSingle returns [Term result = null]{
    Term a1 = null;
}
:result = logicTermReEntry
        { 
            if (!(result.sort() instanceof ProgramSVSort) &&
                result.op() != Op.META_SHADOW &&
                !(result.op() instanceof ProgramVariable || 
                  result.op() instanceof AccessOp)) {
                throw new KeYSemanticException
                ("Term of sort " + result.sort() + 
                 " cannot stand for a location.",
                 getLine(), getColumn(), getFilename());
            }
        }
    ;

normal_dl_term returns [Term a = null]{   JavaBlock jb=null;
    Term a1;
}
:d1:DIAMOND 
        {
            jb=getJavaBlock(d1);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60
        {
            a = tf.createDiamondTerm(jb, a1);
            unbindProgVars();
        }
    |
        d2:BOX
        {
            jb=getJavaBlock(d2);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60   
        {
            a = tf.createBoxTerm(jb, a1);
            unbindProgVars();
        }
    |
        d3:THROUGHOUT
        {
            jb=getJavaBlock(d3);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60
        {
            a = tf.createProgramTerm(Op.TOUT,jb, a1);
            unbindProgVars();
        }
    ;

named_dl_term returns [Term a = null]{
    TermWithBoundVars[] argsWithBoundVars = new TermWithBoundVars[0];
    Term[] terms = null;
    Term term = null;
    JavaBlock jb = null;
}
:id:IDENT
        bl:MODALITYARG ( 
            {
                jb = getJavaBlock(bl);               
                bindProgVars(progVars(jb));
            }
            argsWithBoundVars = argument_list
            {   
                unbindProgVars();
                terms = new Term[argsWithBoundVars.length];
                for (int i=0; i<argsWithBoundVars.length; i++) {
                    terms[i] = argsWithBoundVars[i].term;
                }
                Operator op = null;
                                if (id.getText().charAt(0)=='#') {
                    if (!inSchemaMode()) {
                        throw new KeYSemanticException
                            ("No schema elements allowed outside taclet declarations.",
                             id, getFilename());
                    }
                    op = (SchemaVariable)variables().lookup(new Name(id.getText()));
                } else {
                    op = Op.getModality(id.getText());
                }
                if(op == null) {
                    throw new KeYSemanticException("Unknown modal operator.", id, getFilename());
                }
                if(op.arity() != terms.length) {
                    throw new KeYSemanticException
                        ("The arity of " + op.name() +
                         " is "+op.arity() + ", but you provided " + terms.length + " arguments",
                         getLine(), getColumn(), getFilename());
                }
                a = tf.createProgramTerm(op, jb, terms);
            }
        )
    ;

argument_list returns [TermWithBoundVars[] ts = null]{
    List args = new LinkedList();
    TermWithBoundVars p1, p2;
}
:LPAREN 
        (p1 = argument { args.add(p1);  }

            (COMMA p2 = argument { args.add(p2); } )* )?

        RPAREN
        {
            ts = (TermWithBoundVars[]) args.toArray(new TermWithBoundVars[args.size()]);
        }

    ;

funcpredvarterm returns [Term a = null]{
    Term[] arguments = null;
    TermWithBoundVars[] argsWithBoundVars = null;
    String varfuncid;
    String neg = "";
}
:ch:CHAR_LITERAL {
            String s = ch.getText();
            int intVal = 0;
            if (s.length()==3) {
                intVal = (int)s.charAt(1);
            } else {
                try {
                    intVal = Integer.parseInt(s.substring(3,s.length()-1),16);
                } catch (NumberFormatException ex) {
                    throw new KeYSemanticException
                        ("Expected '"+s+"' to be a character", ch, getFilename());
                }       
            }
            a = tf.createFunctionTerm((Function) functions().lookup(new Name("C")), 
                                      TermParser.toZNotation(""+intVal, functions(), tf).sub(0));
        }
    | { ("-".equals(LT(1).getText()) && TermParser.isNumber(LT(2).getText())) 
	|| 
        (TermParser.isNumber(LT(1).getText()) && !"(".equals(LT(2).getText()))}? 
	({"-".equals(LT(1).getText())}? ARITH_OP {neg = "-";})?
         number:IDENT     
        { a = TermParser.toZNotation(neg+number.getText(), functions(), tf);}    
    | AT a = abbreviation
    | varfuncid = id

        ((LPAREN)=>argsWithBoundVars = argument_list)? 
        //argsWithBoundVars==null indicates no argument list
        //argsWithBoundVars.length==0 indicates open-close-parens ()
        {   
            ArrayOfQuantifiableVariable vars = null;
            if (argsWithBoundVars != null) {
                arguments = new Term[argsWithBoundVars.length];           
                for (int i=0; i<argsWithBoundVars.length; i++) {
                    arguments[i] = argsWithBoundVars[i].term;

                    //For OCL simplification
                    if (argsWithBoundVars[i].boundVars != null) {
                        if ((i == 1 && argsWithBoundVars[i].boundVars.size() == 1)
                            || (i == 2 && argsWithBoundVars[i].boundVars.size() == 2)) {
                            vars = argsWithBoundVars[i].boundVars;
                        } else {
                            throw new antlr.SemanticException("Parsing error: The " + varfuncid + 
                                " expression does not match any OCL collection operation!");
                        }
                    }
                    //
                }
            }
            Operator op = lookupVarfuncId(varfuncid, arguments);
            if (op instanceof ParsableVariable) {
                a = termForParsedVariable((ParsableVariable)op);
            } else  if (op instanceof TermSymbol) {
                if (arguments==null) {
                    arguments = AN_ARRAY_OF_TERMS;
                }
                if (vars == null) {
                    a = tf.createFunctionTerm((TermSymbol)op, arguments);
                }

                //For OCL simplification
                else {
                    if (vars.size() == 2) {
                        a = tf.createTernaryBindingTerm(
                                                        (TermSymbol)op, 
                                                        vars.getQuantifiableVariable(0), 
                                                        vars.getQuantifiableVariable(1), 
                                                        arguments);
                    } else {
                        a = tf.createBinaryBindingTerm(
                                                       (TermSymbol)op, 
                                                       vars.getQuantifiableVariable(0),  
                                                       arguments);
                    }
                } 
                //
            }
        }
    ;

attrid returns [String attr = "";]{}
:({ isPackage(LT(1).getText()) || (LA(2) == COLON && LA(3) == COLON) }? 
            attr = classReference COLON COLON {attr += "::";} )?
        attrID:IDENT {attr += attrID.getText();}
    ;

specialTerm returns [Term result = null]{
    Operator vf = null;
}
:vf = expr_op 
        {   
            result = tf.createFunctionTerm((Function)vf, AN_ARRAY_OF_TERMS);
        }                 
    ;

expr_op returns [Operator v = null]{
    ParsableVariable left_var0 = null;
    ParsableVariable right_var0 = null;
    ParsableVariable unary_var0 = null;
}
:IN_TYPE LPAREN ((left_var0=varId ((id:ARITH_OP | id1:SLASH | id2:IDENT) right_var0 = varId)?) |
            ( "-" unary_var0 = varId) )  RPAREN
        {  
            if (id==null) id=id1;   
            if (unary_var0!=null) {
                // can only be unary minus
                SortedSchemaVariable unary_var = (SortedSchemaVariable) unary_var0;
                Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
                return new InTypeOperator
                    (type, new de.uka.ilkd.key.java.expression.operator.Negative(unary_var));
            }
            else {
                SortedSchemaVariable left_var = (SortedSchemaVariable) left_var0;
                Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
                if (right_var0 != null) {
                    SortedSchemaVariable right_var = (SortedSchemaVariable) right_var0;
                    String op_str = (id == null ? id2.getText() : id.getText());
                    if ("+".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Plus(left_var, right_var));
                    } else if ("-".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Minus(left_var, right_var));
                    } else if ("*".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Times(left_var, right_var));
                    } else if ("/".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Divide(left_var, right_var));
                    } else if ("%".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Modulo(left_var, right_var));
                    } 
                }
                else {
                    return new InTypeOperator(type, left_var);
                }
            }
        }
    | IS_OBJECT  LPAREN (left_var0=varId )  RPAREN
        {
            SortedSchemaVariable var = (SortedSchemaVariable) left_var0;
            return new IsObject(de.uka.ilkd.key.logic.sort.Sort.FORMULA, var);
        }
    ;

varId returns [ParsableVariable v = null]:id:IDENT 
        {   
            v = (ParsableVariable) variables().lookup(new Name(id.getText()));
            if (v == null) {
                throw new NotDeclException("variable", id, 
                                           getFilename());
            }
        }
    ;

// inherited from grammar DeclParser
decls :(INCLUDE 
            ( LDTS COLON LBRACE include[true] (COMMA include[true])* RBRACE
            |
                include[false]
            )
            SEMI)*
        {
            if(parse_includes) return;
            activatedChoices = SetAsListOfChoice.EMPTY_SET;  
        }
        ("with" LPAREN activated_choice (COMMA activated_choice)* RPAREN)?
        {if(onlyWith) return;}
        (
            option_decls
        |    
            sort_decls
        |
            prog_var_decls
        |
            schema_var_decls
        |
            pred_decls
        |
            func_decls
        |
            heur_decls

        ) *
    ;

// inherited from grammar DeclParser
include[boolean ldt] :(absfile:IDENT{ 
                if(parse_includes){
                    addInclude(absfile.getText(),false,ldt);
                }
            }
        | relfile:STRING_LITERAL{ 
                if(parse_includes){
                    addInclude(relfile.getText(),true,ldt);
                }
            })
    ;

// inherited from grammar DeclParser
activated_choice {
    String name;
    Choice c;
}
:cat:IDENT COLON choice:IDENT
        {if(usedChoiceCategories.contains(cat.getText())){
            throw new IllegalArgumentException("You have already chosen a different option for "+cat.getText());
        }
        usedChoiceCategories.add(cat.getText());
        name = cat.getText()+":"+choice.getText();
        c = (Choice) choices().lookup(new Name(name));
        if(c==null){
            throw new NotDeclException("Option", choice,
                                       getFilename());
        }else{
            activatedChoices=activatedChoices.add(c);
            namespaces().setFunctions(functions().extended(c.funcNS().allElements()));
        }
        }
    ;

// inherited from grammar DeclParser
option_decls {
}
:"options" LBRACE choice (SEMI choice)* RBRACE 
    ;

// inherited from grammar DeclParser
choice {
    String cat=null;
}
:category:IDENT {cat=category.getText();} (COLON LBRACE choice_option[cat] (COMMA choice_option[cat])* RBRACE)? 
        {
            if(!category2Default.containsKey(cat)){
                choices().add(new Choice("On",cat));
                choices().add(new Choice("Off",cat)); 
                category2Default.put(cat, cat+":On");               
            }
        }
    ;

// inherited from grammar DeclParser
choice_option[String cat] {
    String name;
}
:choice:IDENT { name=cat+":"+choice.getText();
        Choice c = (Choice) choices().lookup(new Name(name));
        if(c==null){
            c = new Choice(choice.getText(),cat);
            choices().add(c);
        }
        if(!category2Default.containsKey(cat)){
            category2Default.put(cat, name);
        }
        }
    ;

// inherited from grammar DeclParser
sort_decls {
    boolean isObjectSort=false;
    boolean isGenericSort=false;

    //For OCL simplification
    boolean isOclSort = false;

    Sort[] sortExt=new Sort [0];
    Sort[] sortOneOf=new Sort [0];
    ListOfName sortIds;
}
:"sorts" 
        LBRACE ( 
            { 
                isObjectSort=false; 
                isGenericSort=false; 
                isOclSort = false;
                sortExt = new Sort[0];
                sortOneOf = new Sort[0];
            }
            ( "object" {isObjectSort=true;} sortIds = sortIdentifiers
            | "generic" {isGenericSort=true;} sortIds = sortIdentifiers
                ( "oneof" sortOneOf = oneof_sorts )? 
                ("extends" sortExt = extends_sorts )?

            | "ocl" {isOclSort=true;} sortIds = sortIdentifiers

            | sortIds = sortIdentifiers

            )
            SEMI
            {   
                if (!skip_sorts) {
                    Name sort_name;
                    IteratorOfName it = sortIds.iterator ();
                    while ( it.hasNext () ) {
                        sort_name = it.next ();

                        if (sorts().lookup(sort_name) == null) {
                            Sort s;
                            if (isObjectSort) {
                                if ("Object".equals(sort_name)) {
                                    if (sorts().lookup(new Name("Object")) == null) {
                                        s = new ClassInstanceSortImpl(sort_name);
                                    }
                                    s=(Sort)sorts().lookup(new Name("Object"));
                                } else {
                                    s = new ClassInstanceSortImpl(sort_name, 
                                                                  (Sort)sorts().lookup
                                                                  (new Name("Object")));
                                }	
                            }                            
                            //Just constant for now
                            else if (isOclSort) {
                               s = Sort.OCL_ANY;
                            }
                            //
                            else if (isGenericSort) {
                                int        i;
                                SetOfSort  ext   = SetAsListOfSort.EMPTY_SET;
                                SetOfSort  oneOf = SetAsListOfSort.EMPTY_SET;

                                for ( i = 0; i != sortExt.length; ++i )
                                    ext = ext.add ( sortExt[i] );

                                for ( i = 0; i != sortOneOf.length; ++i )
                                    oneOf = oneOf.add ( sortOneOf[i] );

                                try {
                                    s = new GenericSort ( sort_name,
                                                          ext,
                                                          oneOf );
                                } catch ( GenericSupersortException e ) {
                                    throw new GenericSortException ( "sort",
                                                                     "Illegal sort given", e.getIllegalSort (),
                                                                     getFilename (), getLine (), getColumn () );
                                }
                            } else {
                                s = new PrimitiveSort(sort_name);
                            }
                            sorts().add(s);
                            if (s instanceof NonCollectionSort) {
                                NonCollectionSort ns
                                    = (NonCollectionSort)s;
                                sorts().add(ns.getSetSort());
                                sorts().add(ns.getSequenceSort());
                                sorts().add(ns.getBagSort());
                            }
                            if ( s instanceof SortDefiningSymbols )
                                ((SortDefiningSymbols)s).createSymbols ( defaultChoice.funcNS(),
                                                                         sorts() );
                        }

                    }
                }
            }
            
        )* 
        RBRACE
    ;

// inherited from grammar DeclParser
sortIdentifiers returns [ListOfName ids = SLListOfName.EMPTY_LIST]:id:IDENT { ids = ids.append ( new Name(id.getText()) ); }
        (
            COMMA
            id2:IDENT { ids = ids.append ( new Name(id2.getText()) ); }
        ) *
    ;

// inherited from grammar DeclParser
extends_sorts returns [Sort[] extendsSorts = null]{
    List args = new LinkedList();
    Sort s;
}
:s = any_sortId_check[true] { args.add(s); }
        (
            COMMA s = any_sortId_check[true] {args.add(s);}
        ) *
        {
            extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

// inherited from grammar DeclParser
oneof_sorts returns [Sort[] oneOfSorts = null]{
    List args = new LinkedList();
    Sort s;
}
:LBRACE
        s = sortId_check[true] { args.add(s); }
        (
            COMMA s = sortId_check[true] {args.add(s);}
        ) *
        RBRACE
        {
            oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

// inherited from grammar DeclParser
prog_var_decls {
    String var_name;
    String type = null;
}
:{ switchToNormalMode(); }
        "program" "variables"
        LBRACE 
        (
            ((v0:IDENT {type= "" + v0.getText();} (DOT v1:IDENT {type= type + "." + v1.getText();})*)
              |
             (pid:PIDENT {type = pid.getText();} )
            )
            (LBRACKET RBRACKET {
               type = type + "[]";
            })?
            //sortId_check[true]
            var_name = id
            {
                KeYJavaType kjt = getJavaInfo().getKeYJavaType(type);
		ProgramElementName pvName = new ProgramElementName(var_name);
                if (kjt == null) {
                    throw new antlr.SemanticException
                        ("Program Variable " + var_name + 
                         " declared with unknown type " + 
                         type);

                } else if (lookup(pvName)!=null) {
		  // commented out as pv do not have unique name (at the moment)
		    //throw new AmbigiousDeclException
		//	(var_name, getFilename(), getLine(), getColumn()); 
		}
            namespaces().programVariables().add(new ProgramVariable
                (new ProgramElementName(var_name), kjt));

            }
            SEMI
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
schema_var_decls {
    Sort s = null;
    boolean makeVariableSV  = false;
    boolean makeDependingSV = false; // skolem functions
    boolean rigid           = false;
    boolean listModifier    = false;
    String type = "";    
}
:"schema" "variables"
        LBRACE  { switchToSchemaMode(); }
        (
            ( "modal" "operator" one_schema_modal_decl (COMMA one_schema_modal_decl)* SEMI )
        |
            {makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;}
            ( ( "program" ("svlist" {listModifier=true;})? {type = "";}
                ((IDENT IDENT) => id1:IDENT{type += " "+id1.getText();}
                 | id2:IDENT{ type = type.substring(1,type.length());
                              s = (Sort) DeclParser.name2sort.get(type);
                              if (s == null) {
                                  throw new antlr.SemanticException
                                  ("SchemaVariable of sort "+type+ " not found.");
                              }
                              schema_var_decl(id2.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);} 
                )+
                (COMMA id3:IDENT{schema_var_decl(id3.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);})* SEMI
               )
               | "formula"
                  ( "rigid" { rigid = true; } ) ? {s = Sort.FORMULA;type = ""; }
		  (id5:IDENT {schema_var_decl(id5.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);})
		  ( COMMA (id6:IDENT {schema_var_decl(id6.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);}))* SEMI
               | s = any_sortId_check[true]
                  ( "variables" { makeVariableSV = true; }
                    | "depending" { makeDependingSV = true; }
                    | "rigid" { rigid = true; } ) ? {type = "";}
                  (id7:IDENT {schema_var_decl(id7.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);})
	          ( COMMA (id8:IDENT {schema_var_decl(id8.getText(),s,makeVariableSV,makeDependingSV,rigid,listModifier);}))* SEMI
             )
        ) *
        { switchToNormalMode(); }
        RBRACE
    ;

// inherited from grammar DeclParser
one_schema_modal_decl {
    Modality[] mods;
}
:id:IDENT EQUALS mods=modalities_decl {
            SchemaVariable msv = (SchemaVariable)variables().lookup(new Name(id.getText()));
            if(msv != null)
            throw new antlr.SemanticException("Schema variable "+id.getText()+ " already defined");
            int arity = mods[0].arity();
            for(int i=1;i<mods.length;i++)
            if(arity != mods[i].arity())
            throw new antlr.SemanticException("Arity mismatch for schema variable "+id.getText());
            msv = SchemaVariableFactory.createModalOperatorSV(new Name(id.getText()), arity, mods);
            if (inSchemaMode()) {
                variables().add(msv);
            }
        }
    ;

// inherited from grammar DeclParser
modalities_decl returns [Modality[] modalities = null]{
    List args = new LinkedList();
}
:(LBRACE 
            s1:IDENT {
                if(s1.getText().charAt(0) == '#') {
                    ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s1.getText()));
                    if(msv == null)
                        throw new antlr.SemanticException("Schema variable "+s1.getText()+ " not defined");
                    for(int i=0; i<msv.modalities().length; i++)
                        args.add(msv.modalities()[i]);
                }else{
                    Modality op = Op.getModality(s1.getText());
                    if(op == null)
                        throw new antlr.SemanticException("Unrecognised modal operator "+s1.getText());
                    args.add(op);
                }
            } (COMMA s2:IDENT {
                    if(s2.getText().charAt(0) == '#') {
                        ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s2.getText()));
                        if(msv == null)
                            throw new antlr.SemanticException("Schema variable "+s2.getText()+ " not defined");
                        for(int i=0; i<msv.modalities().length; i++)
                            args.add(msv.modalities()[i]);
                    } else {
                        Modality op = Op.getModality(s2.getText());
                        if(op == null)
                            throw new antlr.SemanticException("Unrecognised modal operator "+s2.getText());
                        args.add(op);
                    }
                })* RBRACE
        ) {
            modalities = (Modality[])args.toArray(new Modality[args.size()]);
        }
    ;

// inherited from grammar DeclParser
pred_decls {
    Sort[] argSorts;
    String pred_name;
    Namespace oldFNS=null;
    String ctext = null;
    Choice c=null;
}
:"predicates" (LPAREN cat:IDENT COLON choice:IDENT RPAREN)?
        { if(cat!=null){
            ctext = cat.getText()+":"+choice.getText();
            c = (Choice) choices().lookup(new Name(ctext));
            if(c==null){
                throw new NotDeclException("Option", choice,
                                           getFilename());
            }else{
                namespaces().setFunctions(c.funcNS());
            }
        }else{
            namespaces().setFunctions(defaultChoice.funcNS());
        }
        cat=null;
        }
        LBRACE
        (
            pred_name = id
            argSorts=arg_sorts[!skip_predicates]
            {
                if (!skip_predicates) {
                    Name predicate = new Name(pred_name);		    		    
                    if (lookup(predicate) != null) {
                        throw new AmbigiousDeclException
                        (pred_name, getFilename(), getLine(), getColumn());
                    } else {
                        Function p = new Function(predicate,
                                                  Sort.FORMULA, 
                                                  argSorts);
                        functions().add(p);         
                    }
                } 
            }
            SEMI
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
func_decls {
    Sort[] argSorts;
    Sort retSort;
    String func_name;
    Namespace oldFNS=null;
    String ctext = null;
    Choice c=null;
}
:"functions" (LPAREN cat:IDENT COLON choice:IDENT RPAREN)?
        { 
            if(cat!=null){
                ctext = cat.getText()+":"+choice.getText();
                c = (Choice) choices().lookup(new Name(ctext));
                if(c==null){
                    throw new NotDeclException("Option", choice,
                                               getFilename());
                }
            }else{
                c=defaultChoice;
            }
            namespaces().setFunctions(c.funcNS());
            cat=null;
        }
        LBRACE
        (
            retSort = sortId_check[!skip_functions]
            func_name = id 
            argSorts = arg_sorts[!skip_functions]
            {
                if (!skip_functions) {
                    Name fct_name = new Name(func_name);
                    if (lookup(fct_name) != null) {
                        throw new AmbigiousDeclException
                        (func_name, getFilename(), getLine(), getColumn());
                    } else {
                        functions().add(new Function(fct_name,
     	                                             retSort, 
        	                                     argSorts));
                    }
                } 
            }
            SEMI
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
arg_sorts[boolean checkSort] returns [Sort[] argSorts = null]{
    List args = new LinkedList();
    Sort s;
}
:(
            LPAREN
            s = sortId_check[checkSort] { args.add(s); }
            (
                COMMA s = sortId_check[checkSort] {args.add(s);}
            ) *
            RPAREN
        ) ?
        {
            argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

// inherited from grammar DeclParser
integer_num returns [ int num = 0 ]
options {
	defaultErrorHandler=false;
}
:{
            try {
                Token t = LT(1);
                num = Integer.decode ( t.getText () ).intValue ();
                consume ();
            } catch ( TokenStreamException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            } catch ( NumberFormatException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            } catch ( NullPointerException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            }
        }
    ;

// inherited from grammar DeclParser
heur_decls {
    int     t_priority = 0;
    Integer priority   = null;
}
:"heuristics"
        LBRACE
        (
            { priority = null; }
            id:IDENT
            (
                LPAREN t_priority = integer_num RPAREN 
                { priority = new Integer ( t_priority ); }
            ) ?
            SEMI
            { 
                if (!skip_heuristics) {
                    Heuristic h = new Heuristic(new Name(id.getText()));
                    if(heuristics().lookup(new Name(id.getText()))==null){
                        heuristics().add(h);
                    }
                    if ( priority != null )
                        heur_prioritization.setPriority ( h, priority.intValue () );
                }
            }
        ) *
        RBRACE
    ;

// inherited from grammar DeclParser
sortId returns [Sort s = null]:s = sortId_check[true]
    ;

// inherited from grammar DeclParser
sortId_check[boolean checkSort] returns [Sort s = null]{
    Sort t;
}
:t = sortId_check_help[checkSort]
        s = array_decls[t]
    ;

// inherited from grammar DeclParser
any_sortId_check[boolean checkSort] returns [Sort s = null]{
    Sort t;
}
:t = any_sortId_check_help[checkSort]
        s = array_decls[t]
    ;

// inherited from grammar DeclParser
any_sortId_check_help[boolean checkSort] returns [Sort s = null]{
    String name;
}
:name = id
        {   s= (Sort) sorts().lookup(new Name(name));
            if (checkSort && s == null) {
                throw new NotDeclException("sort", name, 
                                           getFilename(), 
                                           getLine(), getColumn());
            }
        }
    ;

// inherited from grammar DeclParser
sortId_check_help[boolean checkSort] returns [Sort s = null]:s = any_sortId_check_help[checkSort]
        {
            // don't allow generic sorts or collection sorts of
            // generic sorts at this point
            Sort t = s;
            while ( t instanceof CollectionSort )
            t = ((CollectionSort)t).elementSort ();
            if ( t instanceof GenericSort )
            throw new GenericSortException ( "sort",
                                             "Non-generic sort expected", t,
                                             getFilename (), getLine (), getColumn () );
        }
    ;

// inherited from grammar DeclParser
array_decls[Sort p] returns [Sort s = null]{
    int  n=0;
}
:(LBRACKET RBRACKET {n++;})*

        { 
            if(n != 0){
                s = ArraySortImpl.getArraySortForDim(
                    p, n, getJavaInfo().getObjectSort(),
                    getJavaInfo().getCloneableSort());
            }else{
                s = p;
            }
        }
            
    ;

// inherited from grammar DeclParser
id_declaration returns [ IdDeclaration idd = null ]{
    Sort s = null;
}
:id:IDENT
        ( COLON s = sortId_check[true] ) ?
        {
            idd = new IdDeclaration ( id.getText (), s );
        }
    ;

// inherited from grammar DeclParser
id returns [String name = null]:(prefix:IDENT 
	{ name = prefix.getText(); }
        | pprefix : PIDENT { name = pprefix.getText(); })
        ((COLON COLON)=>COLON COLON nameID:IDENT 
            { 
                name += "::" + nameID.getText();
            }
        )?
    ;


